package refit.agreement.pbft;

import refit.agreement.pbft.REFITPBFTMessages.REFITPBFTCommit;
import refit.agreement.pbft.REFITPBFTMessages.REFITPBFTPrePrepare;
import refit.agreement.pbft.REFITPBFTMessages.REFITPBFTPrepare;
import refit.agreement.pbft.REFITPBFTTransition.REFITPBFTTransitionProtocolInstance;
import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITMessage;
import refit.message.REFITMessageType;
import refit.message.REFITNoOpRequest;
import refit.message.REFITOrderRequest;
import refit.replica.order.*;
import refit.util.REFITAssert;


public class REFITPBFT {

	// ############
	// # PROTOCOL #
	// ############

	public static class REFITPBFTProtocol extends REFITOrderProtocol {

		public static final REFITPBFTProtocol instance = new REFITPBFTProtocol();

		@Override
		public REFITOrderProtocolInstance createInstance(REFITOrderStageSlot slot) {
			return new REFITPBFTProtocolInstance(slot, this);
		}

		@Override
		public REFITOrderProtocolTransitionInstance createTransitionInstance(REFITOrderStageSlot slot) {
			return new REFITPBFTTransitionProtocolInstance(slot, this);
		}

	}


	// ############################
	// # PROTOCOL INSTANCE STATES #
	// ############################

	public enum REFITPBFTProtocolInstanceState {
		INITIALIZED,
		PRE_PREPARED,
		PREPARED,
		COMMITTED
	}


	// #####################
	// # PROTOCOL INSTANCE #
	// #####################

	public static class REFITPBFTProtocolInstance extends REFITOrderProtocolInstance {

		protected final REFITOrderProtocol protocol;
		protected final REFITPBFTCertificate certificate;
		protected boolean[] activeOrderRecipients;
		protected boolean[] allOrderReplicas;
		private REFITPBFTProtocolInstanceState instanceState;
		private boolean isPrimary;
		private REFITOrderRequest proposal;

		public REFITPBFTProtocolInstance(REFITOrderStageSlot slot, REFITOrderProtocol protocol) {
			super(slot);
			this.protocol = protocol;
			this.certificate = new REFITPBFTCertificate(REFITConfig.ORDER_GROUP_SIZE);
			REFITAssert.assertTrue(REFITConfig.AUTHENTICATE_MESSAGES, "REFITPBFT requires authenticated messages");
		}

		public REFITPBFTProtocolInstance(REFITOrderStageSlot slot, long instanceID, int viewID, REFITOrderRequest proposal, REFITOrderProtocol protocol) {
			super(slot);
			this.protocol = protocol;
			this.certificate = new REFITPBFTCertificate(REFITConfig.ORDER_GROUP_SIZE);
			init(instanceID, viewID);
			this.proposal = proposal; // Must be after init()
		}


		// ##################
		// # HELPER METHODS #
		// ##################

		@Override
		public String toString() {
			return "PBFT[" + instanceID + "]";
		}

		private void advanceState(REFITPBFTProtocolInstanceState state) {
			if (REFITLogger.LOG_ORDER)
				REFITLogger.logOrder(this, instanceID + " reached " + state + " (" + proposal + ", " + certificate.message + ")");
			this.instanceState = state;
		}

		@Override
		public boolean isProposer() {
			return (isPrimary && (instanceState == REFITPBFTProtocolInstanceState.INITIALIZED) && proposal == null);
		}

		public boolean hasProposal() {
			return instanceState != REFITPBFTProtocolInstanceState.INITIALIZED || proposal != null;
		}

		@Override
		public boolean isPrepared() {
			return instanceState == REFITPBFTProtocolInstanceState.COMMITTED
					|| instanceState == REFITPBFTProtocolInstanceState.PREPARED;
		}

		@Override
		public void assignProposal(REFITOrderRequest message) {
			if (!isProposer()) {
				throw new InternalError("Trying to assign a proposal to a non proposer slot");
			}
			proposal = message;
		}

		protected void orderReplicaMulticast(REFITMessage message) {
			message.setRecipientCount(REFITConfig.ORDER_GROUP_SIZE);
			message.serializeMessage();
			replica.mac.appendMulticastMAC(message, allOrderReplicas);
			replica.replicaMulticast(message, activeOrderRecipients);
		}


		// ######################
		// # LIFE-CYCLE METHODS #
		// ######################

		@Override
		public void init(long instanceID, int viewID) {
			super.init(instanceID, viewID);

			int groupID = REFITOrderGroups.getGroupID(instanceID);
			allOrderReplicas = protocol.getAllOrderRecipients(groupID);
			activeOrderRecipients = protocol.getActiveOrderRecipients(groupID);

			certificate.init(instanceID, viewID, activeOrderRecipients);
			isPrimary = (replica.id == REFITOrderGroups.getPrimary(instanceID, viewID));
			proposal = null;

			if (!allOrderReplicas[replica.id]) {
				throw new InternalError("This replica does not belong to the agreement group!");
			}
			advanceState(REFITPBFTProtocolInstanceState.INITIALIZED);
		}

		@Override
		public REFITOrderProtocolData abort() {
			if (!allOrderReplicas[replica.id]) {
				throw new InternalError("Unexpected");
			}
			// only report prepared requests as possibly committed
			boolean potentiallyDecided = certificate.isPrepared();
			REFITOrderRequest command = potentiallyDecided ? certificate.message : REFITNoOpRequest.INSTANCE;
			return new REFITOrderProtocolData(command, certificate.getProof());
		}


		// ######################
		// # PROTOCOL EXECUTION #
		// ######################

		@Override
		public void execute() {
			if (isPrimary) executePrimary();
			else executeBackup();
		}

		private void executePrimary() {
			switch (instanceState) {
			case INITIALIZED:
				if (proposal == null) {
					// ignore unexpected activation from outside
					break;
				}
				// Send PRE_PREPARE
				REFITPBFTPrePrepare prePrepare = new REFITPBFTPrePrepare(instanceID, viewID, proposal, replica.id);
				orderReplicaMulticast(prePrepare);

				// Add PRE_PREPARE to certificate
				certificate.addPrePrepare(prePrepare, replica.mac);

				// Advance to next state
				advanceState(REFITPBFTProtocolInstanceState.PRE_PREPARED);
				//$FALL-THROUGH$
			case PRE_PREPARED:
				executePreparePhase();
				//$FALL-THROUGH$ --> Try to complete commit phase even if prepare phase has not been completed locally
			case PREPARED:
				executeCommitPhase();
				break;
			default:
				throw new IllegalStateException("nothing to do for state " + instanceState);
			}
		}

		private void executeBackup() {
			switch (instanceState) {
			case INITIALIZED:
				while (true) {
					// Fetch PRE_PREPARE
					REFITPBFTPrePrepare prePrepare = (REFITPBFTPrePrepare) fetchMessage(REFITMessageType.PBFT_PRE_PREPARE, viewID);
					if (prePrepare == null) return;

					// Check whether PRE_PREPARE is from primary
					if (prePrepare.from != REFITOrderGroups.getPrimary(instanceID, viewID)) {
						REFITLogger.logWarning(this, "bad PRE_PREPARE " + prePrepare);
						continue;
					}

					// Check whether the proposed message matches the proposal expected (if there is such a proposal)
					// TODO: It should be save to accept non-no-op proposal if the expected proposal is a no-op
					if ((proposal != null) && !proposal.equals(prePrepare.message)) {
						REFITLogger.logWarning(this, "bad PRE_PREPARE " + prePrepare + " (message mismatch: " + proposal + " expected)");
						continue;
					}

					// Add PRE_PREPARE to certificate
					boolean success = certificate.addPrePrepare(prePrepare, replica.mac);
					if (!success) {
						REFITLogger.logWarning(this, "bad PRE_PREPARE " + prePrepare);
						continue;
					}
					break;
				}

				// Create and distribute PREPARE
				REFITPBFTPrepare ownPrepare = new REFITPBFTPrepare(certificate.prePrepare, replica.id);
				orderReplicaMulticast(ownPrepare);

				// Add PREPARE to certificate
				certificate.addPrepare(ownPrepare, replica.mac);

				// Advance to next state
				advanceState(REFITPBFTProtocolInstanceState.PRE_PREPARED);
				//$FALL-THROUGH$ --> Try to also complete prepare phase
			case PRE_PREPARED:
				executePreparePhase();
				//$FALL-THROUGH$ --> Try to complete commit phase (even if prepare phase has not been completed locally)
			case PREPARED:
				executeCommitPhase();
				break;
			default:
				throw new IllegalStateException("nothing to do for state " + instanceState);
			}
		}

		private void executePreparePhase() {
			while (!certificate.isPrepared()) {
				// Fetch PREPAREs
				REFITPBFTPrepare prepare = (REFITPBFTPrepare) fetchMessage(REFITMessageType.PBFT_PREPARE, viewID);
				if (prepare == null) return;

				// Add PREPARE to certificate
				boolean success = certificate.addPrepare(prepare, replica.mac);
				if (!success) REFITLogger.logWarning(this, "bad PREPARE " + prepare);
			}

			// Create and distribute COMMIT
			REFITPBFTCommit ownCommit = new REFITPBFTCommit(certificate.prePrepare, replica.id);
			orderReplicaMulticast(ownCommit);

			// Add COMMIT to certificate
			certificate.addCommit(ownCommit, replica.mac);

			// Advance to next state
			advanceState(REFITPBFTProtocolInstanceState.PREPARED);
		}

		private void executeCommitPhase() {
			while (!certificate.isCommitted()) {
				// Fetch COMMITs
				REFITPBFTCommit commit = (REFITPBFTCommit) fetchMessage(REFITMessageType.PBFT_COMMIT, viewID);
				if (commit == null) return;

				// Add COMMIT to certificate
				REFITPBFTCertificate.AddResult addResult = certificate.addCommit(commit, replica.mac);
				if (addResult.equals(REFITPBFTCertificate.AddResult.INVALID)) {
					REFITLogger.logWarning(this, "bad COMMIT " + commit);
				}
			}

			// Advance to next state
			advanceState(REFITPBFTProtocolInstanceState.COMMITTED);

			// Instance complete
			completeWithResult(certificate.message, null);
		}

	}

}
