package refit.agreement.pbft;

import java.nio.ByteBuffer;
import java.security.MessageDigest;
import java.util.Arrays;
import java.util.Collection;
import java.util.LinkedList;

import refit.message.REFITMessage;
import refit.message.REFITMessageType;
import refit.message.REFITNoOpRequest;
import refit.message.REFITOrderRequest;
import refit.replica.order.REFITOrderProtocolMessage;
import refit.util.REFITPayload;


public class REFITPBFTMessages {

	// #####################
	// # ABSTRACT MESSAGES #
	// #####################

	public static abstract class REFITPBFTMessage extends REFITOrderProtocolMessage {

		public static final int PBFT_HEADER_SIZE = Integer.BYTES;

		protected ByteBuffer command;


		protected REFITPBFTMessage(REFITMessageType pbftType, long instanceID, int viewID, ByteBuffer command, short from) {
			super(pbftType, instanceID, viewID, from);
			this.command = command;
		}

		protected REFITPBFTMessage(REFITMessageType expectedType, ByteBuffer buffer) {
			super(expectedType, buffer);
			int cmdSize = buffer.getInt();
			this.command = buffer.slice();
			this.command.limit(cmdSize);
		}

		@Override
		public AuthenticationType getAuthenticationType() {
			return AuthenticationType.MAC_AUTHENTICATOR;
		}

		@Override
		protected void serialize(ByteBuffer buffer) {
			super.serialize(buffer);
			serializeCommand(buffer);
		}

		protected void serializeCommand(final ByteBuffer buffer) {
			int cmdSize = command.remaining();
			buffer.putInt(cmdSize);
			// replace command ByteBuffer after serialization to store the data only once
			ByteBuffer newCommand = buffer.slice();
			buffer.put(command);
			newCommand.limit(cmdSize);
			command = newCommand;
		}

		public ByteBuffer getCommand() {
			return command.slice();
		}

		@Override
		protected int getHeaderSize() {
			return (super.getHeaderSize() + PBFT_HEADER_SIZE);
		}

		@Override
		protected int calculatePayloadSize() {
			return (super.calculatePayloadSize() + PBFT_HEADER_SIZE + calculateCommandSize());
		}

		protected int calculateCommandSize() {
			return command.remaining();
		}

		@Override
		public String toString() {
			return String.format("{%s|%d|%d|%d}", type, from, uid.seqNr, viewID);
		}

	}

	// #####################
	// # SPECIFIC MESSAGES #
	// #####################

	public static class REFITPBFTPrePrepare extends REFITPBFTMessage {

		public final transient REFITOrderRequest message;


		public REFITPBFTPrePrepare(long instanceID, int viewID, REFITOrderRequest message, short from) {
			super(REFITMessageType.PBFT_PRE_PREPARE, instanceID, viewID, null, from);
			this.message = message;
		}

		public REFITPBFTPrePrepare(ByteBuffer buffer) {
			super(REFITMessageType.PBFT_PRE_PREPARE, buffer);
			// create buffer slice to keep the invariant that the ByteBuffer passed
			// to a message constructor must be at position zero
			ByteBuffer cmd = getCommand();
			if (cmd.hasRemaining()) {
				this.message = (REFITOrderRequest) REFITMessage.createMessage(cmd);
			} else {
				this.message = REFITNoOpRequest.INSTANCE;
			}
		}


		@Override
		public String toString() {
			return String.format("{%s|%d|%d|%d|%s}", type, from, uid.seqNr, viewID, message);
		}

		@Override
		protected void serializeCommand(final ByteBuffer buffer) {
			if (message.type == REFITMessageType.REQUEST_NO_OP) {
				command = ByteBuffer.allocate(0);
				return;
			}

			int cmdSize = message.calculateMessageSize();
			buffer.putInt(cmdSize);
			// replace command ByteBuffer after serialization to store the data only once
			ByteBuffer newCommand = buffer.slice();
			message.serializeInto(buffer);
			newCommand.limit(cmdSize);
			command = newCommand;
		}

		@Override
		protected int calculateCommandSize() {
			if (message.type == REFITMessageType.REQUEST_NO_OP) {
				return 0;
			} else {
				return message.calculateMessageSize();
			}
		}

		@Override
		protected byte[] calculateHash() {
			// Prepare message header
			ByteBuffer messageHeader = getInternalBuffer();
			messageHeader.limit(getHeaderSize());

			// Prepare message (must be done in advance in order to prevent the hash from
			// being computed during the calculation of the REFITPBFTPrePrepare hash)
			byte[] messageHash = message.getHash();

			// Calculate hash
			MessageDigest digest = REFITPayload.getDigest();
			digest.update(messageHeader);
			digest.update(messageHash);
			return digest.digest();
		}
	}

	public static class REFITPBFTPrepare extends REFITPBFTMessage {

		public REFITPBFTPrepare(REFITPBFTPrePrepare prePrepare, short from) {
			super(REFITMessageType.PBFT_PREPARE, prePrepare.uid.seqNr, prePrepare.viewID,
					ByteBuffer.wrap(prePrepare.message.getHash()), from);
		}

		public REFITPBFTPrepare(ByteBuffer buffer) {
			super(REFITMessageType.PBFT_PREPARE, buffer);
		}

	}

	public static class REFITPBFTCommit extends REFITPBFTMessage {

		public REFITPBFTCommit(REFITPBFTPrePrepare prePrepare, short from) {
			super(REFITMessageType.PBFT_COMMIT, prePrepare.uid.seqNr, prePrepare.viewID,
					ByteBuffer.wrap(prePrepare.message.getHash()), from);
		}

		public REFITPBFTCommit(ByteBuffer buffer) {
			super(REFITMessageType.PBFT_COMMIT, buffer);
		}

	}

	public static class REFITPBFTViewChange extends REFITPBFTMessage {

		public final transient REFITOrderRequest message;
		public final transient REFITPBFTMessage[] proof;


		public REFITPBFTViewChange(long instanceID, int viewID, REFITOrderRequest message, REFITPBFTMessage[] proof, short from) {
			super(REFITMessageType.PBFT_VIEW_CHANGE, instanceID, viewID, createViewChangePayload(message, proof), from);
			this.message = message;
			this.proof = proof;
		}

		public REFITPBFTViewChange(ByteBuffer buffer) {
			super(REFITMessageType.PBFT_VIEW_CHANGE, buffer);
			command.mark();
			if (!command.hasRemaining()) {
				this.message = REFITNoOpRequest.INSTANCE;
				this.proof = new REFITPBFTMessage[0];
			} else {
				this.message = (REFITOrderRequest) REFITMessage.createMessage(command.slice());
				command.position(command.position() + message.getMessageSize());
				short viewCount = command.getShort();
				this.proof = new REFITPBFTMessage[viewCount];
				for (int i = 0; i < proof.length; i++) {
					proof[i] = (REFITPBFTMessage) REFITMessage.createMessage(command.slice());
					command.position(command.position() + proof[i].getMessageSize());
				}
			}
			command.reset();
		}

		@Override
		public AuthenticationType getAuthenticationType() {
			return AuthenticationType.SIGNATURE;
		}

		@Override
		public String toString() {
			return String.format("{%s|%d|%d|%d|%s|%s}", type, from, uid.seqNr, viewID, message.uid, Arrays.toString(proof));
		}

		private static ByteBuffer createViewChangePayload(REFITMessage message, REFITPBFTMessage[] proof) {
			// Short cut for no-ops
			if (message.type == REFITMessageType.REQUEST_NO_OP) {
				return ByteBuffer.allocate(0);
			}

			// Determine size
			int size = message.calculateMessageSize() + Short.BYTES;
			for (REFITPBFTMessage msg : proof) size += msg.calculateMessageSize();

			// Pack messages
			ByteBuffer buffer = ByteBuffer.allocate(size);
			buffer.put(message.getBuffer());
			buffer.putShort((short) proof.length);
			for (final REFITPBFTMessage msg : proof) {
				buffer.put(msg.getBuffer());
			}
			buffer.flip();
			return buffer;
		}

		@Override
		protected byte[] calculateHash() {
			// Prepare message header
			ByteBuffer messageHeader = getInternalBuffer();
			messageHeader.limit(getHeaderSize());

			// Prepare message (must be done in advance in order to prevent the hash from
			// being computed during the calculation of the REFITPBFTPrePrepare hash)
			byte[] messageHash = message.getHash();

			if (message.type != REFITMessageType.REQUEST_NO_OP) {
				// Prepare proof hashes
				for (REFITPBFTMessage proofMessage : proof) proofMessage.getHash();
			}

			// Calculate hash
			MessageDigest digest = REFITPayload.getDigest();
			digest.update(messageHeader);
			digest.update(messageHash);
			if (message.type != REFITMessageType.REQUEST_NO_OP) {
				for (REFITPBFTMessage proofMessage : proof) digest.update(proofMessage.getHash());
			}
			return digest.digest();
		}

	}

	public static class REFITPBFTNewView extends REFITPBFTMessage {

		public final transient Collection<REFITPBFTViewChange> proof;


		public REFITPBFTNewView(long instanceID, int viewID, Collection<REFITPBFTViewChange> proof, short from) {
			super(REFITMessageType.PBFT_NEW_VIEW, instanceID, viewID, serializeViewChanges(proof), from);
			this.proof = proof;
		}

		public REFITPBFTNewView(ByteBuffer buffer) {
			super(REFITMessageType.PBFT_NEW_VIEW, buffer);
			this.proof = new LinkedList<>();
			short nrOfViewChanges = command.getShort();
			command.mark();
			for (int i = 0; i < nrOfViewChanges; i++) {
				REFITPBFTViewChange viewChange = new REFITPBFTViewChange(command.slice());
				proof.add(viewChange);
				command.position(command.position() + viewChange.getMessageSize());
			}
			command.reset();
		}

		@Override
		public AuthenticationType getAuthenticationType() {
			return AuthenticationType.SIGNATURE;
		}

		private static ByteBuffer serializeViewChanges(Collection<REFITPBFTViewChange> viewChanges) {
			// Determine size
			int size = Short.BYTES;
			for (REFITPBFTViewChange viewChange : viewChanges) size += viewChange.getMessageSize();

			// Pack VIEW_CHANGES
			ByteBuffer buffer = ByteBuffer.allocate(size);
			buffer.putShort((short) viewChanges.size());
			for (final REFITPBFTViewChange viewChange : viewChanges) {
				buffer.put(viewChange.getBuffer());
			}
			buffer.flip();
			return buffer;
		}

		@Override
		protected byte[] calculateHash() {
			// Prepare message header
			ByteBuffer messageHeader = getInternalBuffer();
			messageHeader.limit(getHeaderSize());

			// Prepare VIEW_CHANGE hashes
			for (REFITPBFTMessage viewChange : proof) viewChange.getHash();

			// Calculate hash
			MessageDigest digest = REFITPayload.getDigest();
			digest.update(messageHeader);
			for (REFITPBFTViewChange viewChange : proof) digest.update(viewChange.getHash());
			return digest.digest();
		}

	}

}
