package refit.agreement.pbft;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import refit.agreement.pbft.REFITPBFT.REFITPBFTProtocolInstance;
import refit.agreement.pbft.REFITPBFTMessages.REFITPBFTMessage;
import refit.agreement.pbft.REFITPBFTMessages.REFITPBFTNewView;
import refit.agreement.pbft.REFITPBFTMessages.REFITPBFTViewChange;
import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITMessageType;
import refit.message.REFITOrderRequest;
import refit.replica.order.*;


public class REFITPBFTTransition {

	// ############################
	// # PROTOCOL INSTANCE STATES #
	// ############################

	public enum REFITPBFTTransitionProtocolState {
		WAITING,
		SEND_VIEW_CHANGE,
		VIEW_CHANGE_SENT
	}


	// #####################
	// # PROTOCOL INSTANCE #
	// #####################

	public static class REFITPBFTTransitionProtocolInstance extends REFITOrderProtocolTransitionInstance {

		// calculates ceil((n+f+1)/2)
		private static final int QUORUM_THRESHOLD = (REFITConfig.ORDER_GROUP_SIZE + REFITConfig.FAULTS_TO_TOLERATE + 1 + 1) / 2;
		private static final int VERIFICATION_THRESHOLD = REFITConfig.FAULTS_TO_TOLERATE + 1;

		private final Map<Integer, REFITPBFTTransitionCertificate> certificates;
		private final int[] highestReceivedViewChange;
		private boolean[] orderRecipients;
		private final REFITOrderProtocol protocol;
		private REFITPBFTViewChange ownViewChange;
		private REFITPBFTTransitionProtocolState instanceState;
		private boolean viewChangeCouldComplete = false;


		public REFITPBFTTransitionProtocolInstance(REFITOrderStageSlot slot, REFITOrderProtocol protocol) {
			super(slot);
			this.protocol = protocol;
			this.certificates = new HashMap<>();
			this.highestReceivedViewChange = new int[REFITConfig.TOTAL_NR_OF_REPLICAS];
			Arrays.fill(highestReceivedViewChange, -1);
			this.instanceState = REFITPBFTTransitionProtocolState.WAITING;
		}


		// ##################
		// # HELPER METHODS #
		// ##################

		@Override
		public String toString() {
			return "VWCH[" + instanceID + "-" + viewID + "]";
		}

		private void advanceState(REFITPBFTTransitionProtocolState state) {
			this.instanceState = state;
		}


		// ###############################
		// # TRANSITION-SPECIFIC METHODS #
		// ###############################

		protected REFITPBFTViewChange createOwnViewChange(REFITOrderRequest message, REFITPBFTMessage[] proof) {
			return new REFITPBFTViewChange(instanceID, viewID, message, proof, replica.id);
		}

		protected REFITPBFTTransitionCertificate createCertificate() {
			return new REFITPBFTTransitionCertificate(VERIFICATION_THRESHOLD, QUORUM_THRESHOLD, REFITConfig.ORDER_GROUP_SIZE, orderRecipients);
		}

		protected boolean checkViewChange(REFITPBFTViewChange viewChange) {
			return (viewID <= viewChange.viewID)
					// don't allow a server to repeat old view changes
					&& (highestReceivedViewChange[viewChange.from] < viewChange.viewID);
		}


		// ######################
		// # LIFE-CYCLE METHODS #
		// ######################


		@Override
		public void init(long instanceID, int viewID) {
			super.init(instanceID, viewID);
			int groupID = REFITOrderGroups.getGroupID(instanceID);
			orderRecipients = protocol.getAllOrderRecipients(groupID);
			Arrays.fill(highestReceivedViewChange, -1);
		}

		@Override
		public void updateProtocolData(REFITOrderProtocolData data) {
			ownViewChange = createOwnViewChange(data.message, (REFITPBFTMessage[]) data.proof);
			advanceState(REFITPBFTTransitionProtocolState.SEND_VIEW_CHANGE);
		}

		@Override
		public void updateView(int newViewID) {
			if (newViewID <= viewID) {
				throw new InternalError("View must increase! at " + viewID + " got " + newViewID);
			}
			viewID = newViewID;
			switchToViewID(newViewID);
			ownViewChange = createOwnViewChange(ownViewChange.message, ownViewChange.proof);
			advanceState(REFITPBFTTransitionProtocolState.SEND_VIEW_CHANGE);
		}

		private void switchToViewID(int viewID) {
			// Garbage-collect old certificates
			certificates.keySet().removeIf(integer -> integer < viewID);

			viewChangeCouldComplete = couldViewChangeComplete();
			if (viewChangeCouldComplete) {
				learnCompletableViewChange();
			}
		}

		private boolean couldViewChangeComplete() {
			int currentViewIDCount = 0;
			for (Integer receivedViewID : highestReceivedViewChange) {
				if (receivedViewID == viewID) {
					currentViewIDCount++;
				}
			}
			if (instanceState.equals(REFITPBFTTransitionProtocolState.VIEW_CHANGE_SENT)) {
				currentViewIDCount++;
			}
			return currentViewIDCount >= QUORUM_THRESHOLD;
		}


		@Override
		public void execute() {
			switch (instanceState) {
			case SEND_VIEW_CHANGE:
				sendViewChange();
				// exit if this completed the view change
				if (handleViewChangeIfPrimary(ownViewChange)) return;

				// Advance to next state
				advanceState(REFITPBFTTransitionProtocolState.VIEW_CHANGE_SENT);
				//$FALL-THROUGH$
			case WAITING: // watch for view changes until receiving f+1 view-changes
			case VIEW_CHANGE_SENT:
				// Process NEW_VIEWs
				while (true) {
					// Get NEW_VIEW
					REFITPBFTNewView newView = (REFITPBFTNewView) fetchMessage(REFITMessageType.PBFT_NEW_VIEW);
					if (newView == null) break;
					if (REFITLogger.LOG_ORDER) REFITLogger.logOrder(this, "received NEW_VIEW " + newView);

					// Check NEW_VIEW
					// prevent reverting to an old view, this is necessary to guarantee progress
					// only listen to newer views while waiting
					if (newView.viewID < viewID || (newView.viewID == viewID && instanceState == REFITPBFTTransitionProtocolState.WAITING))
						continue;
					if (newView.from != REFITOrderGroups.getPrimary(instanceID, newView.viewID)) continue;
					if (!replica.mac.verifySignature(newView, orderRecipients)) continue;

					// Verify and get decision
					for (REFITPBFTViewChange viewChange : newView.proof) {
						// prevent creating a NEW_VIEW from old VIEW_CHANGEs, also prevents viewID exhaustion
						// the VIEW_CHANGEs must match the view of NEW_VIEW to ensure that a quorum of servers
						// is available after the view change
						if (viewChange.viewID != newView.viewID) break;
						Collection<REFITPBFTViewChange> stabilityProof = tryToCompleteViewChange(viewChange);
						if (stabilityProof != null) return;
					}
					REFITLogger.logWarning(this, "bad NEW_VIEW " + newView);
				}

				// Process VIEW_CHANGEs
				while (true) {
					// Get VIEW_CHANGE
					REFITPBFTViewChange viewChange = (REFITPBFTViewChange) fetchMessage(REFITMessageType.PBFT_VIEW_CHANGE);
					if (viewChange == null) break;
					if (REFITLogger.LOG_ORDER)
						REFITLogger.logOrder(this, "received VIEW_CHANGE " + viewChange + " (" + (replica.id == REFITOrderGroups.getPrimary(instanceID, viewChange.viewID)) + ")");

					// Check VIEW_CHANGE
					if (!checkViewChange(viewChange)) continue;
					if (viewChange.viewID == viewID && instanceState == REFITPBFTTransitionProtocolState.WAITING)
						continue;
					if (!replica.mac.verifySignature(viewChange, orderRecipients)) continue;
					highestReceivedViewChange[viewChange.from] = viewChange.viewID;

					// start new view change after receiving f+1 view change messages with a higher viewID that the current one
					// use the lowest of these viewIDs
					int higherViewIDCount = 0;
					int newViewID = Integer.MAX_VALUE;
					for (Integer receivedViewID : highestReceivedViewChange) {
						if (receivedViewID > viewID) {
							higherViewIDCount++;
							newViewID = Math.min(newViewID, receivedViewID);
						}
					}

					if (higherViewIDCount >= VERIFICATION_THRESHOLD) {
						synchronizeNewView(newViewID);
						return;
					} else if (!viewChangeCouldComplete && couldViewChangeComplete()) {
						viewChangeCouldComplete = true;
						learnCompletableViewChange();
					}

					// exit if this completed the new_view message
					if (handleViewChangeIfPrimary(viewChange)) break;
				}
				break;
			default:
				throw new IllegalStateException("nothing to do for state " + instanceState);
			}
		}

		private void sendViewChange() {
			ownViewChange.serializeMessage();
			replica.mac.appendSignature(ownViewChange);

			// Broadcast own VIEW_CHANGE, the primary will process it, the others get informed for liveness
			if (REFITLogger.LOG_ORDER) REFITLogger.logOrder(this, "broadcast " + ownViewChange);
			replica.replicaMulticast(ownViewChange, orderRecipients);
		}

		private boolean handleViewChangeIfPrimary(REFITPBFTViewChange viewChange) {
			boolean isViewChangePrimary = (replica.id == REFITOrderGroups.getPrimary(instanceID, viewChange.viewID));
			if (isViewChangePrimary) {
				Collection<REFITPBFTViewChange> proof = tryToCompleteViewChange(viewChange);
				if (proof != null) {
					// broadcast the new view to restart the protocol on the other servers
					broadcastNewView(viewChange.viewID, proof);
					return true;
				}
			}
			return false;
		}

		private Collection<REFITPBFTViewChange> tryToCompleteViewChange(REFITPBFTViewChange viewChange) {
			// Get certificate
			REFITPBFTTransitionCertificate certificate = certificates.get(viewChange.viewID);
			if (certificate == null) {
				boolean isPrimary = (replica.id == REFITOrderGroups.getPrimary(instanceID, viewChange.viewID));
				certificate = createCertificate();
				certificates.put(viewChange.viewID, certificate);
			}

			// Add VIEW_CHANGE to certificate
			REFITPBFTTransitionCertificate.AddResult addResult = certificate.addViewChange(viewChange, replica.mac);
			if (!addResult.equals(REFITPBFTTransitionCertificate.AddResult.ADDED)) {
				if (addResult.equals(REFITPBFTTransitionCertificate.AddResult.INVALID)) {
					REFITLogger.logWarning(this, "bad VIEW_CHANGE " + viewChange);
				}
				return null;
			}

			// Check whether the certificate has become stable
			if (!certificate.isStable()) return null;
			if (REFITLogger.LOG_ORDER)
				REFITLogger.logOrder(this, "certificate for view " + viewChange.viewID + " became stable");

			// Switch to new PBFT instance
			REFITPBFTProtocolInstance pbftInstance = new REFITPBFTProtocolInstance(slot, instanceID, viewChange.viewID,
					certificate.getDecision(), protocol);
			if (REFITLogger.LOG_ORDER)
				REFITLogger.logOrder(this, "switch to PBFT instance for " + certificate.getDecision());
			advanceState(REFITPBFTTransitionProtocolState.WAITING);
			learnCompleteViewChange(viewChange.viewID, protocol, pbftInstance);

			return certificate.getProof();
		}

		private void learnCompleteViewChange(int viewID, REFITOrderProtocol protocol, REFITPBFTProtocolInstance pbftInstance) {
			synchronizeNewView(viewID);
			if (!viewChangeCouldComplete) {
				learnCompletableViewChange();
				viewChangeCouldComplete = true;
			}
			learnViewChange(protocol, pbftInstance);
		}

		private void broadcastNewView(int viewID, Collection<REFITPBFTViewChange> stabilityProof) {
			// Create and distribute NEW_VIEW
			REFITPBFTNewView newView = new REFITPBFTNewView(instanceID, viewID, stabilityProof, replica.id);
			newView.serializeMessage();
			replica.mac.appendSignature(newView);
			replica.replicaMulticast(newView, orderRecipients);
		}
	}

}
