package refit.agreement.steward;

import refit.agreement.steward.REFITStewardMessages.REFITStewardCommit;
import refit.agreement.steward.REFITStewardMessages.REFITStewardPrePrepare;
import refit.agreement.steward.REFITStewardMessages.REFITStewardPrepare;
import refit.agreement.steward.REFITStewardTransition.REFITStewardTransitionProtocolInstance;
import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITMessage;
import refit.message.REFITMessageType;
import refit.message.REFITNoOpRequest;
import refit.message.REFITOrderRequest;
import refit.replica.order.*;
import refit.util.REFITAssert;
import threshsig.SigShare;


public class REFITSteward {

	// ############
	// # PROTOCOL #
	// ############

	public static class REFITStewardProtocol extends REFITOrderProtocol {

		public static final REFITStewardProtocol instance = new REFITStewardProtocol();

		@Override
		public REFITOrderProtocolInstance createInstance(REFITOrderStageSlot slot) {
			return new REFITStewardProtocolInstance(slot, this);
		}

		@Override
		public REFITOrderProtocolTransitionInstance createTransitionInstance(REFITOrderStageSlot slot) {
			return new REFITStewardTransitionProtocolInstance(slot);
		}

	}


	// ############################
	// # PROTOCOL INSTANCE STATES #
	// ############################

	public enum REFITStewardProtocolInstanceState {
		INITIALIZED,
		PRE_PREPARED,
		PREPARED,
		COMMITTED
	}


	// #####################
	// # PROTOCOL INSTANCE #
	// #####################

	public static class REFITStewardProtocolInstance extends REFITOrderProtocolInstance {

		protected final REFITOrderProtocol protocol;
		protected final REFITStewardCertificate certificate;
		protected final boolean[] orderRecipients;
		private REFITStewardProtocolInstanceState instanceState;
		private boolean isPrimary;
		private REFITOrderRequest proposal;

		public REFITStewardProtocolInstance(REFITOrderStageSlot slot, REFITOrderProtocol protocol) {
			super(slot);
			this.protocol = protocol;
			orderRecipients = protocol.getAllOrderRecipients(0);
			this.certificate = new REFITStewardCertificate(replica.id, REFITConfig.ORDER_GROUP_SIZE);
			REFITAssert.assertTrue(REFITConfig.AUTHENTICATE_MESSAGES && REFITConfig.SIGNED_REQUESTS,
					"steward requires authenticated messages");
		}


		// ##################
		// # HELPER METHODS #
		// ##################

		@Override
		public String toString() {
			return "STEW[" + instanceID + "]";
		}

		private void advanceState(REFITStewardProtocolInstanceState state) {
			if (REFITLogger.LOG_ORDER)
				REFITLogger.logOrder(this, instanceID + " reached " + state + " (" + proposal + ", " + certificate.message + ")");
			this.instanceState = state;
		}

		@Override
		public boolean isProposer() {
			return (isPrimary && (instanceState == REFITStewardProtocolInstanceState.INITIALIZED) && proposal == null);
		}

		public boolean hasProposal() {
			return instanceState != REFITStewardProtocolInstanceState.INITIALIZED || proposal != null;
		}

		@Override
		public boolean isPrepared() {
			return instanceState == REFITStewardProtocolInstanceState.COMMITTED
					|| instanceState == REFITStewardProtocolInstanceState.PREPARED;
		}

		@Override
		public void assignProposal(REFITOrderRequest message) {
			if (!isProposer()) {
				throw new InternalError("Trying to assign a proposal to a non proposer slot");
			}
			proposal = message;
		}

		protected void orderReplicaMulticast(REFITMessage message) {
			message.serializeMessage();
			replica.mac.appendSignature(message);
			replica.replicaMulticast(message, orderRecipients);
		}


		// ######################
		// # LIFE-CYCLE METHODS #
		// ######################

		@Override
		public void init(long instanceID, int viewID) {
			super.init(instanceID, viewID);
			certificate.init(instanceID, viewID, orderRecipients);
			isPrimary = (replica.id == REFITOrderGroups.getPrimary(instanceID, viewID));
			proposal = null;

			if (!orderRecipients[replica.id]) {
				throw new InternalError("This replica does not belong to the agreement group!");
			}
			advanceState(REFITStewardProtocolInstanceState.INITIALIZED);
		}

		@Override
		public REFITOrderProtocolData abort() {
			// only report prepared requests as possibly committed
			boolean potentiallyDecided = certificate.isPrepared();
			REFITOrderRequest command = potentiallyDecided ? certificate.message : REFITNoOpRequest.INSTANCE;
			return new REFITOrderProtocolData(command, certificate.getProof());
		}


		// ######################
		// # PROTOCOL EXECUTION #
		// ######################

		@Override
		public void execute() {
			if (isPrimary) executePrimary();
			else executeBackup();
		}

		private void executePrimary() {
			switch (instanceState) {
			case INITIALIZED:
				if (proposal == null) {
					// ignore unexpected activation from outside
					break;
				}
				// Send PRE_PREPARE
				REFITStewardPrePrepare prePrepare = new REFITStewardPrePrepare(instanceID, viewID, proposal, replica.id);
				orderReplicaMulticast(prePrepare);

				// Add PRE_PREPARE to certificate
				certificate.addPrePrepare(prePrepare, replica.mac);

				// Advance to next state
				advanceState(REFITStewardProtocolInstanceState.PRE_PREPARED);
				//$FALL-THROUGH$
			case PRE_PREPARED:
				executePreparePhase();
				//$FALL-THROUGH$ --> Try to complete commit phase even if prepare phase has not been completed locally
			case PREPARED:
				executeCommitPhase();
				break;
			default:
				throw new IllegalStateException("nothing to do for state " + instanceState);
			}
		}

		private void executeBackup() {
			switch (instanceState) {
			case INITIALIZED:
				while (true) {
					// Fetch PRE_PREPARE
					REFITStewardPrePrepare prePrepare = (REFITStewardPrePrepare) fetchMessage(REFITMessageType.STEWARD_PRE_PREPARE, viewID);
					if (prePrepare == null) return;

					// Check whether PRE_PREPARE is from primary
					if (prePrepare.from != REFITOrderGroups.getPrimary(instanceID, viewID)) {
						REFITLogger.logWarning(this, "bad PRE_PREPARE " + prePrepare);
						continue;
					}

					// Check whether the proposed message matches the proposal expected (if there is such a proposal)
					// TODO: It should be save to accept non-no-op proposal if the expected proposal is a no-op
					if ((proposal != null) && !proposal.equals(prePrepare.message)) {
						REFITLogger.logWarning(this, "bad PRE_PREPARE " + prePrepare + " (message mismatch: " + proposal + " expected)");
						continue;
					}

					// Add PRE_PREPARE to certificate
					boolean success = certificate.addPrePrepare(prePrepare, replica.mac);
					if (!success) {
						REFITLogger.logWarning(this, "bad PRE_PREPARE " + prePrepare);
						continue;
					}
					break;
				}

				// Create and distribute PREPARE
				REFITStewardPrepare ownPrepare = new REFITStewardPrepare(certificate.prePrepare, replica.id);
				orderReplicaMulticast(ownPrepare);

				// Add PREPARE to certificate
				certificate.addPrepare(ownPrepare, replica.mac);

				// Advance to next state
				advanceState(REFITStewardProtocolInstanceState.PRE_PREPARED);
				//$FALL-THROUGH$ --> Try to also complete prepare phase
			case PRE_PREPARED:
				executePreparePhase();
				//$FALL-THROUGH$ --> Try to complete commit phase (even if prepare phase has not been completed locally)
			case PREPARED:
				executeCommitPhase();
				break;
			default:
				throw new IllegalStateException("nothing to do for state " + instanceState);
			}
		}

		private void executePreparePhase() {
			while (!certificate.isPrepared()) {
				// Fetch PREPAREs
				REFITStewardPrepare prepare = (REFITStewardPrepare) fetchMessage(REFITMessageType.STEWARD_PREPARE, viewID);
				if (prepare == null) return;

				// Add PREPARE to certificate
				boolean success = certificate.addPrepare(prepare, replica.mac);
				if (!success) REFITLogger.logWarning(this, "bad PREPARE " + prepare);
			}

			SigShare share = replica.mac.calcSigShare(certificate.prePrepare.calcShortHash());

			// Create and distribute COMMIT
			REFITStewardCommit ownCommit = new REFITStewardCommit(certificate.prePrepare, replica.id, share);
			orderReplicaMulticast(ownCommit);

			// Add COMMIT to certificate
			certificate.addCommit(ownCommit, replica.mac);

			// Advance to next state
			advanceState(REFITStewardProtocolInstanceState.PREPARED);
		}

		private void executeCommitPhase() {
			while (!certificate.isCommitted()) {
				// Fetch COMMITs
				REFITStewardCommit commit = (REFITStewardCommit) fetchMessage(REFITMessageType.STEWARD_COMMIT, viewID);
				if (commit == null) return;

				// Add COMMIT to certificate
				REFITStewardCertificate.AddResult addResult = certificate.addCommit(commit, replica.mac);
				if (addResult.equals(REFITStewardCertificate.AddResult.INVALID)) {
					REFITLogger.logWarning(this, "bad COMMIT " + commit);
				}
			}

			// Advance to next state
			advanceState(REFITStewardProtocolInstanceState.COMMITTED);

			// Instance complete
			completeWithResult(certificate.message, certificate.getCombinedSig());
		}

	}

}
