package refit.agreement.steward;

import java.math.BigInteger;
import java.nio.ByteBuffer;
import java.util.List;

import refit.agreement.steward.REFITStewardMessages.REFITStewardCommit;
import refit.agreement.steward.REFITStewardMessages.REFITStewardMessage;
import refit.agreement.steward.REFITStewardMessages.REFITStewardPrePrepare;
import refit.agreement.steward.REFITStewardMessages.REFITStewardPrepare;
import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITMessageAuthentication;
import refit.message.REFITOrderRequest;
import refit.replica.stewardorder.REFITStewardGroups;
import refit.util.REFITBallotBox;
import threshsig.SigShare;
import threshsig.ThresholdSigException;

public class REFITStewardCertificate {

	public enum AddResult {
		INVALID,
		DUPLICATE,
		ADDED
	}

	private final int groupId;
	private int PREPARED_THRESHOLD;
	private int COMMITTED_THRESHOLD;
	private boolean[] legitimateSenders;

	private long instanceID;
	private int viewID;

	public REFITStewardPrePrepare prePrepare;
	public REFITOrderRequest message;
	private REFITBallotBox<Short, ByteBuffer, REFITStewardMessage> prepares;
	private REFITBallotBox<Short, ByteBuffer, REFITStewardCommit> commits;

	private BigInteger combinedSig;


	public REFITStewardCertificate(int replicaId, int groupSize) {
		groupId = REFITStewardGroups.stewardOrderGroupID(replicaId);
		// calculates ceil((n+f+1)/2)
		PREPARED_THRESHOLD = (groupSize + REFITConfig.FAULTS_TO_TOLERATE + 1 + 1) / 2;
		COMMITTED_THRESHOLD = (groupSize + REFITConfig.FAULTS_TO_TOLERATE + 1 + 1) / 2;
		if (groupSize < PREPARED_THRESHOLD || groupSize < COMMITTED_THRESHOLD) {
			throw new AssertionError(
					String.format("Group size %d is too small for the expected number of faults %d",
							groupSize, REFITConfig.FAULTS_TO_TOLERATE));
		}
		this.prepares = new REFITBallotBox<>(PREPARED_THRESHOLD);
		this.commits = new REFITBallotBox<>(COMMITTED_THRESHOLD);
	}


	@Override
	public String toString() {
		return "{" + viewID + " " + instanceID + ": " + isPrepared() + " " + isCommitted() + "}";
	}

	public void init(long instanceID, int viewID, boolean[] legitimateSenders) {
		this.instanceID = instanceID;
		this.viewID = viewID;
		this.prePrepare = null;
		this.message = null;
		this.prepared = false;
		this.committed = false;
		this.legitimateSenders = legitimateSenders;

		prepares.clear();
		commits.clear();
	}


	// ###################
	// # ADDING MESSAGES #
	// ###################

	public boolean addPrePrepare(REFITStewardPrePrepare prePrepare, REFITMessageAuthentication mac) {
		if (prePrepare == null) return false;
		if (this.prePrepare != null) return false;
		if (message != null) return false;
		if (!mac.verifySignature(prePrepare, legitimateSenders)) return false;
		if (!mac.verifySignatureRequest(prePrepare.message)) return false;
		this.prePrepare = prePrepare;
		message = prePrepare.message;
		prepares.add(prePrepare.from, ByteBuffer.wrap(prePrepare.calcShortHash()), prePrepare);

		for (int i = 0; i < PARTICIPATION.length; i++) PARTICIPATION[i]--;
		PARTICIPATION[prePrepare.from] = PARTICIPATION_INTERVAL;

		checkPrepared();
		return true;
	}

	public boolean addPrepare(REFITStewardPrepare prepare, REFITMessageAuthentication mac) {
		if (prepare == null) return false;
		if (prepares.hasVoted(prepare.from)) return false;
		if (!mac.verifySignature(prepare, legitimateSenders)) return false;
		prepares.add(prepare.from, prepare.getCommand(), prepare);

		for (int i = 0; i < PARTICIPATION.length; i++) PARTICIPATION[i]--;
		PARTICIPATION[prepare.from] = PARTICIPATION_INTERVAL;

		checkPrepared();
		return true;
	}

	public AddResult addCommit(REFITStewardCommit commit, REFITMessageAuthentication mac) {
		if (commit == null) return AddResult.INVALID;
		if (commits.hasVoted(commit.from)) return AddResult.DUPLICATE;
		if (!mac.verifySignature(commit, legitimateSenders)) return AddResult.INVALID;
		commits.add(commit.from, ByteBuffer.wrap(commit.prepareHash), commit);

		for (int i = 0; i < PARTICIPATION.length; i++) PARTICIPATION[i]--;
		PARTICIPATION[commit.from] = PARTICIPATION_INTERVAL;

		checkCommitted(mac);
		return AddResult.ADDED;
	}


	// ######################
	// # CERTIFICATE STATES #
	// ######################

	private boolean prepared;
	private boolean committed;


	private static final int PARTICIPATION_INTERVAL = 2 * REFITConfig.CHECKPOINT_INTERVAL;
	private static final int[] PARTICIPATION = new int[REFITConfig.TOTAL_NR_OF_REPLICAS];

	static {
		for (int i = 0; i < PARTICIPATION.length; i++) PARTICIPATION[i] = PARTICIPATION_INTERVAL;
	}


	public boolean isPrepared() {
		return prepared;
	}

	public boolean preparedVotesDiffer() {
		return prepares.votesDiffer();
	}

	public boolean isCommitted() {
		return committed;
	}

	private void checkPrepared() {
		if (prepared) return;
		if (prePrepare == null) return;
		ByteBuffer decision = prepares.getDecision();
		if (decision == null) return;
		if (!prepares.getDecidingBallots().contains(prePrepare)) return;
		prepared = true;
	}

	private void checkCommitted(REFITMessageAuthentication mac) {
		if (committed) return;
		if (!prepared) return;
		if (prePrepare == null) return;

//		if(!REFITConfig.USE_PASSIVE_REPLICATION) {
//			for(int i = 0; i < PARTICIPATION.length; i++) {
//				if(PARTICIPATION[i] <= 0) return;
//			}
//		}

		ByteBuffer decision = commits.getDecision();
		if (decision == null) return;
		// In the current implementation, a non-faulty replica that
		// has been provided with a faulty PRE_PREPARE will never
		// reach the status "committed" in the corresponding instance,
		// even if it is able to obtain 2f+1 correct COMMITs; see the
		// following code line. This is not a safety problem but
		// implementing a solution (which would require obtaining
		// the correct message) would lead to a more robust protocol.
		byte[] shortHash = prePrepare.calcShortHash();
		if (!decision.equals(ByteBuffer.wrap(shortHash))) return;

		SigShare[] sigs = new SigShare[COMMITTED_THRESHOLD];
		List<REFITStewardCommit> desc = commits.getDecidingBallots();
		for (int i = 0; i < sigs.length; i++) {
			sigs[i] = desc.get(i).share;
		}

		try {
			combinedSig = mac.combineShares(groupId, shortHash, sigs);
		} catch (ThresholdSigException e) {
			REFITLogger.logError(this, "Failed to combine signature shares");
			// FIXME: implement handling of corrupted signatures
			return;
		}

		committed = true;
	}

	public BigInteger getCombinedSig() {
		return combinedSig;
	}

	// ##################
	// # PROGRESS PROOF #
	// ##################

	public REFITStewardMessage[] getProof() {
		if (isPrepared()) return createProof(prepares, PREPARED_THRESHOLD);
		return new REFITStewardMessage[0];
	}

	private static REFITStewardMessage[] createProof(REFITBallotBox<?, ?, ? extends REFITStewardMessage> ballotBox, int proofSize) {
		REFITStewardMessage[] proof = new REFITStewardMessage[proofSize];
		int i = 0;
		for (REFITStewardMessage message : ballotBox.getDecidingBallots()) {
			proof[i] = message;
			if (++i >= proofSize) break;
		}
		return proof;
	}
}
