package refit.agreement.steward;

import java.math.BigInteger;
import java.nio.ByteBuffer;
import java.security.MessageDigest;

import refit.message.REFITMessage;
import refit.message.REFITMessageType;
import refit.message.REFITNoOpRequest;
import refit.message.REFITOrderRequest;
import refit.replica.order.REFITOrderProtocolMessage;
import refit.util.REFITBigIntHelper;
import refit.util.REFITPayload;
import threshsig.SigShare;
import threshsig.Verifier;


public class REFITStewardMessages {

	// #####################
	// # ABSTRACT MESSAGES #
	// #####################

	public static abstract class REFITStewardMessage extends REFITOrderProtocolMessage {

		public static final int STEWARD_HEADER_SIZE = Integer.BYTES;

		protected ByteBuffer command;


		protected REFITStewardMessage(REFITMessageType pbftType, long instanceID, int viewID, ByteBuffer command, short from) {
			super(pbftType, instanceID, viewID, from);
			this.command = command;
		}

		protected REFITStewardMessage(REFITMessageType expectedType, ByteBuffer buffer) {
			super(expectedType, buffer);
			int cmdSize = buffer.getInt();
			this.command = buffer.slice();
			this.command.limit(cmdSize);
			buffer.position(buffer.position() + cmdSize);
		}

		@Override
		public AuthenticationType getAuthenticationType() {
			return AuthenticationType.SIGNATURE;
		}

		@Override
		protected void serialize(ByteBuffer buffer) {
			super.serialize(buffer);
			serializeCommand(buffer);
		}

		protected void serializeCommand(final ByteBuffer buffer) {
			int cmdSize = command.remaining();
			buffer.putInt(cmdSize);
			// replace command ByteBuffer after serialization to store the data only once
			ByteBuffer newCommand = buffer.slice();
			buffer.put(command);
			newCommand.limit(cmdSize);
			command = newCommand;
		}

		public ByteBuffer getCommand() {
			return command.slice();
		}

		@Override
		protected int getHeaderSize() {
			return (super.getHeaderSize() + STEWARD_HEADER_SIZE);
		}

		@Override
		protected int calculatePayloadSize() {
			return (super.calculatePayloadSize() + STEWARD_HEADER_SIZE + calculateCommandSize());
		}

		protected int calculateCommandSize() {
			return command.remaining();
		}

		@Override
		public String toString() {
			return String.format("{%s|%d|%d|%d}", type, from, uid.seqNr, viewID);
		}

	}

	// #####################
	// # SPECIFIC MESSAGES #
	// #####################

	public static class REFITStewardPrePrepare extends REFITStewardMessage {

		public final transient REFITOrderRequest message;
		public static final int SHORT_HASH_LENGTH = REFITPayload.HASH_SIZE + Integer.BYTES + Integer.BYTES + Long.BYTES;


		public REFITStewardPrePrepare(long instanceID, int viewID, REFITOrderRequest message, short from) {
			super(REFITMessageType.STEWARD_PRE_PREPARE, instanceID, viewID, null, from);
			this.message = message;
		}

		public REFITStewardPrePrepare(ByteBuffer buffer) {
			super(REFITMessageType.STEWARD_PRE_PREPARE, buffer);
			// create buffer slice to keep the invariant that the ByteBuffer passed
			// to a message constructor must be at position zero
			ByteBuffer cmd = getCommand();
			if (cmd.hasRemaining()) {
				this.message = (REFITOrderRequest) REFITMessage.createMessage(cmd);
			} else {
				this.message = REFITNoOpRequest.INSTANCE;
			}
		}


		@Override
		public String toString() {
			return String.format("{%s|%d|%d|%d|%s}", type, from, uid.seqNr, viewID, message);
		}

		@Override
		protected void serializeCommand(final ByteBuffer buffer) {
			if (message.type == REFITMessageType.REQUEST_NO_OP) {
				command = ByteBuffer.allocate(0);
				return;
			}

			int cmdSize = message.calculateMessageSize();
			buffer.putInt(cmdSize);
			// replace command ByteBuffer after serialization to store the data only once
			ByteBuffer newCommand = buffer.slice();
			message.serializeInto(buffer);
			newCommand.limit(cmdSize);
			command = newCommand;
		}

		@Override
		protected int calculateCommandSize() {
			if (message.type == REFITMessageType.REQUEST_NO_OP) {
				return 0;
			} else {
				return message.calculateMessageSize();
			}
		}

		@Override
		protected byte[] calculateHash() {
			// Prepare message header
			ByteBuffer messageHeader = getInternalBuffer();
			messageHeader.limit(getHeaderSize());

			// Prepare message (must be done in advance in order to prevent the hash from
			// being computed during the calculation of the REFITStewardPrePrepare hash)
			byte[] messageHash = message.getHash();

			// Calculate hash
			MessageDigest digest = REFITPayload.getDigest();
			digest.update(messageHeader);
			digest.update(messageHash);
			return digest.digest();
		}

		public byte[] calcShortHash() {
			ByteBuffer buffer = ByteBuffer.allocate(SHORT_HASH_LENGTH);
			buffer.putInt(0);
			buffer.putInt(viewID);
			buffer.putLong(uid.seqNr);
			buffer.put(message.getHash());
			return buffer.array();
		}
	}

	public static class REFITStewardPrepare extends REFITStewardMessage {

		public REFITStewardPrepare(REFITStewardPrePrepare prePrepare, short from) {
			super(REFITMessageType.STEWARD_PREPARE, prePrepare.uid.seqNr, prePrepare.viewID,
					ByteBuffer.wrap(prePrepare.calcShortHash()), from);
		}

		public REFITStewardPrepare(ByteBuffer buffer) {
			super(REFITMessageType.STEWARD_PREPARE, buffer);
		}

	}

	public static class REFITStewardCommit extends REFITStewardMessage {

		public final SigShare share;
		public final byte[] prepareHash;

		public REFITStewardCommit(REFITStewardPrePrepare prePrepare, short from, SigShare share) {
			super(REFITMessageType.STEWARD_COMMIT, prePrepare.uid.seqNr, prePrepare.viewID,
					null, from);
			this.share = share;
			this.prepareHash = prePrepare.calcShortHash();
		}

		public REFITStewardCommit(ByteBuffer buffer) {
			super(REFITMessageType.STEWARD_COMMIT, buffer);
			// create buffer slice to keep the invariant that the ByteBuffer passed
			// to a message constructor must be at position zero
			ByteBuffer cmd = getCommand();
			prepareHash = new byte[REFITStewardPrePrepare.SHORT_HASH_LENGTH];
			cmd.get(prepareHash);

			int shareId = cmd.getInt();
			BigInteger sig = getBigInt(cmd);
			BigInteger z = getBigInt(cmd);
			BigInteger c = getBigInt(cmd);

			// FIXME: should reconstruct share and group verifier using the signers info
			Verifier verif = new Verifier(z, c, null, null);
			share = new SigShare(shareId, sig, verif);
		}

		@Override
		protected void serializeCommand(ByteBuffer buffer) {
			int cmdSize = calculateCommandSize();
			buffer.putInt(cmdSize);
			// replace command ByteBuffer after serialization to store the data only once
			ByteBuffer newCommand = buffer.slice();
			buffer.put(prepareHash);

			buffer.putInt(share.getId());
			putBigInt(buffer, share.getSig());

			Verifier verif = share.getSigVerifier();
			putBigInt(buffer, verif.getZ());
			putBigInt(buffer, verif.getC());

			newCommand.limit(cmdSize);
			command = newCommand;
		}

		@Override
		protected int calculateCommandSize() {
			Verifier verif = share.getSigVerifier();
			return prepareHash.length + Integer.BYTES + getLen(share.getSig()) + getLen(verif.getZ()) + getLen(verif.getC());
		}

		private void putBigInt(ByteBuffer buffer, BigInteger bigint) {
			REFITBigIntHelper.putBigInt(buffer, bigint);
		}

		private BigInteger getBigInt(ByteBuffer buffer) {
			return REFITBigIntHelper.getBigInt(buffer);
		}

		private int getLen(BigInteger bigint) {
			return REFITBigIntHelper.getLen(bigint);
		}

	}
}
