package refit.agreement.weave;

import java.util.ArrayList;
import java.util.List;

import refit.agreement.weave.REFITWeaveMessages.REFITWeaveMessageCommit;
import refit.agreement.weave.REFITWeaveMessages.REFITWeaveMessagePrepare;
import refit.agreement.weave.REFITWeaveTransition.REFITWeaveTransitionProtocolInstance;
import refit.client.REFITClientLibrary.RequestMode;
import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.*;
import refit.replica.order.*;
import refit.util.REFITAssert;


public class REFITWeave {

	// ############
	// # PROTOCOL #
	// ############

	public static class REFITWeaveProtocol extends REFITOrderProtocol {

		public static final REFITWeaveProtocol instance = new REFITWeaveProtocol();

		@Override
		public REFITOrderProtocolInstance createInstance(REFITOrderStageSlot slot) {
			return new REFITWeaveProtocolInstance(slot, this);
		}

		@Override
		public REFITOrderProtocolTransitionInstance createTransitionInstance(REFITOrderStageSlot slot) {
			return new REFITWeaveTransitionProtocolInstance(slot);
		}

	}

	public static class REFITWeaveProtocolInstance extends REFITOrderProtocolInstance {

		private static final int COMMIT_QUORUM_SIZE = (REFITConfig.MENCIUS_MODE) ? (REFITConfig.ORDER_GROUP_SIZE / 2 + 1) : (REFITConfig.FAULTS_TO_TOLERATE + 1);

		private enum State {
			INITIALIZED,
			PREPARED,
			COMMITTED,
		}

		protected final REFITOrderProtocol protocol;
		private boolean isPrimary;
		private State state;
		private REFITOrderRequest proposal;
		private REFITWeaveMessagePrepare prepare;
		private boolean commitSent;
		private int currentCommits;
		private List<Short> commitsReceivedFrom;
		private boolean viewAborted = false;

		protected boolean[] allOrderReplicas;

		public REFITWeaveProtocolInstance(REFITOrderStageSlot slot, REFITOrderProtocol protocol) {
			super(slot);
			this.protocol = protocol;
			REFITAssert.assertTrue(!REFITConfig.AUTHENTICATE_MESSAGES, "WEAVE cannot authenticate messages");
		}

		// ##################
		// # HELPER METHODS #
		// ##################

		@Override
		public String toString() {
			return "WEAVE[" + instanceID + "]";
		}

		@Override
		public boolean isProposer() {
			return (isPrimary && (state == State.INITIALIZED) && proposal == null);
		}

		public boolean hasProposal() {
			return state != State.INITIALIZED || proposal != null;
		}

		@Override
		public boolean isPrepared() {
			return state == State.PREPARED || state == State.COMMITTED;
		}

		@Override
		public void assignProposal(final REFITOrderRequest message) {
			REFITAssert.assertTrue(isProposer());
			proposal = message;
		}

		private void orderReplicaMulticast(REFITMessage message) {
			message.serializeMessage();
			replica.mac.appendNullMAC(message);
			replica.replicaMulticast(message, allOrderReplicas);
		}

		// ######################
		// # LIFE-CYCLE METHODS #
		// ######################

		@Override
		public void init(long instanceID, int viewID) {
			super.init(instanceID, viewID);

			int groupID = REFITOrderGroups.getGroupID(instanceID);
			allOrderReplicas = protocol.getAllOrderRecipients(groupID);

			isPrimary = (replica.id == REFITOrderGroups.getPrimary(instanceID, viewID));
			state = State.INITIALIZED;
			proposal = null;
			prepare = null;
			commitSent = false;
			currentCommits = 0;
			commitsReceivedFrom = new ArrayList<>();

			if (!allOrderReplicas[replica.id]) {
				throw new InternalError("Unexpected");
			}
		}

		public REFITOrderProtocolData abort() {
			viewAborted = true;
			throw new UnsupportedOperationException("Not implemented");
		}


		// ######################
		// # PROTOCOL EXECUTION #
		// ######################

		@Override
		public void execute() {
			REFITAssert.assertTrue(!viewAborted, "Execute called on slot with aborted view");
			if (isPrimary) executePrimary();
			else executeFollower();
		}

		private void executePrimary() {
			receivePrepares();
			switch (state) {
			case INITIALIZED:
				// ignore unexpected activation from outside
				if (proposal == null) return;

				prepare = new REFITWeaveMessagePrepare(instanceID, viewID, proposal, replica.id);

				if (REFITConfig.MENCIUS_MODE && proposal.type == REFITMessageType.REQUEST_NO_OP) {
					// a SKIP message can only be triggered by a message with a higher seqNr;
					// the corresponding prepare or commit messages count as implicit acks and each replica
					// broadcasts its prepare or commit message to all other replicas. That is each replica
					// automatically and immediately sends its implicit SKIP messages
					state = State.COMMITTED;
					completeWithResult(proposal, null);
					return;
				}

				orderReplicaMulticast(prepare);
				// Own prepare counts as commit
				currentCommits++;

				state = State.PREPARED;

				// Creating the prepare counts as receiving it
				prepareReceived(prepare.from);
				//$FALL-THROUGH$
			case PREPARED:
				receiveCommits();
				tryComplete();
				break;
			case COMMITTED:
				receiveCommits();
				break;
			default:
				REFITAssert.fail("Unexpected call in state " + state + " in primary");
			}
		}

		private void executeFollower() {
			switch (state) {
			case INITIALIZED:
				receivePrepares();
				//$FALL-THROUGH$
			case PREPARED:
				receiveCommits();
				if (!commitSent && prepare != null) {
					trySendCommit();
				}
				tryComplete();
				break;
			case COMMITTED:
				receiveCommits();
				break;
			default:
				REFITAssert.fail("Unexpected call in state " + state + " in follower");
			}
		}

		private void receivePrepares() {
			while (true) {
				REFITWeaveMessagePrepare prepare =
						(REFITWeaveMessagePrepare) fetchMessage(REFITMessageType.WEAVE_PREPARE, viewID);
				if (prepare == null) return;

				if (!REFITWeaveVerifyMessage.verifyPrepare(prepare)) {
					REFITLogger.logWarning(this, "Invalid order protocol PREPARE message " + prepare);
					continue;
				}

				if (proposal == null) {
					this.prepare = prepare;
					proposal = prepare.request;

					if (REFITConfig.MENCIUS_MODE && proposal.type == REFITMessageType.REQUEST_NO_OP) {
						prepareReceived(this.prepare.from);
						state = State.COMMITTED;
						// don't send commit
						commitSent = true;
						completeWithResult(proposal, null);
						return;
					} else {
						// Leader prepare counts as commit
						currentCommits++;
						state = State.PREPARED;

						prepareReceived(this.prepare.from);
					}
				}
			}
		}

		private void trySendCommit() {
			final REFITWeaveMessageCommit commit = new REFITWeaveMessageCommit(instanceID, viewID, replica.id);
			orderReplicaMulticast(commit);

			currentCommits++;
			commitSent = true;
		}

		private void receiveCommits() {
			if (prepare == null) return;
			// nothing to do for SKIP messages
			if (REFITConfig.MENCIUS_MODE && proposal.type == REFITMessageType.REQUEST_NO_OP) return;

			while (true) {
				final REFITWeaveMessageCommit commitMessage =
						(REFITWeaveMessageCommit) fetchMessage(REFITMessageType.WEAVE_COMMIT, viewID);
				if (commitMessage == null) return;

				if (!REFITWeaveVerifyMessage.verifyCommit(commitMessage, prepare)) {
					REFITLogger.logWarning(this, "Invalid order protocol COMMIT message");
					continue;
				}

				if (!commitsReceivedFrom.contains(commitMessage.from)) {
					commitsReceivedFrom.add(commitMessage.from);
					currentCommits++;

					if (isPrimary) {
						commitReceived(commitMessage.from);
					}
				}
			}
		}

		private void sendRequestAckReplies() {
			// NO_OPs are proposed by the replica itself so nothing needs to be sent to a client
			switch (proposal.type) {
			case REQUEST_NO_OP:
				break;
			case BATCH:
				for (REFITRequest request : ((REFITBatch) proposal).requests) {
					sendIndividualAck(request);
				}
				break;
			case REQUEST:
				sendIndividualAck((REFITRequest) proposal);
				break;
			default:
				throw new IllegalStateException("Unexpected proposal message type");
			}
		}

		private void sendIndividualAck(REFITRequest request) {
			if (!request.mode.equals(RequestMode.WRITE_ACK)) return;

			short groupID = REFITOrderGroups.getGroupID(instanceID);
			REFITAckReply requestAckReply = new REFITAckReply(request.uid, replica.id, instanceID, groupID, request.replyReplicaID);
			ackStable(requestAckReply);
		}

		private void tryComplete() {
			if (REFITConfig.MENCIUS_MODE && proposal != null && proposal.type == REFITMessageType.REQUEST_NO_OP) return;

			if (currentCommits >= COMMIT_QUORUM_SIZE && (isPrimary || commitSent)) {
				state = State.COMMITTED;
				// ACK can be sent only after a commit certificate is reached to ensure that the request cannot be lost
				sendRequestAckReplies();
				completeWithResult(proposal, null);
			}
		}
	}
}
