package refit.application.counter;

import java.nio.ByteBuffer;
import java.util.HashSet;
import java.util.Set;

import refit.application.REFITApplicationServer;
import refit.application.REFITResult;
import refit.config.REFITConfig;
import refit.message.REFITUniqueID;
import refit.replica.checkpoint.REFITCheckpointObject;
import refit.replica.fuzzycheckpoint.REFITFuzzyCheckpointContainer;

public class REFITCounterServer implements REFITApplicationServer {

	private final int[] counters;


	public REFITCounterServer() {
		this.counters = new int[REFITConfig.TOTAL_NR_OF_CLIENTS];
	}


	@Override
	public void init() {
		// Do nothing
	}

	@Override
	public boolean checkReadOnly(ByteBuffer payload) {
		return payload.remaining() > 0 && payload.get(payload.position()) != 0;
	}

	@Override
	public REFITResult processRequest(REFITUniqueID uid, long agreementSeqNr, ByteBuffer payload, boolean createUpdate) {
		// Create reply
		short counterID = (short) (uid.nodeID - REFITConfig.TOTAL_NR_OF_REPLICAS);
		ByteBuffer result = ByteBuffer.allocate(Math.max(REFITConfig.REPLY_SIZE, Integer.BYTES));
		boolean readOnly = checkReadOnly(payload);

		if (!readOnly) {
			++counters[counterID];
		}
		result.putInt(counters[counterID]);
		result.flip();

		ByteBuffer update = null; // Leave (update == null) for read-only requests
		if ((createUpdate || REFITConfig.ENABLE_DIFFERENTIAL_CHECKPOINTING) && !readOnly) {
			if (REFITConfig.UPDATE_BASED_CHECKPOINTING) {
				// Create update
				ByteBuffer updateBuffer = ByteBuffer.allocate(Math.max(REFITConfig.UPDATE_SIZE, Short.BYTES + Integer.BYTES));
				updateBuffer.putShort(counterID);
				updateBuffer.putInt(counters[counterID]);
				updateBuffer.flip();
				update = updateBuffer;
			} else {
				changedObjects.add(counterID);
			}
		}

		// Return result
		return new REFITResult(result, update);
	}

	@Override
	public String toString() {
		return "CNTR";
	}

	@Override
	public void applyUpdate(ByteBuffer update) {
		short counterID = update.getShort();
		counters[counterID] = update.getInt();
	}

	private Set<Short> changedObjects = new HashSet<>();

	@Override
	public Set<Short> takeChangelist() {
		Set<Short> changed = changedObjects;
		changedObjects = new HashSet<>();
		return changed;
	}

	@Override
	public REFITCheckpointObject getObject(Object key) {
		short counterID = (short) key;

		ByteBuffer updateBuffer = ByteBuffer
				.allocate(Math.max(REFITConfig.UPDATE_SIZE, Short.BYTES + Integer.BYTES));
		updateBuffer.putShort(counterID);
		updateBuffer.putInt(counters[counterID]);
		updateBuffer.flip();
		return new REFITCheckpointObject(updateBuffer);
	}

	@Override
	public REFITCheckpointObject[] createCheckpoint() {
		ByteBuffer checkpoint = ByteBuffer.allocate(Integer.BYTES * counters.length);
		for (int counter : counters) checkpoint.putInt(counter);
		return new REFITCheckpointObject[]{new REFITCheckpointObject(checkpoint.array())};
	}

	@Override
	public void applyCheckpoint(final REFITCheckpointObject[] checkpoint) {
//		System.out.println("Apply checkpoint");
		ByteBuffer checkpointBuffer = checkpoint[0].getBuffer();
		for (int i = 0; i < counters.length; i++) counters[i] = checkpointBuffer.getInt();
	}

	@Override
	public REFITFuzzyCheckpointContainer<? extends REFITCheckpointObject, ? extends REFITCheckpointObject> createFuzzyCheckpoint() {
		// FIXME implement
		throw new UnsupportedOperationException("Not implemented yet");
	}
}
