package refit.application.probe;

import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.Set;

import refit.application.REFITLeaderRecommendation;
import refit.application.REFITResult;
import refit.application.REFITSystemApplicationServer;
import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITProbeData;
import refit.message.REFITUniqueID;
import refit.replica.checkpoint.REFITCheckpointObject;
import refit.replica.fuzzycheckpoint.REFITFuzzyCheckpointContainer;

public class REFITProbeSystemServer implements REFITSystemApplicationServer {

	private static class ProbeCollection {
		private long[] probeStamps = new long[REFITConfig.TOTAL_NR_OF_CLIENTS];
		private REFITProbeData[] probes = new REFITProbeData[REFITConfig.TOTAL_NR_OF_CLIENTS];
		private Short[] replyTime = new Short[REFITConfig.TOTAL_NR_OF_CLIENTS];
		private Short[][] passiveReplyTime = new Short[REFITConfig.TOTAL_NR_OF_REPLICAS][REFITConfig.TOTAL_NR_OF_CLIENTS];
	}

	private ProbeCollection[] data = new ProbeCollection[REFITConfig.TOTAL_NR_OF_REPLICAS];
	private long probeCounter = 0;


	private static final int STABILITY_THRESHOLD = REFITConfig.FAULTS_TO_TOLERATE + 1;
	private static final int PREPARED_THRESHOLD = (REFITConfig.TOTAL_NR_OF_REPLICAS + REFITConfig.FAULTS_TO_TOLERATE + 1 + 1) / 2;
	private static final int COMMITTED_THRESHOLD = (REFITConfig.TOTAL_NR_OF_REPLICAS + REFITConfig.FAULTS_TO_TOLERATE + 1 + 1) / 2;


	public REFITProbeSystemServer() {
		for (int i = 0; i < data.length; i++) {
			data[i] = new ProbeCollection();
		}
	}

	@Override
	public String toString() {
		return "PRBSS";
	}

	@Override
	public void init() {
	}

	@Override
	public boolean checkReadOnly(ByteBuffer payload) {
		// delivery of probe data always modifies state
		return false;
	}

	@Override
	public REFITResult processRequest(REFITUniqueID uid, long agreementSeqNr, ByteBuffer payload, boolean createUpdate) {
		short clientID = (short) (uid.nodeID - REFITConfig.TOTAL_NR_OF_REPLICAS);
		processProbe(payload, clientID);

		payload.rewind();
		return new REFITResult(ByteBuffer.wrap(new byte[0]), (createUpdate) ? createUpdate(clientID, payload) : null);
	}

	private void processProbe(ByteBuffer bb, short clientID) {
		for (int i = 0; i < REFITConfig.TOTAL_NR_OF_REPLICAS; i++) {
			REFITProbeData probe = new REFITProbeData(bb.slice());
			bb.position(bb.position() + probe.getMessageSize());
			ProbeCollection collection = data[probe.primaryID];
			collection.probes[clientID] = probe;
			collection.replyTime[clientID] = calculateQuorumTime(probe.replyTimes, STABILITY_THRESHOLD);
			collection.probeStamps[clientID] = probeCounter;
			if (REFITConfig.USE_PASSIVE_REPLICATION) {
				short[] replyTimes = calculatePassiveReplyTimes(probe);
				for (int j = 0; j < REFITConfig.TOTAL_NR_OF_REPLICAS; j++) {
					collection.passiveReplyTime[j][clientID] = replyTimes[j];
				}
			}
			probeCounter++;

//        if ((probeCounter % 100) == 0) {
			printProbe(clientID, probe);
			printRecommendation();
//        }
		}
	}

	private void printProbe(short clientID, REFITProbeData probe) {
		ProbeCollection collection = data[probe.primaryID];
		REFITLogger.logDebug(this, "Raw probe from client " + clientID + " in " + collection.replyTime[clientID] + " with leader " + probe.primaryID);

		REFITLogger.logDebug(this, "Raw replies " + Arrays.toString(probe.replyTimes));
		for (int i = 0; i < probe.preparedTimes.length; i++) {
			short[] prepares = probe.preparedTimes[i];
			REFITLogger.logDebug(this, "Raw prepares " + i + " " + Arrays.toString(prepares));
		}
		for (int i = 0; i < probe.committedTimes.length; i++) {
			short[] commits = probe.committedTimes[i];
			REFITLogger.logDebug(this, "Raw commits " + i + " " + Arrays.toString(commits));
		}
	}

	private void printRecommendation() {
		REFITLeaderRecommendation recommendation = queryLeader(-1, -REFITConfig.PROBE_WARMUP);
		if (recommendation != null) {
			REFITLogger.logDebug(this, "Recommend leader " + recommendation.leader + " with " + recommendation.percentile);
			if (REFITConfig.USE_PASSIVE_REPLICATION) {
				REFITLogger.logDebug(this, "Recommend passive replicas " + Arrays.toString(recommendation.passive));
			}
		} else {
			REFITLogger.logDebug(this, "No recommendation");
		}
	}

	private short calculateQuorumTime(short[] quorumTimes, int quorumSize) {
		short[] times = Arrays.copyOf(quorumTimes, quorumTimes.length);
		Arrays.sort(times);
		return times[quorumSize - 1];
	}

	private short[] calculatePassiveReplyTimes(REFITProbeData probe) {
		if (REFITConfig.FAULTS_TO_TOLERATE != 1) {
			throw new AssertionError("Not supported!");
		}

		short[] prepareQuorumTime = quorumTimes(probe.preparedTimes, PREPARED_THRESHOLD);
		short[] commitQuorumTime = quorumTimes(probe.committedTimes, COMMITTED_THRESHOLD);

		short[] replyTimes = new short[REFITConfig.TOTAL_NR_OF_REPLICAS];
		for (int i = 0; i < REFITConfig.TOTAL_NR_OF_REPLICAS; i++) {
			if (i == probe.primaryID) {
				replyTimes[i] = REFITConfig.PROBE_TIMEOUT;
				continue;
			}

			// drop messages of passive replica
			short[][] modPrepareTimes = copy(probe.preparedTimes);
			strikeoutReplica(i, modPrepareTimes);
			short[] modPrepareQuorumTime = quorumTimes(modPrepareTimes, PREPARED_THRESHOLD);

			// propagate delays
			short[][] modCommitTimes = copy(probe.committedTimes);
			for (int j = 0; j < REFITConfig.TOTAL_NR_OF_REPLICAS; j++) {
				for (int k = 0; k < REFITConfig.TOTAL_NR_OF_REPLICAS; k++) {
					int delta = modPrepareQuorumTime[j] - prepareQuorumTime[j];
					modCommitTimes[k][j] += delta;
				}
			}
			// drop messages of passive replica
			strikeoutReplica(i, modCommitTimes);
			short[] modCommitQuorumTime = quorumTimes(modCommitTimes, COMMITTED_THRESHOLD);

			// propagate delays
			short[] modReplyTimes = Arrays.copyOf(probe.replyTimes, probe.replyTimes.length);
			for (int j = 0; j < REFITConfig.TOTAL_NR_OF_REPLICAS; j++) {
				int delta = modCommitQuorumTime[j] - commitQuorumTime[j];
				modReplyTimes[j] += delta;
			}
			modReplyTimes[i] = REFITConfig.PROBE_TIMEOUT;

			replyTimes[i] = calculateQuorumTime(modReplyTimes, STABILITY_THRESHOLD);
		}
		return replyTimes;
	}

	private void strikeoutReplica(int i, short[][] modPrepareTimes) {
		for (int j = 0; j < REFITConfig.TOTAL_NR_OF_REPLICAS; j++) {
			modPrepareTimes[i][j] = REFITConfig.PROBE_TIMEOUT;
			modPrepareTimes[j][i] = REFITConfig.PROBE_TIMEOUT;
		}
	}

	private short[] quorumTimes(short[][] times, int threshold) {
		short[] prepareQuorumTime = new short[REFITConfig.TOTAL_NR_OF_REPLICAS];
		for (int i = 0; i < prepareQuorumTime.length; i++) {
			prepareQuorumTime[i] = calculateQuorumTime(times[i], threshold);
		}
		return prepareQuorumTime;
	}

	private short[][] copy(short[][] times) {
		short[][] copy = new short[times.length][];
		for (int i = 0; i < times.length; i++) {
			copy[i] = Arrays.copyOf(times[i], times[i].length);
		}
		return copy;
	}

	private ByteBuffer createUpdate(short clientID, ByteBuffer payload) {
		byte[] updateData = new byte[Short.BYTES + payload.limit()];
		ByteBuffer bb = ByteBuffer.wrap(updateData);
		bb.putShort(clientID);
		bb.put(payload);
		bb.flip();
		return bb;
	}

	@Override
	public void applyUpdate(ByteBuffer bb) {
		short clientID = bb.getShort();
		processProbe(bb, clientID);
	}


	@Override
	public Set<?> takeChangelist() {
		throw new InternalError("not implemented yet");
	}

	@Override
	public REFITCheckpointObject getObject(Object key) {
		throw new InternalError("not implemented yet");
	}

	@Override
	public REFITFuzzyCheckpointContainer<? extends REFITCheckpointObject, ? extends REFITCheckpointObject> createFuzzyCheckpoint() {
		throw new InternalError("not implemented yet");
	}


	private byte[] checkpointBuffer = new byte[5 * 1000 * 1000];

	@Override
	public REFITCheckpointObject[] createCheckpoint() {
		ByteBuffer bb = ByteBuffer.wrap(checkpointBuffer);

		for (ProbeCollection collection : data) {
			for (int i = 0; i < collection.probes.length; i++) {
				REFITProbeData probe = collection.probes[i];
				if (probe != null) {
					bb.put((byte) 1);
					bb.putShort(collection.replyTime[i]);
					bb.putLong(collection.probeStamps[i]);
					if (REFITConfig.USE_PASSIVE_REPLICATION) {
						for (int j = 0; j < REFITConfig.TOTAL_NR_OF_REPLICAS; j++) {
							bb.putShort(collection.passiveReplyTime[j][i]);
						}
					}
					bb.put(probe.getBuffer());
				} else {
					bb.put((byte) 0);
				}
			}
		}

		bb.flip();
		return new REFITCheckpointObject[]{new REFITCheckpointObject(bb)};
	}


	@Override
	public void applyCheckpoint(REFITCheckpointObject[] checkpoint) {
		ByteBuffer bb = checkpoint[0].getBuffer();

		for (ProbeCollection collection : data) {
			for (int i = 0; i < collection.probes.length; i++) {
				boolean isPresent = bb.get() != 0;
				if (isPresent) {
					collection.replyTime[i] = bb.getShort();
					collection.probeStamps[i] = bb.getLong();
					if (REFITConfig.USE_PASSIVE_REPLICATION) {
						for (int j = 0; j < REFITConfig.TOTAL_NR_OF_REPLICAS; j++) {
							collection.passiveReplyTime[j][i] = bb.getShort();
						}
					}
					collection.probes[i] = new REFITProbeData(bb);
				} else {
					collection.replyTime[i] = null;
					collection.probes[i] = null;
					collection.probeStamps[i] = 0;
					if (REFITConfig.USE_PASSIVE_REPLICATION) {
						for (int j = 0; j < REFITConfig.TOTAL_NR_OF_REPLICAS; j++) {
							collection.passiveReplyTime[j][i] = null;
						}
					}
				}
			}
		}
	}

	@Override
	public REFITLeaderRecommendation queryLeader(int lastLeader, long targetProbeCounter) {
		// warmup phase after startup and view changes
		if (lastLeader >= 0 && probeCounter < targetProbeCounter + REFITConfig.PROBE_WARMUP) {
			return null;
		}

		garbageCollectOldProbes();
		if (REFITConfig.USE_PASSIVE_REPLICATION) {
			return calculatePassiveLeader(lastLeader);
		} else {
			return calculateActiveLeader(lastLeader);
		}
	}

	private void garbageCollectOldProbes() {
		// FIXME more efficient variant?
		for (ProbeCollection collection : data) {
			for (int i = 0; i < collection.probes.length; i++) {
				REFITProbeData probe = collection.probes[i];
				if (probe != null && collection.probeStamps[i] + REFITConfig.PROBE_MAX_AGE < probeCounter) {
					collection.replyTime[i] = null;
					collection.probes[i] = null;
					collection.probeStamps[i] = 0;
					for (int j = 0; j < REFITConfig.TOTAL_NR_OF_REPLICAS; j++) {
						collection.passiveReplyTime[j][i] = null;
					}
				}
			}
		}
	}

	private REFITLeaderRecommendation calculateActiveLeader(int lastLeader) {
		short[] replicaPercentiles = new short[REFITConfig.TOTAL_NR_OF_REPLICAS];
		for (int i = 0; i < replicaPercentiles.length; i++) {
			replicaPercentiles[i] = calculatePercentile(data[i].replyTime);
		}

		short bestPercentile = replicaPercentiles[0];
		short bestReplica = 0;
		for (short i = 1; i < replicaPercentiles.length; i++) {
			if (replicaPercentiles[i] < bestPercentile) {
				bestPercentile = replicaPercentiles[i];
				bestReplica = i;
			}
		}

		// keep previous leader unless the new one is significantly better
		if (lastLeader >= 0 && replicaPercentiles[lastLeader] < bestPercentile + REFITConfig.PROBING_DECISION_HYSTERESIS) {
			bestPercentile = replicaPercentiles[lastLeader];
			bestReplica = (short) lastLeader;
		}

		if (bestPercentile >= REFITConfig.PROBE_TIMEOUT) {
			// no data
			return null;
		}

		if (lastLeader >= 0) {
			REFITLogger.logDebug(this, "Last " + lastLeader + " with " + replicaPercentiles[lastLeader]);
		} else {
			REFITLogger.logDebug(this, "Active leaders " + Arrays.toString(replicaPercentiles));
		}
		return new REFITLeaderRecommendation(bestReplica, null, probeCounter, bestPercentile);
	}

	private short calculatePercentile(Short[] values) {
		int count = 0;
		for (Short value : values) {
			if (value != null) {
				count++;
			}
		}

		if (count == 0) {
			return REFITConfig.PROBE_TIMEOUT;
		}

		short[] compact = new short[count];
		count = 0;
		for (Short value : values) {
			if (value != null) {
				compact[count++] = value;
			}
		}

		Arrays.sort(compact);
		int index = ((compact.length - 1) * REFITConfig.PROBE_PERCENTILE + 99) / 100;
		return compact[index];
	}

	private REFITLeaderRecommendation calculatePassiveLeader(int lastLeader) {
		short[][] replicaPercentiles = new short[REFITConfig.TOTAL_NR_OF_REPLICAS][REFITConfig.TOTAL_NR_OF_REPLICAS];
		for (int i = 0; i < replicaPercentiles.length; i++) {
			for (int j = 0; j < replicaPercentiles[0].length; j++) {
				replicaPercentiles[i][j] = calculatePercentile(data[i].passiveReplyTime[j]);
			}
		}

		short bestPercentile = replicaPercentiles[0][1];
		short bestReplica = 0;
		short bestPassiveReplica = 1;
		for (short i = 0; i < replicaPercentiles.length; i++) {
			for (short j = 0; j < replicaPercentiles[0].length; j++) {
				if (i == j) continue;
				if (replicaPercentiles[i][j] < bestPercentile) {
					bestPercentile = replicaPercentiles[i][j];
					bestReplica = i;
					bestPassiveReplica = j;
				}
			}

			if (lastLeader == -1) {
				REFITLogger.logDebug(this, "Passive leader " + i + " percentiles " + Arrays.toString(replicaPercentiles[i]));
			}
		}

		// FIXME implement old leader hysteresis

		if (bestPercentile >= REFITConfig.PROBE_TIMEOUT) {
			// no data
			return null;
		}
		return new REFITLeaderRecommendation(bestReplica, new short[]{bestPassiveReplica}, probeCounter, bestPercentile);
	}
}
