package refit.application.rezk.common;

import java.nio.ByteBuffer;
import java.util.Arrays;

import refit.application.rezk.util.REFITByteString;
import refit.application.rezk.util.REFITSerialization;
import refit.application.rezk.util.REFITSerialization.REFITSerializationSize;
import refit.message.REFITBaseMessage;


public class REZKRequest extends REFITBaseMessage {

	public final REZKOperation operation;
	public final REFITByteString path;

	public byte[] data;
	public boolean watch;
	public boolean isSequential;
	public int version;


	public REZKRequest(REZKOperation operation, REFITByteString path) {
		super();
		this.operation = operation;
		this.path = path;
	}


	// ###########
	// # SETTERS #
	// ###########

	public void setData(byte[] data) {
		this.data = data;
	}

	public void setWatch(boolean watch) {
		this.watch = watch;
	}

	public void setSequential(boolean sequential) {
		this.isSequential = sequential;
	}

	public void setVersion(int version) {
		this.version = version;
	}


	// ###################################
	// # SERIALIZATION & DESERIALIZATION #
	// ###################################

	public REZKRequest(ByteBuffer buffer) {
		super(buffer);
		// Deserialize common attributes
		this.operation = REZKOperation.values[buffer.get()];
		this.path = REFITSerialization.getByteString(buffer, REFITSerializationSize.TINY);

		// Deserialize specific attributes
		switch (operation) {
		case CREATE:
			this.isSequential = REFITSerialization.getBoolean(buffer);
			this.data = REFITSerialization.getByteArray(buffer, REFITSerializationSize.SMALL);
			break;
		case DELETE:
		case INCREMENT:
		case DEQUEUE:
			// Nothing more to do
			break;
		case EXISTS:
		case GET_CHILDREN:
		case GET_DATA:
			this.watch = REFITSerialization.getBoolean(buffer);
			break;
		case SET_DATA:
			this.version = buffer.getInt();
			this.data = REFITSerialization.getByteArray(buffer, REFITSerializationSize.SMALL);
			break;
		default:
			throw new InternalError("Unexpected operation: " + operation);
		}
	}

	@Override
	protected void serialize(ByteBuffer buffer) {
		super.serialize(buffer);
		// Serialize common attributes
		REFITSerialization.putByte(buffer, operation.ordinal());
		REFITSerialization.putByteString(buffer, path, REFITSerializationSize.TINY);

		// Serialize specific attributes
		switch (operation) {
		case CREATE:
			REFITSerialization.putBoolean(buffer, isSequential);
			REFITSerialization.putByteArray(buffer, data, REFITSerializationSize.SMALL);
			break;
		case DELETE:
		case INCREMENT:
		case DEQUEUE:
			// Nothing more to do
			break;
		case EXISTS:
		case GET_CHILDREN:
		case GET_DATA:
			REFITSerialization.putBoolean(buffer, watch);
			break;
		case SET_DATA:
			buffer.putInt(version);
			REFITSerialization.putByteArray(buffer, data, REFITSerializationSize.SMALL);
			break;
		default:
			throw new InternalError("Unexpected operation: " + operation);
		}
	}

	@Override
	protected int getHeaderSize() {
		int size = super.getHeaderSize() + Byte.BYTES
				+ REFITSerialization.getByteStringSize(path, REFITSerializationSize.TINY);
		switch (operation) {
		case CREATE:
			size += REFITSerialization.getBooleanSize()
					+ REFITSerialization.getByteArraySize(data, REFITSerializationSize.SMALL);
			break;
		case DELETE:
		case INCREMENT:
		case DEQUEUE:
			break;
		case EXISTS:
		case GET_CHILDREN:
		case GET_DATA:
			size += REFITSerialization.getBooleanSize();
			break;
		case SET_DATA:
			size += Integer.BYTES + REFITSerialization.getByteArraySize(data, REFITSerializationSize.SMALL);
			break;
		default:
			throw new InternalError("Unexpected operation: " + operation);
		}
		return size;
	}

	@Override
	protected int calculatePayloadSize() {
		return getHeaderSize();
	}

	@Override
	public String toString() {
		return "REZKRequest [operation=" + operation + ", path=" + path + ", data=" + Arrays.toString(data) + ", watch="
				+ watch + ", isSequential=" + isSequential + ", version=" + version + "]";
	}
}
