package refit.application.rezk.datatree;

import java.nio.ByteBuffer;
import java.util.*;

import refit.application.rezk.REZKFuzzyCheckpointContainer;
import refit.application.rezk.REZKServer;
import refit.application.rezk.common.REZKException;
import refit.application.rezk.common.REZKStat;
import refit.application.rezk.common.REZKUpdate;
import refit.application.rezk.util.REFITByteString;
import refit.replica.checkpoint.REFITCheckpointObject;


public class REZKCachedDataTree implements REZKDataTree {

	private final REZKDataTree tree;
	private final Map<REFITByteString, REZKStat> stats;
	private final Map<REFITByteString, Set<REFITByteString>> children;


	public REZKCachedDataTree(REZKDataTree tree) {
		this.tree = tree;
		this.stats = new HashMap<>();
		this.children = new HashMap<>();
	}


	// ########
	// # TREE #
	// ########

	@Override
	public void init() throws REZKException {
		stats.clear();
		tree.init();
	}

	@Override
	public void shutdown() throws REZKException {
		tree.shutdown();
	}


	// ###################
	// # REZKClient OPERATIONS #
	// ###################

	@Override
	public void create(REFITByteString path, byte[] data, REZKStat stat) throws REZKException {
		tree.create(path, data, stat);
		stats.put(path, stat);
		addChild(path);
	}

	@Override
	public void delete(REFITByteString path) throws REZKException {
		tree.delete(path);
		stats.remove(path);
		removeChild(path);
	}

	@Override
	public Set<REFITByteString> getChildren(REFITByteString path) throws REZKException {
		return children.get(path);
	}

	@Override
	public byte[] getData(REFITByteString path, REZKStat stat) throws REZKException {
		return tree.getData(path, stat);
	}

	@Override
	public void setData(REFITByteString path, byte[] data, REZKStat stat) throws REZKException {
		tree.setData(path, data, stat);
		stats.put(path, stat);
	}


	// #########
	// # STATS #
	// #########

	@Override
	public REZKStat getStat(REFITByteString path) throws REZKException {
		return stats.get(path);
	}

	@Override
	public void updateStat(REFITByteString path, REZKStat stat) throws REZKException {
		tree.updateStat(path, stat);
		stats.put(path, stat);
	}


	// ###########
	// # UPDATES #
	// ###########

	@Override
	public void applyUpdates(Collection<REZKUpdate> updates) throws REZKException {
		// Apply updates
		tree.applyUpdates(updates);

		// Update cache
		for (REZKUpdate update : updates) {
			switch (update.type) {
			case CREATE:
			case SET_STAT:
				stats.put(update.path, update.stat);
				addChild(update.path);
				break;
			case DELETE:
				stats.remove(update.path);
				removeChild(update.path);
				break;
			case SET_DATA:
				// Do nothing
				break;
			default:
				System.err.println("Unexpected update type: " + update.type);
			}
		}
	}


	// ###############
	// # CHECKPOINTS #
	// ###############

	@Override
	public void createCheckpoint(final Queue<ByteBuffer> buffer) throws REZKException {
		tree.createCheckpoint(buffer);
	}

	@Override
	public void createFuzzyCheckpoint(REZKFuzzyCheckpointContainer checkpoint) throws REZKException {
		tree.createFuzzyCheckpoint(checkpoint);
	}

	@Override
	public void createFuzzyDiffCheckpoint(Set<?> changelist, Map<Object, REFITCheckpointObject> changedObjects) throws REZKException {
		tree.createFuzzyDiffCheckpoint(changelist, changedObjects);
	}

	@Override
	public void createFuzzyUpdateCheckpoint(Map<?, REFITCheckpointObject> changelist, Collection<REFITCheckpointObject> changedObjects) throws REZKException {
		tree.createFuzzyUpdateCheckpoint(changelist, changedObjects);
	}

	@Override
	public Map<REFITByteString, REZKStat> applyCheckpoint(ByteBuffer buffer) throws REZKException {
		// Apply checkpoint
		Map<REFITByteString, REZKStat> newStats = tree.applyCheckpoint(buffer);

		// Update cache
		stats.clear();
		for (Map.Entry<REFITByteString, REZKStat> entry : newStats.entrySet()) {
			stats.put(entry.getKey(), entry.getValue());
			addChild(entry.getKey());
		}
		return newStats;
	}


	// ############
	// # CHILDREN #
	// ############

	private static final REFITByteString SLASH = new REFITByteString("/");

	private void addChild(REFITByteString path) throws REZKException {
		// Create cache entry for this node
		children.put(path, new LinkedHashSet<REFITByteString>());
		if (SLASH.equals(path)) return;

		// Update children
		REFITByteString parentPath = REZKServer.getParentPath(path);
		Set<REFITByteString> childrenSet = children.get(parentPath);
		REFITByteString child = REZKServer.getChildPath(path);
		childrenSet.add(child);
	}

	private void removeChild(REFITByteString path) throws REZKException {
		// Remove cache entry for this node
		children.remove(path);
		if (SLASH.equals(path)) return;

		// Update children
		REFITByteString parentPath = REZKServer.getParentPath(path);
		Set<REFITByteString> childrenSet = children.get(parentPath);
		REFITByteString child = REZKServer.getChildPath(path);
		childrenSet.remove(child);
	}

}
