package refit.application.rezk.datatree;

import java.nio.ByteBuffer;
import java.sql.*;
import java.util.*;

import refit.application.rezk.REZKFuzzyCheckpointContainer;
import refit.application.rezk.REZKServer;
import refit.application.rezk.common.REZKException;
import refit.application.rezk.common.REZKStat;
import refit.application.rezk.common.REZKUpdate;
import refit.application.rezk.util.REFITByteString;
import refit.application.rezk.util.REFITSerialization;
import refit.application.rezk.util.REFITSerialization.REFITSerializationSize;
import refit.replica.checkpoint.REFITCheckpointObject;


public abstract class REZKSQLDataTree implements REZKDataTree {

	public static final String TABLE_NAME = "rezk";

	private Connection database;
	private REZKSQLOperationContainer operations;

	// ########
	// # TREE #
	// ########

	protected abstract Connection initDatabase() throws REZKException;

	protected abstract void initTable() throws REZKException;

	@Override
	public void init() throws REZKException {
		// Initialize database
		database = initDatabase();
		initTable();

		// Prepare statements
		operations = new REZKSQLOperationContainer(database);
	}

	@Override
	public void shutdown() throws REZKException {
		try {
			database.close();
			database = null;
		} catch (SQLException sqle) {
			sqle.printStackTrace();
			throw new REZKException(sqle.getMessage());
		}
	}


	// ###################
	// # REZKClient OPERATIONS #
	// ###################

	@Override
	public void create(REFITByteString path, byte[] data, REZKStat stat) throws REZKException {
		try {
			REFITByteString parentPath = REZKServer.getParentPath(path);
			PreparedStatement statement = operations.CREATE;
			statement.setBytes(1, path.getBytes());
			statement.setBytes(2, parentPath.getBytes());
			statement.setBytes(3, data);
			statement.setInt(4, stat.version);
			statement.setInt(5, stat.dataLength);
			statement.setInt(6, stat.nrOfChildren);
			statement.setLong(7, stat.lastAccess);
			executePreparedCommand(statement);
		} catch (SQLException sqle) {
			sqle.printStackTrace();
			throw new REZKException(sqle.getMessage());
		}
	}

	@Override
	public void delete(REFITByteString path) throws REZKException {
		try {
			PreparedStatement statement = operations.DELETE;
			statement.setBytes(1, path.getBytes());
			executePreparedCommand(statement);
		} catch (SQLException sqle) {
			sqle.printStackTrace();
			throw new REZKException(sqle.getMessage());
		}
	}

	@Override
	public Set<REFITByteString> getChildren(REFITByteString path) throws REZKException {
		try {
			// Execute query
			PreparedStatement statement = operations.GET_CHILDREN;
			statement.setBytes(1, path.getBytes());
			executePreparedCommand(statement);

			// Handle result
			Set<REFITByteString> children = new HashSet<>();
			ResultSet result = statement.getResultSet();
			while (result.next()) {
				REFITByteString childPath = new REFITByteString(result.getBytes(1));
				children.add(REZKServer.getChildPath(childPath));
			}
			result.close();
			return children;
		} catch (SQLException sqle) {
			sqle.printStackTrace();
			throw new REZKException(sqle.getMessage());
		}
	}

	@Override
	public byte[] getData(REFITByteString path, REZKStat stat) throws REZKException {
		try {
			// Execute query
			PreparedStatement statement = operations.GET_DATA;
			statement.setBytes(1, path.getBytes());
			executePreparedCommand(statement);

			// Handle result
			ResultSet result = statement.getResultSet();
			if (!result.next()) return null;
			stat.setVersion(result.getInt(2));
			stat.setDataLength(result.getInt(3));
			stat.setNrOfChildren(result.getInt(4));
			stat.setLastAccess(result.getLong(5));
			byte[] bytes = result.getBytes(1);
			result.close();
			return bytes;
		} catch (SQLException sqle) {
			sqle.printStackTrace();
			throw new REZKException(sqle.getMessage());
		}
	}

	@Override
	public void setData(REFITByteString path, byte[] data, REZKStat stat) throws REZKException {
		try {
			PreparedStatement statement = operations.SET_DATA;
			statement.setBytes(1, data);
			statement.setInt(2, stat.version);
			statement.setInt(3, stat.dataLength);
			statement.setInt(4, stat.nrOfChildren);
			statement.setLong(5, stat.lastAccess);
			statement.setBytes(6, path.getBytes());
			executePreparedCommand(statement);
		} catch (SQLException sqle) {
			sqle.printStackTrace();
			throw new REZKException(sqle.getMessage());
		}
	}


	// #########
	// # STATS #
	// #########

	@Override
	public REZKStat getStat(REFITByteString path) throws REZKException {
		try {
			// Execute query
			PreparedStatement statement = operations.GET_STAT;
			statement.setBytes(1, path.getBytes());
			executePreparedCommand(statement);

			// Handle result
			REZKStat stat = new REZKStat();
			ResultSet result = statement.getResultSet();
			if (!result.next()) return null;
			stat.setVersion(result.getInt(1));
			stat.setDataLength(result.getInt(2));
			stat.setNrOfChildren(result.getInt(3));
			stat.setLastAccess(result.getLong(4));
			result.close();
			return stat;
		} catch (SQLException sqle) {
			sqle.printStackTrace();
			throw new REZKException(sqle.getMessage());
		}
	}

	@Override
	public void updateStat(REFITByteString path, REZKStat stat) throws REZKException {
		try {
			PreparedStatement statement = operations.SET_STAT;
			statement.setInt(1, stat.version);
			statement.setInt(2, stat.dataLength);
			statement.setInt(3, stat.nrOfChildren);
			statement.setLong(4, stat.lastAccess);
			statement.setBytes(5, path.getBytes());
			executePreparedCommand(statement);
		} catch (SQLException sqle) {
			sqle.printStackTrace();
			throw new REZKException(sqle.getMessage());
		}
	}


	// ###########
	// # UPDATES #
	// ###########

	@Override
	public void applyUpdates(Collection<REZKUpdate> updates) throws REZKException {
		try {
			// Preparations
			boolean executeCreates = false;
			boolean executeDeletes = false;
			boolean executeSetDatas = false;
			boolean executeSetStats = false;

			// Prepare statements
			for (REZKUpdate update : updates) {
				// Create statement
				PreparedStatement statement;
				switch (update.type) {
				case CREATE:
					REFITByteString parentPath = REZKServer.getParentPath(update.path);
					statement = operations.CREATE;
					statement.setBytes(1, update.path.getBytes());
					statement.setBytes(2, parentPath.getBytes());
					statement.setBytes(3, update.dataArray());
					assert update.stat != null;
					statement.setInt(4, update.stat.version);
					statement.setInt(5, update.stat.dataLength);
					statement.setInt(6, update.stat.nrOfChildren);
					statement.setLong(7, update.stat.lastAccess);
					executeCreates = true;
					break;
				case DELETE:
					statement = operations.DELETE;
					statement.setBytes(1, update.path.getBytes());
					executeDeletes = true;
					break;
				case SET_DATA:
					statement = operations.SET_DATA;
					statement.setBytes(1, update.dataArray());
					assert update.stat != null;
					statement.setInt(2, update.stat.version);
					statement.setInt(3, update.stat.dataLength);
					statement.setInt(4, update.stat.nrOfChildren);
					statement.setLong(5, update.stat.lastAccess);
					executeSetDatas = true;
					break;
				case SET_STAT:
					statement = operations.SET_STAT;
					assert update.stat != null;
					statement.setInt(1, update.stat.version);
					statement.setInt(2, update.stat.dataLength);
					statement.setInt(3, update.stat.nrOfChildren);
					statement.setLong(4, update.stat.lastAccess);
					executeSetStats = true;
					break;
				default:
					throw new REZKException("Unexpected update type: " + update.type);
				}

				// Add statement to batch
				statement.addBatch();
			}

			// Execute batches
			if (executeCreates) operations.CREATE.executeBatch();
			if (executeDeletes) operations.DELETE.executeBatch();
			if (executeSetDatas) operations.SET_DATA.executeBatch();
			if (executeSetStats) operations.SET_STAT.executeBatch();
		} catch (SQLException sqle) {
			sqle.printStackTrace();
			throw new REZKException(sqle.getMessage());
		}
	}


	// ###############
	// # CHECKPOINTS #
	// ###############

	@Override
	public void createCheckpoint(Queue<ByteBuffer> queue) throws REZKException {
		try (ResultSet result = getFullSQLCheckpoint()) {
			while (result.next()) {
				// Store path and parent path
				final REFITByteString path = new REFITByteString(result.getBytes(1));
				final byte[] data = result.getBytes(2);

				// Store stat
				REZKStat stat = new REZKStat();
				stat.setVersion(result.getInt(3));
				stat.setDataLength(result.getInt(4));
				stat.setNrOfChildren(result.getInt(5));
				stat.setLastAccess(result.getLong(6));

				int size = REFITSerialization.getByteStringSize(path, REFITSerializationSize.TINY)
						+ stat.getSerializedSize()
						+ REFITSerialization.getByteArraySize(data, REFITSerializationSize.SMALL);

				ByteBuffer buffer = ByteBuffer.allocate(size);
				REFITSerialization.putByteString(buffer, path, REFITSerializationSize.TINY);
				stat.serialize(buffer);
				REFITSerialization.putByteArray(buffer, data, REFITSerializationSize.SMALL);
				buffer.flip();
				queue.add(buffer);
			}
		} catch (SQLException sqle) {
			sqle.printStackTrace();
			throw new REZKException(sqle.getMessage());
		}
	}

	public static int DELAY_STEP = 500;
	public static int DELAY_TIME = 2;

	@Override
	public void createFuzzyCheckpoint(REZKFuzzyCheckpointContainer checkpoint) throws REZKException {
		// _rowid_ or one of its aliases is used by sqlite to identify individual rows
		// its value may change in SQLite only if the database is vacuumed but that case
		// can be handled / controlled by the application

		int ctr = 0;
		long rowid = -1;
		boolean progress;
		try {
			do {
				progress = false;
				operations.CHECKPOINT_FUZZY.setLong(1, rowid);
				try (ResultSet result = getSQLCheckpoint(operations.CHECKPOINT_FUZZY)) {
					while (result.next()) {
						progress = true;
						// Store path and parent path
						REFITByteString path = new REFITByteString(result.getBytes(1));
						REZKUpdate update = new REZKUpdate(REZKUpdate.REZKUpdateType.CREATE, path);
						update.setData(result.getBytes(2));

						update.stat.setVersion(result.getInt(3));
						update.stat.setDataLength(result.getInt(4));
						update.stat.setNrOfChildren(result.getInt(5));
						update.stat.setLastAccess(result.getLong(6));

						update.serializeMessage();

						checkpoint.addData(update);

						rowid = result.getLong(7);

						ctr++;
						if ((ctr % DELAY_STEP) == 0) {
							try {
								Thread.sleep(DELAY_TIME);
							} catch (InterruptedException e) {
								e.printStackTrace();
							}
						}
					}
				}
			} while (progress);
		} catch (SQLException sqle) {
			sqle.printStackTrace();
			throw new REZKException(sqle.getMessage());
		}
	}

	@Override
	public void createFuzzyDiffCheckpoint(Set<?> changelist, Map<Object, REFITCheckpointObject> changedObjects) throws REZKException {
		int ctr = 0;
		try {
			for (Object o : changelist) {
				REFITByteString path = ((REFITByteString) o);
				// Execute query
				PreparedStatement statement = operations.GET_DATA_FUZZY;
				statement.setBytes(1, path.getBytes());
				executePreparedCommand(statement);

				// Handle result
				ResultSet result = statement.getResultSet();
				if (!result.next()) {
					// Nothing to do for missing objects
					continue;
				}

				REZKUpdate update = new REZKUpdate(REZKUpdate.REZKUpdateType.CREATE, path);
				update.setData(result.getBytes(1));

				update.stat.setVersion(result.getInt(2));
				update.stat.setDataLength(result.getInt(3));
				update.stat.setNrOfChildren(result.getInt(4));
				update.stat.setLastAccess(result.getLong(5));
				result.close();

				update.serializeMessage();

				changedObjects.put(path, update);

				ctr++;
				if ((ctr % DELAY_STEP) == 0) {
					try {
						Thread.sleep(DELAY_TIME);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
			}
		} catch (SQLException sqle) {
			sqle.printStackTrace();
			throw new REZKException(sqle.getMessage());
		}
	}

	@Override
	public void createFuzzyUpdateCheckpoint(Map<?, REFITCheckpointObject> changelist, Collection<REFITCheckpointObject> changedObjects) throws REZKException {
		int ctr = 0;
		try {
			for (Map.Entry<?, REFITCheckpointObject> e : changelist.entrySet()) {
				REFITByteString path = ((REFITByteString) e.getKey());
				REZKUpdate update;

				if (((REZKUpdate) e.getValue()).type == REZKUpdate.REZKUpdateType.STAT_STUB) {
					// Execute query
					PreparedStatement statement = operations.GET_STAT_FUZZY;
					statement.setBytes(1, path.getBytes());
					executePreparedCommand(statement);

					// Handle result
					ResultSet result = statement.getResultSet();
					if (!result.next()) {
						// Nothing to do for missing objects
						continue;
					}

					update = new REZKUpdate(REZKUpdate.REZKUpdateType.SET_STAT, path);
					update.stat.setVersion(result.getInt(1));
					update.stat.setDataLength(result.getInt(2));
					update.stat.setNrOfChildren(result.getInt(3));
					update.stat.setLastAccess(result.getLong(4));
					result.close();
				} else {
					// Execute query
					PreparedStatement statement = operations.GET_DATA_FUZZY;
					statement.setBytes(1, path.getBytes());
					executePreparedCommand(statement);

					// Handle result
					ResultSet result = statement.getResultSet();
					if (!result.next()) {
						// Nothing to do for missing objects
						continue;
					}

					update = new REZKUpdate(REZKUpdate.REZKUpdateType.CREATE, path);
					update.setData(result.getBytes(1));

					update.stat.setVersion(result.getInt(2));
					update.stat.setDataLength(result.getInt(3));
					update.stat.setNrOfChildren(result.getInt(4));
					update.stat.setLastAccess(result.getLong(5));
					result.close();
				}

				update.serializeMessage();
				changedObjects.add(update);
				ctr++;
				if ((ctr % DELAY_STEP) == 0) {
					try {
						Thread.sleep(DELAY_TIME);
					} catch (InterruptedException f) {
						f.printStackTrace();
					}
				}
			}
		} catch (SQLException sqle) {
			sqle.printStackTrace();
			throw new REZKException(sqle.getMessage());
		}
	}

	private ResultSet getFullSQLCheckpoint() throws REZKException, SQLException {
		return getSQLCheckpoint(operations.CHECKPOINT);
	}

	private ResultSet getSQLCheckpoint(PreparedStatement statement) throws REZKException, SQLException {
		// Execute query
		executePreparedCommand(statement);

		// Handle result
		return statement.getResultSet();
	}

	@Override
	public Map<REFITByteString, REZKStat> applyCheckpoint(final ByteBuffer buffer) throws REZKException {
		// Initialize table
		Map<REFITByteString, REZKStat> stats = new HashMap<>();
		initTable();
		if (!buffer.hasRemaining()) return stats;

		try {
			// Prepare statements
			PreparedStatement statement = operations.CREATE;
			while (buffer.hasRemaining()) {
				// Get node information
				REFITByteString path = REFITSerialization.getByteString(buffer, REFITSerializationSize.TINY);
				REFITByteString parentPath = REZKServer.getParentPath(path);
				REZKStat stat = new REZKStat(buffer);
				byte[] data = REFITSerialization.getByteArray(buffer, REFITSerializationSize.SMALL);

				// Prepare statement
				statement.setBytes(1, path.getBytes());
				statement.setBytes(2, parentPath.getBytes());
				statement.setBytes(3, data);
				statement.setInt(4, stat.version);
				statement.setInt(5, stat.dataLength);
				statement.setInt(6, stat.nrOfChildren);
				statement.setLong(7, stat.lastAccess);

				// Add statement to batch
				statement.addBatch();

				// Add stat to stats
				stats.put(path, stat);
			}

			// Execute batch
			statement.executeBatch();
			statement.close();
			return stats;
		} catch (SQLException sqle) {
			sqle.printStackTrace();
			throw new REZKException(sqle.getMessage());
		}
	}


	// ###################
	// # DATABASE ACCESS #
	// ###################

	public void executeBatchCommand(String[] commands) throws REZKException {
		try {
			Statement statement = database.createStatement();
			for (String command : commands) statement.addBatch(command);
			statement.executeBatch();
			statement.close();
		} catch (SQLException sqle) {
			sqle.printStackTrace();
			throw new REZKException(sqle.getMessage());
		}
	}

	private void executePreparedCommand(PreparedStatement statement) throws REZKException {
		try {
			statement.execute();
		} catch (SQLException sqle) {
			sqle.printStackTrace();
			throw new REZKException(sqle.getMessage());
		}
	}

}
