package refit.client;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.concurrent.LinkedBlockingQueue;

import refit.client.policy.REFITClientPolicy;
import refit.communication.REFITNetworkEndpoint;
import refit.communication.REFITNetworkEndpointWorker;
import refit.communication.REFITNetworkInboundRouter;
import refit.communication.REFITNodeAddress;
import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITMessage;
import refit.message.REFITMessageType;
import refit.message.REFITProbeData;
import refit.replica.REFITStage;
import refit.scheduler.REFITSchedulerGroup;
import refit.scheduler.REFITSchedulerTaskType;
import refit.util.REFITIntervalStatistics;

public abstract class REFITClientLibrary extends REFITStage implements REFITNetworkInboundRouter {

	public enum RequestMode {
		READ,
		READ_WEAK,
		WRITE,
		WRITE_ACK,
		PROBE,
		SYSTEM,
		;

		public RequestMode toStrongMode() {
			// used to upgrade READ_WEAK to READ when REFITConfig.ONLY_CONSISTENT is set
			if (this == READ_WEAK) {
				return READ;
			}
			return this;
		}

		public boolean isTotalOrder() {
			// return whether the request should yield linearizable results
			switch (this) {
			case READ:
			case WRITE:
			case WRITE_ACK:
			case SYSTEM:
				return true;
			case READ_WEAK:
			case PROBE:
				return false;
			default:
				throw new InternalError();
			}
		}

		public boolean isReadOnly() {
			// return whether the request will modify the application state or not
			switch (this) {
			case READ:
			case READ_WEAK:
				return true;
			case WRITE:
			case WRITE_ACK:
			case PROBE:
			case SYSTEM:
				return false;
			default:
				throw new InternalError();
			}
		}
	}

	public enum ResultState {
		NONE, ACK, STABLE
	}

	private final REFITSchedulerGroup group;
	private final short clientID;
	protected final REFITNetworkEndpoint endpoint;
	protected final REFITIntervalStatistics statistics;
	protected final REFITClientPolicy policy;
	protected REFITClientTask client;

	protected final LinkedBlockingQueue<Object> sidechannel;

	public REFITClientLibrary(short clientID, REFITSchedulerGroup group, REFITIntervalStatistics statistics,
	                          REFITClientPolicy policy, boolean isFirstClient, LinkedBlockingQueue<Object> sidechannel) {
		super(REFITSchedulerTaskType.CLIENT_LIBRARY, group);
		this.group = group;
		this.clientID = clientID;
		this.endpoint = createNetwork(isFirstClient);
		if (endpoint != null) {
			endpoint.initCompletedSender.addListener(this);
		}
		this.statistics = statistics;
		this.policy = policy;
		this.sidechannel = sidechannel;
	}

	protected REFITNetworkEndpoint createNetwork(boolean isFirstClient) {
		HashSet<Short> initialNodes = new HashSet<>();
		boolean[] connects = getInitialConnects();
		ArrayList<REFITNodeAddress> initialConnects = new ArrayList<>();
		for (short i = 0; i < REFITConfig.TOTAL_NR_OF_REPLICAS; i++) {
			if (connects != null && !connects[i]) continue;
			initialNodes.add(i);
			initialConnects.add(new REFITNodeAddress(i, REFITConfig.ADDRESSES[i]));
		}

		return new REFITNetworkEndpoint(this,
				REFITConfig.TOTAL_NR_OF_REPLICAS,
				(i) -> new REFITNetworkEndpointWorker.ConnectionParameters(REFITSchedulerTaskType.CLIENT_ENDPOINT_WORKER,
						REFITConfig.CLIENT_RECEIVE_BUFFER_SIZE, REFITConfig.CLIENT_SEND_BUFFER_SIZE), REFITConfig.CLIENT_INDIVIDUAL_NETWORK_CONNECTORS,
				initialNodes, false, initialConnects, false, !isFirstClient);
	}

	protected boolean[] getInitialConnects() {
		return null;
	}

	public void setClient(REFITClientTask client) {
		this.client = client;
	}

	@Override
	public REFITSchedulerGroup getGroup() {
		return group;
	}

	@Override
	public short getNodeID() {
		return clientID;
	}

	public REFITIntervalStatistics getStatistics() {
		return statistics;
	}

	public abstract long getProgress();

	public REFITClientPolicy getPolicy() {
		return policy;
	}

	protected void completeDelayedConnect() {
		// should be called when receiving a REFITInitialConnectionsCompleted instruction for an other node
		endpoint.completeInitialConnects();
	}

	public void addInitialCompleteListener(REFITStage listener) {
		endpoint.initCompletedSender.addListener(listener);
	}

	// ###################
	// # DECODE MESSAGES #
	// ###################

	@Override
	public void receiveMessage(ByteBuffer message, short nodeID) {
		// Create reply
		REFITMessage reply = REFITMessage.createMessage(message);
		if (reply.type != REFITMessageType.REPLY && reply.type != REFITMessageType.ACK_REPLY) {
			REFITLogger.logWarning(this, "Dropping message of unexpected type from replica " + nodeID + " " + reply.type);
			return;
		}

		insertMessage(reply);
	}

	protected void sendToReplica(REFITMessage message, short replicaID) {
		endpoint.unicastMessage(message, replicaID);
	}

	protected void sendToReplicas(REFITMessage message, boolean[] replicas) {
		for (short i = 0; i < REFITConfig.TOTAL_NR_OF_REPLICAS; i++) {
			if (replicas[i]) {
				endpoint.unicastMessage(message, i);
			}
		}
	}

	protected void sendToAllReplicas(REFITMessage message) {
		for (short i = 0; i < REFITConfig.TOTAL_NR_OF_REPLICAS; i++) {
			endpoint.unicastMessage(message, i);
		}
	}

	public LinkedBlockingQueue<Object> getSidechannel() {
		return sidechannel;
	}


	// ###################
	// # ABSTRACT METHOD #
	// ###################

	public ByteBuffer invoke(ByteBuffer request, RequestMode mode) throws Exception {
		return invoke(request, -1, mode);
	}

	public abstract ByteBuffer invoke(ByteBuffer request, long executeAfter, RequestMode mode) throws Exception;

	public void invokeAsync(ByteBuffer request, RequestMode mode) {
		invokeAsync(request, -1, mode);
	}

	public abstract void invokeAsync(ByteBuffer request, long executeAfter, RequestMode mode);

	public abstract REFITProbeData invokeProbe() throws Exception;

	public abstract boolean isResultStable();

	public abstract boolean isReadWriteConflict();

	public abstract ByteBuffer getResult();

	public abstract long getAssignedSeqNr();
}
