package refit.client;

import java.nio.ByteBuffer;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;

import refit.client.policy.REFITClientPolicy;
import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.*;
import refit.message.REFITInstruction.REFITInitialConnectionsCompleted;
import refit.message.REFITInstruction.REFITInvocationInstruction;
import refit.replica.REFITReplicaGroups;
import refit.replica.stewardorder.REFITStewardGroups;
import refit.scheduler.REFITSchedulerGroup;
import refit.util.REFITIntervalStatistics;
import refit.util.REFITTime;


public class REFITLibSteward extends REFITClientLibrary {

	private short replyReplicaID;
	private final REFITMessageAuthentication messageAuthentication;
	private final REFITReplyCertificate certificate;
	private final REFITStewardReplyCertificate consReadCertificate;
	private boolean consRead;
	private final Semaphore stableReplyLock;
	private boolean readWriteConflict;

	private boolean isConnected;
	private REFITInvocationInstruction delayedInstruction;

	private final short groupID;
	private long nextSeqNr;
	private long nextReadSeqNr;
	private REFITRequest request;


	public REFITLibSteward(short clientID, REFITSchedulerGroup group, REFITIntervalStatistics statistics,
	                       REFITClientPolicy policy, boolean isFirstClient, LinkedBlockingQueue<Object> sidechannel) {
		super(clientID, group, statistics, policy, isFirstClient, sidechannel);

		short replyOffset = (short) (clientID % (REFITConfig.TOTAL_NR_OF_REPLICAS - (REFITConfig.USE_PASSIVE_REPLICATION ? REFITConfig.FAULTS_TO_TOLERATE : 0)));
		for (short i = 0; i < REFITConfig.TOTAL_NR_OF_REPLICAS; i++) {
			boolean isPassive = false;
			for (short passiveReplica : REFITConfig.PASSIVE_REPLICAS) {
				if (i == passiveReplica) {
					isPassive = true;
					break;
				}
			}
			if (isPassive) continue;

			if (replyOffset > 0) {
				replyOffset--;
			} else {
				this.replyReplicaID = i;
				break;
			}
		}

		this.replyReplicaID = replyOffset;
		this.messageAuthentication = new REFITMessageAuthentication(clientID);
		this.groupID = REFITStewardGroups.stewardClientEGID(clientID);
		REFITLogger.logEvent(this, "Using execution group " + groupID);

		this.certificate = new REFITReplyCertificate(messageAuthentication, statistics, REFITStewardGroups.stewardOrderGroupByID(groupID));
		this.consReadCertificate = new REFITStewardReplyCertificate(messageAuthentication, statistics);
		this.stableReplyLock = new Semaphore(0);

		this.nextSeqNr = 0L;
		this.invocationTimeStamp = -1L;
		this.isConnected = false;
	}

	@Override
	public long getProgress() {
		return nextSeqNr;
	}

	@Override
	public String toString() {
		return String.format("LIBSTW[%03d]", getNodeID());
	}


	// ######################
	// # CLIENT INTERACTION #
	// ######################

	private long invocationTimeStamp;

	@Override
	public ByteBuffer invoke(ByteBuffer request, long executeAfter, RequestMode mode) throws InterruptedException {
		// Insert invocation request
		invokeAsync(request, REFITConfig.ONLY_CONSISTENT ? -1 : executeAfter, REFITConfig.ONLY_CONSISTENT ? mode.toStrongMode() : mode);

		// Wait until reply is stable
		while (true) {
			boolean success = stableReplyLock.tryAcquire(REFITConfig.CLIENT_REQUEST_TIMEOUT, TimeUnit.MILLISECONDS);
			if (success) break;

			// Retry
			REFITInstruction invocation = new REFITInvocationInstruction();
			insertMessage(invocation);
		}

		// Return result
		return (consRead) ? consReadCertificate.result.getPayload() : ((REFITReply) certificate.result).getPayload();
	}

	@Override
	public void invokeAsync(ByteBuffer request, long executeAfter, RequestMode mode) {
		REFITInstruction invocation = new REFITInvocationInstruction(request, executeAfter, mode);
		insertMessage(invocation);
		invocationTimeStamp = REFITTime.currentTimeMillis.getAsLong();
	}

	@Override
	public REFITProbeData invokeProbe() throws InterruptedException {
		throw new UnsupportedOperationException("No reason to probe the order stage");
	}

	@Override
	public boolean isResultStable() {
		// Check whether result is stable
		boolean resultStable = (consRead) ? consReadCertificate.isStable() : certificate.isStable();
		if (resultStable) return true;

		// Check for retry
		long now = REFITTime.currentTimeMillis.getAsLong();
		if ((now - invocationTimeStamp) > REFITConfig.CLIENT_REQUEST_TIMEOUT) {
			// Retry
			REFITInstruction invocation = new REFITInvocationInstruction();
			insertMessage(invocation);
		}
		return false;
	}

	@Override
	public ByteBuffer getResult() {
		stableReplyLock.acquireUninterruptibly();
		return (consRead) ? consReadCertificate.result.getPayload() : ((REFITReply) certificate.result).getPayload();
	}

	@Override
	public boolean isReadWriteConflict() {
		return readWriteConflict;
	}

	@Override
	public long getAssignedSeqNr() {
		return certificate.result.sequenceNumber();
	}

	// #######################
	// # REPLICA INTERACTION #
	// #######################

	@Override
	protected void handleMessage(REFITMessage message) {
		switch (message.type) {
		case REPLY:
			handleReply((REFITReply) message);
			break;
		case INSTRUCTION:
			handleInstruction((REFITInstruction) message);
			break;
		default:
			REFITLogger.logError(this, "drop message of unexpected type " + message.type);
		}
	}

	private void handleReply(REFITReply reply) {
		if (consRead) {
			// Ignore replies that arrive after the result has become stable
			if (request == null || consReadCertificate.isStable()) return;

			// Add reply to certificate
			consReadCertificate.add(reply);
			if (!consReadCertificate.isStable()) return;
		} else {
			// Ignore replies that arrive after the result has become stable
			if (request == null || certificate.isStable()) return;

			// Add reply to certificate
			certificate.add(reply);
			if (!certificate.isStable()) return;
		}

		// Release lock
		stableReplyLock.release();
		if (client != null) client.progress();
	}


	private void handleInstruction(REFITInstruction instruction) {
		switch (instruction.instructionType) {
		case INVOCATION:
			handeInvocation(((REFITInvocationInstruction) instruction));
			break;
		case INITIAL_CONNECTIONS_COMPLETED:
			REFITInitialConnectionsCompleted init = (REFITInitialConnectionsCompleted) instruction;
			if (init.nodeID == getNodeID()) {
				handleInitCompleted();
			} else {
				completeDelayedConnect();
			}
			break;
		default:
			REFITLogger.logError(this, "drop instruction of unexpected type " + instruction.instructionType);
		}
	}

	private void handleInitCompleted() {
		isConnected = true;
		if (delayedInstruction != null) {
			insertMessage(delayedInstruction);
			delayedInstruction = null;
		}
	}

	private void handeInvocation(REFITInvocationInstruction instruction) {
		// block instructions until we are connected
		if (!isConnected) {
			if (delayedInstruction == null) {
				delayedInstruction = instruction;
			}
			return;
		}

		if (instruction.request != null) {
			// Prepare request
			REFITUniqueID uid;
			boolean isTotalOrder = instruction.mode.isTotalOrder();
			if (isTotalOrder) {
				uid = new REFITUniqueID(getNodeID(), nextSeqNr++);
			} else {
				uid = new REFITUniqueID(getNodeID(), nextReadSeqNr++);
			}

			request = new REFITRequest(uid, getNodeID(), groupID, instruction.executeAfter, instruction.request, replyReplicaID, instruction.mode);
			consRead = instruction.mode.equals(RequestMode.READ);
			if (consRead) {
				consReadCertificate.init(uid);
			} else {
				certificate.init(uid, isTotalOrder, request.mode.equals(RequestMode.WRITE_ACK));
			}
			readWriteConflict = false;

			request.serializeMessage();
			if (isTotalOrder) {
				// use signatures to prevent executors from damaging the request's authenticator
				messageAuthentication.appendSignature(request);
			} else {
				// using MACs is sufficient as every executor processes the command on its own
				// add MAC to wrapper to unify code paths
				messageAuthentication.appendNullMAC(request);
			}

			if (isTotalOrder && !consRead) {
				short leaderID = (short) REFITStewardGroups.stewardOrderGroupStartId(groupID);

				// wrap
				REFITRequestWrapper wrapper = new REFITRequestWrapper(uid, getNodeID(), request);
				wrapper.serializeMessage();
				messageAuthentication.appendUnicastMAC(wrapper, leaderID);

				sendToReplica(wrapper, leaderID);
			} else {
				boolean[] executionReplicas = (consRead) ? REFITReplicaGroups.getAllReplicas() : REFITStewardGroups.stewardOrderGroupByID(groupID);
				for (short i = 0; i < executionReplicas.length; i++) {
					if (!executionReplicas[i]) continue;
					// wrap
					REFITRequestWrapper wrapper = new REFITRequestWrapper(uid, getNodeID(), request);
					wrapper.serializeMessage();
					messageAuthentication.appendUnicastMAC(wrapper, i);

					sendToReplica(wrapper, i);
				}
			}
		} else if (request != null) {
			// Retry; just resend normal requests
			if (consRead) {
				REFITLogger.logWarning(this, String.format("retry %s (%d site replies received)", request, consReadCertificate.mergedReplies.getVoteCount()));
			} else {
				REFITLogger.logWarning(this, String.format("retry %s (%d replies received)", request, certificate.replies.getVoteCount()));
			}
			if (!request.mode.isTotalOrder()) {
				// Send read-only as normal request instead, if we couldn't get a reply in time - doesn't need to be marked panic the first time
				REFITUniqueID uid = new REFITUniqueID(getNodeID(), nextSeqNr++);
				REFITLogger.logWarning(this, "Failed receiving reply to read-only request " + request + ", resending as read-write");
				// Reinitialize certificates
				certificate.init(uid, true, false);
				consRead = false;

				request = new REFITRequest(uid, getNodeID(), groupID, request.executeAfter, request.getPayload(), (short) -1, RequestMode.WRITE);
				readWriteConflict = true;

				// use signatures to prevent executors from damaging the request's authenticator
				request.serializeMessage();
				messageAuthentication.appendSignature(request);
			}

			boolean[] executionReplicas = REFITStewardGroups.stewardOrderGroupByID(0);
			for (short i = 0; i < executionReplicas.length; i++) {
				if (!executionReplicas[i]) continue;
				// wrap
				REFITRequestWrapper wrapper = new REFITRequestWrapper(request.uid, getNodeID(), request);
				wrapper.serializeMessage();
				messageAuthentication.appendUnicastMAC(wrapper, i);

				sendToReplica(wrapper, i);
			}
		}
		progress();
	}

}
