package refit.client;

import java.nio.ByteBuffer;
import java.util.List;

import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITMessageAuthentication;
import refit.message.REFITReply;
import refit.message.REFITUniqueID;
import refit.replica.REFITReplicaGroups;
import refit.replica.stewardorder.REFITStewardGroups;
import refit.util.REFITBallotBox;
import refit.util.REFITIntervalStatistics;


public class REFITStewardReplyCertificate {

	private static final int STABILITY_THRESHOLD = REFITConfig.FAULTS_TO_TOLERATE + 1;
	private static final int SITE_STABILITY_THRESHOLD = REFITStewardGroups.stewardOrderGroupCount() / 2 + 1;

	public final REFITBallotBox<Short, ByteBuffer, REFITReply>[] replies;
	public final REFITBallotBox<Short, ByteBuffer, REFITReply> mergedReplies;

	private final REFITMessageAuthentication messageAuthentication;
	private REFITUniqueID uid;
	public REFITReply result;
	private final boolean[] legitimateSenders;

	private final REFITIntervalStatistics statistics;

	public REFITStewardReplyCertificate(REFITMessageAuthentication messageAuthentication, REFITIntervalStatistics statistics) {
		//noinspection unchecked
		this.replies = (REFITBallotBox<Short, ByteBuffer, REFITReply>[]) new REFITBallotBox[REFITStewardGroups.stewardOrderGroupCount()];
		for (int i = 0; i < replies.length; i++) {
			this.replies[i] = new REFITBallotBox<>(STABILITY_THRESHOLD);
		}
		this.mergedReplies = new REFITBallotBox<>(SITE_STABILITY_THRESHOLD);
		this.messageAuthentication = messageAuthentication;
		this.statistics = statistics;
		this.legitimateSenders = REFITReplicaGroups.getAllReplicas();
	}


	public void init(REFITUniqueID uid) {
		for (int i = 0; i < replies.length; i++) {
			replies[i].clear();
		}
		mergedReplies.clear();
		this.uid = uid;
		result = null;
	}

	public boolean add(REFITReply reply) {
		// Check reply
		if (reply == null) return false;
		if (!uid.equals(reply.uid) || !reply.mode.equals(REFITClientLibrary.RequestMode.READ)) return false;
		if (!messageAuthentication.verifyUnicastMAC(reply, legitimateSenders)) return false;

		// Prepare vote
		ByteBuffer vote = ByteBuffer.wrap(reply.getPayloadHash());

		// Check whether the reply is a late full reply
		ByteBuffer decision = mergedReplies.getDecision();
		if (reply.isFullReply && vote.equals(decision)) {
			result = reply;
			return true;
		}

		int group = REFITStewardGroups.stewardOrderGroupID(reply.from);

		// Cast vote
		boolean success = replies[group].add(reply.from, vote, reply);
		if (success) checkResult();
		return success;
	}

	public boolean isStable() {
		// FIXME merge viewID
		return (result != null);
	}

	private void checkResult() {
		if (isStable()) return;
		for (int i = 0; i < replies.length; i++) {
			if (mergedReplies.hasVoted((short) i)) continue;

			if (REFITConfig.ENABLE_DEBUG_CHECKS && replies[i].votesDiffer())
				REFITLogger.logWarning("[REPLY]", "received different results for request " + uid);
			List<REFITReply> correctReplies = replies[i].getDecidingBallots();
			if (correctReplies == null) continue;

			REFITReply rep = correctReplies.get(0);
			for (REFITReply reply : correctReplies) {
				if (!reply.isFullReply) continue;
				rep = reply;
				break;
			}

			ByteBuffer vote = ByteBuffer.wrap(rep.getPayloadHash());
			mergedReplies.add((short) i, vote, rep);
		}

		// FIXME: handling of hash replys is partially broken
		if (REFITConfig.ENABLE_DEBUG_CHECKS) if (mergedReplies.votesDiffer())
			REFITLogger.logWarning("[REPLY]", "received different site results for request " + uid);
		List<REFITReply> correctReplies = mergedReplies.getDecidingBallots();
		if (correctReplies == null) return;
		for (REFITReply reply : correctReplies) {
			if (!reply.isFullReply) continue;
			result = reply;
			break;
		}
		if (REFITConfig.CLIENT_REPLY_STATISTICS) updateStats(correctReplies);
	}

	private void updateStats(List<REFITReply> correctReplies) {
		for (REFITReply reply : correctReplies) {
			statistics.trackReply(reply.from);
		}
	}

}
