package refit.client.policy;

import java.util.Arrays;
import java.util.concurrent.Semaphore;

import refit.client.REFITClientTask;
import refit.config.REFITConfig;
import refit.util.REFITAssert;
import refit.util.REFITTime;

public class REFITOpenLoopPolicy implements REFITClientPolicy {

	private final long durationInMs;
	private final int clientIDOffset;
	private final short clientRate;
	private final Semaphore clientThrottle;
	private final boolean[] firstRun;
	private final int clientProcessID;

	public REFITOpenLoopPolicy(long durationInMs, int clientProcessID, int nrOfClients, int clientIDOffset) {
		this.durationInMs = durationInMs;
		this.clientIDOffset = clientIDOffset;
		this.clientProcessID = clientProcessID;

		clientRate = REFITConfig.POLICY_OPEN_LOOP_RATE[clientProcessID];
		REFITAssert.assertTrue(clientRate > 0, "Invalid client rate for open loop policy");

		clientThrottle = new Semaphore(0);
		firstRun = new boolean[nrOfClients];
		Arrays.fill(firstRun, true);
	}

	@Override
	public void waitForMyTurn(short nodeID) throws InterruptedException {
		int offsetID = nodeID - clientIDOffset;
		if (firstRun[offsetID]) {

			if (clientProcessID == REFITConfig.GROUP_LATE_START_GID) {
				Thread.sleep(80 * 1000);
			}

			// don't wait on first run
			firstRun[offsetID] = false;
			return;
		}

		clientThrottle.acquire();
	}

	@Override
	public void execute() throws InterruptedException {
		long expectedEnd = REFITTime.currentTimeMillis.getAsLong() + durationInMs;

		double tokenPerMS = clientRate / 1000.;
		// allow at most half second bursts
		int tokenLimit = (int) Math.ceil(tokenPerMS * 500);

		double tokenRemainder = 0;
		long lastTime = REFITTime.currentTimeMillis.getAsLong();
		while (true) {
			long currentTime = REFITTime.currentTimeMillis.getAsLong();
			if (currentTime >= expectedEnd) break;

			double newTokens = (currentTime - lastTime) * tokenPerMS + tokenRemainder;
			int newFullTokens = (int) Math.floor(newTokens);

			int availablePermits = clientThrottle.availablePermits();
			int permitsToAdd = Math.min(newFullTokens, tokenLimit - availablePermits);
			clientThrottle.release(permitsToAdd);

			lastTime = currentTime;
			tokenRemainder = newTokens - newFullTokens;

			REFITTime.sleep(10);
		}
	}

	@Override
	public boolean shouldWaitForMyTurn(short nodeID, REFITClientTask task) {
		throw new UnsupportedOperationException("client task are not supported by this policy");
	}
}
