package refit.communication;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.nio.channels.SocketChannel;
import java.util.List;
import java.util.Queue;
import java.util.concurrent.ConcurrentLinkedQueue;

import refit.config.REFITConfig;
import refit.replica.REFITOutbox;
import refit.scheduler.REFITSchedulerTask;
import refit.scheduler.REFITSchedulerTaskType;
import refit.util.REFITAssert;
import refit.util.REFITTime;


public class REFITNetworkEndpointWorker extends REFITSchedulerTask {

	private final REFITNetworkInboundRouter replica;
	private final REFITNetworkEndpoint endpoint;
	private final REFITConnection connection;
	private final String instanceName;
	private final short replicaID;
	private final short nodeID;
	private REFITNodeAddress address;
	private final Queue<DelayedBuffer> delayQueue; // testing only!

	private boolean initCalled;

	public REFITNetworkEndpointWorker(REFITNetworkInboundRouter replica, REFITNetworkEndpoint endpoint, short replicaID, short nodeID, ConnectionParameters parameters) {
		super(parameters.taskType, replica.getGroup());
		this.replica = replica;
		this.endpoint = endpoint;
		this.replicaID = replicaID;
		this.nodeID = nodeID;
		this.connection = new REFITConnection(parameters.receiveBufferSize, parameters.sendBufferSize, REFITUnmarshaller.instance, nodeID);
		this.initCalled = false;
		this.instanceName = String.format("NWK%02d", nodeID);
		this.delayQueue = (REFITConfig.FAKE_WAN_DELAYS) ? new ConcurrentLinkedQueue<>() : null;
	}


	@Override
	public String toString() {
		return instanceName;
	}

	public boolean init(SocketChannel socketChannel, REFITNodeAddress address) {
		if (address.nodeID != nodeID) {
			throw new IllegalArgumentException("Got connection for wrong node!");
		}
		this.address = address;
		try {
			connection.init(socketChannel);
		} catch (IOException e) {
			e.printStackTrace();
			return false;
		}
		initCalled = true;
		if (delayQueue != null) {
			delayQueue.clear();
		}

		progress();
		return true;
	}

	public boolean isOpen() {
		return connection.isOpen();
	}

	// ##################
	// # SCHEDULER TASK #
	// ##################

	@Override
	public void selectKey(SelectionKey key) {
		if (REFITConfig.ENABLE_DEBUG_CHECKS && key != connection.selectionKey) {
			throw new IllegalArgumentException("Called for wrong selection key!");
		}
		connection.select();
	}

	@Override
	public void execute() {
		if (initCalled) {
			// register() should only be called from the thread that owns the selector
			// otherwise registering a selectionKey may deadlock
			connection.register(this);
			initCalled = false;
		}

		// Handle sending
		connection.send();

		// Handle receiving
		List<ByteBuffer> messages = connection.receive();
		for (ByteBuffer message : messages) {
			// Forward request
			replica.receiveMessage(message, nodeID);
			event();
		}

		if (!connection.isOpen() && address != null) {
			REFITNodeAddress address = this.address;
			shutdown();
			endpoint.queueReconnect(address);
		}
	}

	// ###########
	// # SENDING #
	// ###########

	public void enqueueMessage(ByteBuffer message) {
		if (REFITConfig.FAKE_WAN_DELAYS) {
			int myLoc = REFITConfig.LOCATION[replicaID];
			int otherLoc = REFITConfig.LOCATION[nodeID];
			long delay = REFITConfig.INTER_LOCATION_DELAYS_MS[myLoc][otherLoc];
			REFITAssert.assertTrue(delay >= 0, "Got negative delay!");
			long sendAfterTime = REFITTime.currentTimeMillis.getAsLong() + delay;
			delayQueue.add(new DelayedBuffer(sendAfterTime, message));

			if (!hasTimeout()) {
				setTimeoutAbsolute(sendAfterTime);
			}
		} else {
			connection.enqueue(message);
		}

		progress();
	}

	public boolean boundedEnqueueMessage(ByteBuffer message, REFITOutbox outbox) {
		boolean success = connection.boundedEnqueue(message, outbox);
		progress();
		return success;
	}

	@Override
	public void timeout() {
		super.timeout();

		long currentTime = REFITTime.currentTimeMillis.getAsLong();
		while (!delayQueue.isEmpty() && currentTime >= delayQueue.peek().sendAfterTime) {
			DelayedBuffer buffer = delayQueue.poll();
			connection.enqueue(buffer.buffer);
		}
		if (!delayQueue.isEmpty()) {
			setTimeoutAbsolute(delayQueue.peek().sendAfterTime);
		}
	}

	public void shutdown() {
		if (connection.isOpen()) {
			connection.closeConnection();
		}
		address = null;
	}

	public static class ConnectionParameters {
		public final REFITSchedulerTaskType taskType;
		public final int receiveBufferSize;
		public final int sendBufferSize;

		public ConnectionParameters(REFITSchedulerTaskType taskType, int receiveBufferSize, int sendBufferSize) {
			this.taskType = taskType;
			this.receiveBufferSize = receiveBufferSize;
			this.sendBufferSize = sendBufferSize;
		}
	}

	private static class DelayedBuffer {
		public final long sendAfterTime;
		public final ByteBuffer buffer;

		public DelayedBuffer(long sendAfterTime, ByteBuffer buffer) {
			this.sendAfterTime = sendAfterTime;
			this.buffer = buffer;
		}
	}
}
