package refit.replica;

import java.lang.reflect.InvocationTargetException;
import java.net.SocketAddress;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.EnumMap;
import java.util.HashSet;
import java.util.concurrent.CountDownLatch;

import refit.application.REFITApplicationServer;
import refit.application.REFITSystemApplicationServer;
import refit.communication.*;
import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITMessage;
import refit.message.REFITMessageAuthentication;
import refit.message.REFITMessageType;
import refit.message.REFITReplyBase;
import refit.scheduler.REFITHangDetector;
import refit.scheduler.REFITScheduler;
import refit.scheduler.REFITSchedulerGroup;
import refit.scheduler.REFITSchedulerTaskType;


public abstract class REFITBoundedBaseReplica implements REFITScheduledRouter, REFITNetworkInboundRouter {
	public final short id;
	public final REFITMessageAuthentication mac;
	public final REFITSchedulerGroup group;

	final REFITNetworkEndpoint network;

	private REFITScheduler[] schedulers;
	private EnumMap<REFITMessageType, REFITObserver<REFITMessage>> inbounds;
	private REFITObserver<REFITMessage> clientsInbound;


	public REFITBoundedBaseReplica(short id, boolean isSingleThreaded, boolean createClientNetwork) {
		// Configure replica
		this.id = id;
		this.mac = new REFITMessageAuthentication(id);
		this.group = new REFITSchedulerGroup(isSingleThreaded);

		this.network = setupNetwork(createClientNetwork);

		this.inbounds = new EnumMap<>(REFITMessageType.class);
		this.clientsInbound = new REFITObserver<>();
	}

	private REFITNetworkEndpoint setupNetwork(boolean createClientNetwork) {
		HashSet<Short> initialNodes = new HashSet<>();
		for (short i = 0; i < REFITConfig.TOTAL_NR_OF_REPLICAS; i++) {
			if (i != id) initialNodes.add(i);
		}
		ArrayList<REFITNodeAddress> initialConnects = new ArrayList<>();
		for (short i = (short) (id + 1); i < REFITConfig.TOTAL_NR_OF_REPLICAS; i++) {
			final SocketAddress addr = (REFITConfig.LOCATION[i] == REFITConfig.LOCATION[id]) ? REFITConfig.INT_ADDRESSES[i] : REFITConfig.ADDRESSES[i];
			initialConnects.add(new REFITNodeAddress(i, addr));
		}

		// Create client and replica networks
		return new REFITNetworkEndpoint(this,
				REFITConfig.TOTAL_NR_OF_REPLICAS + REFITConfig.TOTAL_NR_OF_CLIENTS,
				(i) -> {
					if (i < REFITConfig.TOTAL_NR_OF_REPLICAS) {
						return new REFITNetworkEndpointWorker.ConnectionParameters(REFITSchedulerTaskType.REPLICA_NETWORK_ENDPOINT_WORKER,
								REFITConfig.REPLICA_NETWORK_RECEIVE_BUFFER_SIZE, REFITConfig.REPLICA_NETWORK_SEND_BUFFER_SIZE);
					} else {
						return new REFITNetworkEndpointWorker.ConnectionParameters(REFITSchedulerTaskType.CLIENT_NETWORK_ENDPOINT_WORKER,
								REFITConfig.CLIENT_NETWORK_RECEIVE_BUFFER_SIZE, 0);
					}
				}, REFITConfig.REPLICA_NETWORK_CONNECTORS,
				initialNodes, !createClientNetwork, initialConnects, true, false);
	}

	@Override
	public REFITSchedulerGroup getGroup() {
		return group;
	}

	@Override
	public short getNodeID() {
		return id;
	}

	@Override
	public String toString() {
		return "RPLCA";
	}


	// #################
	// # SETUP HELPERS #
	// #################

	protected abstract void assignTasks(REFITScheduler[] schedulers);

	protected REFITApplicationServer createApplication() {
		try {
			return REFITConfig.APPLICATION_SERVER.getConstructor().newInstance();
		} catch (NoSuchMethodException | IllegalAccessException | InstantiationException | InvocationTargetException e) {
			e.printStackTrace();
			System.exit(1);
			// Strange java things...
			return null;
		}
	}

	protected REFITSystemApplicationServer createSystemApplication() {
		try {
			return REFITConfig.SYSTEM_APPLICATION_SERVER.getConstructor().newInstance();
		} catch (NoSuchMethodException | IllegalAccessException | InstantiationException | InvocationTargetException e) {
			e.printStackTrace();
			System.exit(1);
			// Strange java things...
			return null;
		}
	}

	// ###########################
	// # CLIENT NETWORK ENDPOINT #
	// ###########################

	protected void receiveMessageFromClient(REFITMessage message) {
		clientsInbound.broadcast(message);
	}

	public void sendMessageToClient(REFITReplyBase reply, short clientID) {
		if (REFITConfig.ENABLE_DEBUG_CHECKS && clientID < REFITConfig.TOTAL_NR_OF_REPLICAS) {
			throw new AssertionError("Got client reply " + reply + " for id " + clientID + " which is not a clientID!");
		}
		// Complains if the client is not connected
		network.unicastMessage(reply, clientID);
	}

	protected void addToClientInbound(REFITStage stage) {
		clientsInbound.addListener(stage);
	}


	// ############################
	// # REPLICA NETWORK ENDPOINT #
	// ############################

	protected void addToInitialConnectObserver(REFITStage stage) {
		network.initCompletedSender.addListener(stage);
	}

	protected void addToInbound(REFITMessageType type, REFITStage stage) {
		if (!inbounds.containsKey(type)) {
			inbounds.put(type, new REFITObserver<>());
		}
		inbounds.get(type).addListener(stage);
	}

	protected void addClientInbound(REFITMessageType type) {
		if (!inbounds.containsKey(type)) {
			inbounds.put(type, new REFITObserver<>());
		}
		inbounds.get(type).addListenersFrom(clientsInbound);
	}

	@Override
	public void receiveMessage(ByteBuffer message, short nodeID) {
		if (nodeID < REFITConfig.TOTAL_NR_OF_REPLICAS) {
			// Create message
			REFITMessage msg = REFITMessage.createMessage(message);
			if (REFITLogger.LOG_COMMUNICATION)
				REFITLogger.logCommunication(this, "received message " + message + " from " + msg.from);

			// Forward message
			receiveMessageFromReplica(msg);
		} else {
			// Create request
			REFITMessage request = REFITMessage.createMessage(message);
			if (request.type != REFITMessageType.REQUEST && request.type != REFITMessageType.REQUEST_WRAPPER) {
				REFITLogger.logWarning(this, "Dropping message of unexpected type from client " + nodeID + " " + request.type);
				return;
			}
			if (REFITConfig.COLLECT_MESSAGE_STATISTICS) request.getContext().statistics.track("REC");

			if (request.from == nodeID) {
				receiveMessageFromClient(request);
			} else {
				REFITLogger.logWarning(this, "Faked sender identity " + request + " " + request.from + " " + nodeID);
			}
		}
	}

	protected void receiveMessageFromReplica(REFITMessage message) {
		// check message type and then group type
		REFITObserver<REFITMessage> observer = inbounds.get(message.type);
		if (observer == null) {
			observer = inbounds.get(message.groupType());
		}
		if (observer != null) {
			observer.broadcast(message);
		} else {
			REFITLogger.logError(this, "drop message of unexpected type " + message.type);
		}
	}

	public boolean replicaBoundedUnicast(REFITMessage message, short replicaID, REFITOutbox outbox) {
		return network.unicastMessageBounded(message, replicaID, outbox);
	}

	// ################
	// # START / STOP #
	// ################

	public void start() {
		// Create schedulers
		schedulers = new REFITScheduler[REFITConfig.NR_OF_REPLICA_SCHEDULERS];
		CountDownLatch startupSync = new CountDownLatch(REFITConfig.NR_OF_REPLICA_SCHEDULERS);
		REFITHangDetector hangDetector = (REFITConfig.TRACK_SCHEDULER_HANGS) ? new REFITHangDetector() : null;
		if (hangDetector != null) hangDetector.start();
		for (int i = 0; i < schedulers.length; i++) {
			schedulers[i] = new REFITScheduler(startupSync, hangDetector);
			schedulers[i].setName("RPLC" + id + "-" + i);
		}

		assignTasks(schedulers);

		group.assertAllTaskAreAssigned();

		// Mark as running
		REFITLogger.logEvent(this, "READY");

		// Start schedulers
		for (REFITScheduler scheduler : schedulers) scheduler.start();
	}

	public void stop() {
		for (REFITScheduler scheduler : schedulers) {
			scheduler.interrupt();
		}
		for (REFITScheduler scheduler : schedulers) {
			try {
				scheduler.join();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		network.shutdown();
	}


	// ########
	// # MAIN #
	// ########

	// NOTE: Most replica implementations are better suited by the REFITReplicaRunner
	public static void standaloneMain(String[] args) {
		// Create and initialize replica
		short replicaID = Short.parseShort(args[0]);
		REFITReplica replica = new REFITReplica(replicaID, (REFITConfig.NR_OF_REPLICA_SCHEDULERS == 1));

		// Start replica
		replica.start();
	}
}
