package refit.replica;

import refit.agreement.steward.REFITSteward;
import refit.application.REFITApplicationServer;
import refit.application.REFITSystemApplicationServer;
import refit.message.REFITMessageType;
import refit.replica.checkpoint.REFITCheckpointStage;
import refit.replica.client.REFITStewardClientStage;
import refit.replica.execution.REFITExecutionStage;
import refit.replica.order.REFITOrderStage;
import refit.replica.stewardorder.REFITStewardGroups;
import refit.replica.stewardorder.REFITStewardOrderStage;
import refit.scheduler.REFITScheduler;
import refit.scheduler.REFITSchedulerTaskType;

public class REFITStewardReplica extends REFITBaseReplica {
	private final REFITStewardClientStage clientStage;
	private final REFITOrderStage orderStage;
	private final REFITCheckpointStage execCheckpointStage;
	private final REFITExecutionStage execExecutionStage;
	private final REFITStewardOrderStage stewardOrderStage;

	public final int GLOBAL_LEADER = 0;
	public final int TOTAL_SITES = REFITStewardGroups.stewardOrderGroupCount();

	public REFITStewardReplica(short id, boolean isSingleThreaded) {
		super(id, isSingleThreaded, true);
		mac.loadStewardKeys();

		// create and initialize application
		final REFITApplicationServer application = createApplication();
		application.init();
		final REFITSystemApplicationServer systemApplication = createSystemApplication();
		systemApplication.init();

		execExecutionStage = new REFITExecutionStage(this, application, systemApplication);
		execCheckpointStage = new REFITCheckpointStage(this, REFITStewardGroups.stewardOrderGroup(id),
				REFITStewardGroups.stewardOrderGroup(id), null);
		addToInbound(REFITMessageType.CHECKPOINT, execCheckpointStage);
		addToInbound(REFITMessageType.CHECKPOINT_STATE, execCheckpointStage);
		execExecutionStage.checkpointSender.addListener(execCheckpointStage);
		execCheckpointStage.progressSender.addListener(execExecutionStage);
		execCheckpointStage.stableCPSender.addListener(execExecutionStage);
		execCheckpointStage.stableCPStateSender.addListener(execExecutionStage);

		int groupID = REFITStewardGroups.stewardOrderGroupID(id);
		int leaderGroupID = REFITStewardGroups.stewardOrderGroupID(GLOBAL_LEADER);
		orderStage = (groupID == leaderGroupID) ? new REFITOrderStage(this, REFITSteward.REFITStewardProtocol.instance, (short) 0, null) : null;
		if (orderStage != null) {
			addToInbound(REFITMessageType.ORDER, orderStage);
			execCheckpointStage.progressSender.addListener(orderStage);
		}

		stewardOrderStage = new REFITStewardOrderStage(this, GLOBAL_LEADER,
				leaderGroupID, TOTAL_SITES);
		execCheckpointStage.progressSender.addListener(stewardOrderStage);
		if (orderStage != null)
			orderStage.requestSender.addListener(stewardOrderStage);
		stewardOrderStage.requestSender.addListener(execExecutionStage);
		addToInbound(REFITMessageType.STEWARD_PROPOSAL, stewardOrderStage);
		addToInbound(REFITMessageType.STEWARD_ACCEPT_SHARE, stewardOrderStage);
		addToInbound(REFITMessageType.STEWARD_ACCEPT, stewardOrderStage);

		clientStage = new REFITStewardClientStage(this, (short) GLOBAL_LEADER);
		addToClientInbound(clientStage);
		// treat message as directly received from a client
		addClientInbound(REFITMessageType.REQUEST_WRAPPER);

		execExecutionStage.replySenders[0].addListener(clientStage);
		clientStage.executionSender.addListener(execExecutionStage);
		if (orderStage != null) {
			if (GLOBAL_LEADER == id) {
				clientStage.requestSender.addListener(orderStage);
				clientStage.panicSender.addListener(orderStage);
			}
			orderStage.viewChangeSender.addListener(clientStage);
			orderStage.viewChangeSender.addListener(execExecutionStage);
		} else {
			stewardOrderStage.configSender.addListener(execExecutionStage);
		}
	}

	@Override
	protected void assignTasks(REFITScheduler[] schedulers) {
		// Assign tasks to schedulers
		schedulers[1 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.NETWORK_ENDPOINT);
		schedulers[1 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.CLIENT_NETWORK_ENDPOINT_WORKER);
		schedulers[2 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.CLIENT_STAGE);
		schedulers[3 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.REPLICA_NETWORK_ENDPOINT_WORKER);
		schedulers[0 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.ORDER_STAGE);
		schedulers[0 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.CHECKPOINT_STAGE);
		schedulers[0 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.EXECUTION_STAGE);
		schedulers[0 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.COUPLING_STAGE);
		schedulers[0 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.EXEC_COUPLING_STAGE);
		schedulers[0 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.CHANNEL_SENDER_STAGE);
		schedulers[0 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.CHANNEL_RECEIVER_STAGE);
	}
}
