package refit.replica;

import java.util.Arrays;

import refit.agreement.weave.REFITWeave.REFITWeaveProtocol;
import refit.application.REFITApplicationServer;
import refit.application.REFITSystemApplicationServer;
import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITMessageType;
import refit.replica.checkpoint.REFITCheckpointStage;
import refit.replica.client.REFITClientStage;
import refit.replica.execution.REFITExecutionStage;
import refit.replica.fuzzycheckpoint.REFITFuzzyCheckpointStage;
import refit.replica.order.REFITMenciusSplitter;
import refit.replica.order.REFITOrderGroups;
import refit.replica.order.REFITOrderStage;
import refit.replica.order.REFITTimingTracker;
import refit.replica.update.REFITUpdateStage;
import refit.scheduler.REFITScheduler;
import refit.scheduler.REFITSchedulerTaskType;
import refit.util.REFITAssert;

public class REFITWeaveReplica extends REFITBaseReplica {
	private final REFITClientStage[] clientStage;
	private final REFITOrderStage[] orderStage;
	private final REFITMenciusSplitter menciusSplitter;

	private final REFITUpdateStage updateStage;
	private final REFITCheckpointStage checkpointStage;
	private final REFITExecutionStage executionStage;
	private final REFITFuzzyCheckpointStage fuzzyCheckpointStage;

	private final REFITGroupMessageRouter msgRouter;

	static {
		// minimal sanity check
		REFITAssert.assertTrue(REFITConfig.LOCATION_BASED_ORDER_GROUPS
				&& REFITConfig.TOTAL_NR_OF_REPLICAS == REFITConfig.ORDER_GROUP_SIZE
				&& REFITConfig.TOTAL_NR_OF_REPLICAS == REFITConfig.EXEC_GROUP_SIZE, "Incompatible configuration");
	}

	public REFITWeaveReplica(short id, boolean isSingleThreaded) {
		super(id, isSingleThreaded, true);

		for (int i = 0; i < REFITOrderGroups.COUNT; i++) {
			REFITLogger.logDebug(this, "Order group " + i + ": " + Arrays.toString(REFITConfig.ORDER_GROUPS[i]));
		}

		msgRouter = new REFITGroupMessageRouter(this);
		addToClientInbound(msgRouter);
		addToInbound(REFITMessageType.ORDER, msgRouter);

		// Create global stages
		// create and initialize application
		final REFITApplicationServer application = createApplication();
		application.init();
		final REFITSystemApplicationServer systemApplication = createSystemApplication();
		systemApplication.init();

		this.executionStage = new REFITExecutionStage(this, application, systemApplication);
		this.fuzzyCheckpointStage = (REFITConfig.ENABLE_DIFFERENTIAL_CHECKPOINTING || REFITConfig.ENABLE_FUZZY_CHECKPOINTS)
				? new REFITFuzzyCheckpointStage(this, application) : null;

		this.updateStage = REFITConfig.USE_PASSIVE_REPLICATION ? new REFITUpdateStage(this, REFITReplicaGroups.getExecutionReplicas(0).clone()) : null;
		if (updateStage != null) {
			addToInbound(REFITMessageType.UPDATE, updateStage);
			addToInbound(REFITMessageType.BATCH_UPDATE, updateStage);
		}
		this.checkpointStage = new REFITCheckpointStage(this, REFITReplicaGroups.getExecutionReplicas(0),
				REFITReplicaGroups.getAllReplicas(), null);
		addToInbound(REFITMessageType.CHECKPOINT, checkpointStage);
		addToInbound(REFITMessageType.CHECKPOINT_STATE, checkpointStage);

		// Create group stages
		clientStage = new REFITClientStage[REFITOrderGroups.COUNT];
		orderStage = new REFITOrderStage[REFITOrderGroups.COUNT];

		short myPrimaryOrderGroupID = REFITOrderGroups.getPrimaryGroup(id);
		REFITTimingTracker tracker = (REFITConfig.MENCIUS_MODE) ? null : new REFITTimingTracker(this, myPrimaryOrderGroupID);
		if (tracker != null) {
			addToInitialConnectObserver(tracker);
		}

		menciusSplitter = (REFITConfig.MENCIUS_MODE) ? new REFITMenciusSplitter(this) : null;

		for (short groupID = 0; groupID < REFITOrderGroups.COUNT; groupID++) {
			// also setup client and coupling stage on execution replicas
			clientStage[groupID] = new REFITClientStage(this, groupID);
			msgRouter.requestSenders[groupID].addListener(clientStage[groupID]);
			executionStage.replySenders[groupID].addListener(clientStage[groupID]);
			clientStage[groupID].readRequestSender.addListener(executionStage);
			orderStage[groupID] = (REFITConfig.MENCIUS_MODE && groupID > 0) ? null : new REFITOrderStage(this,
					REFITWeaveProtocol.instance, (REFITConfig.MENCIUS_MODE) ? -1 : groupID, tracker);

			REFITOrderStage order = orderStage[(REFITConfig.MENCIUS_MODE) ? 0 : groupID];
			msgRouter.orderSenders[groupID].addListener(order);

			clientStage[groupID].requestSender.addListener(order);
			clientStage[groupID].panicSender.addListener(order);

			if (!REFITConfig.MENCIUS_MODE || groupID == 0) {
				order.requestSender.addListener(executionStage);
				order.viewChangeSender.addListener(executionStage);
				checkpointStage.progressSender.addListener(order);

				REFITStage client = (REFITConfig.MENCIUS_MODE) ? menciusSplitter : clientStage[groupID];
				order.ackSender.addListener(client);
				order.viewChangeSender.addListener(client);
				if (updateStage != null) {
					order.viewChangeSender.addListener(updateStage);
				}
			}

			if (menciusSplitter != null) {
				menciusSplitter.viewChangeSenders[groupID].addListener(clientStage[groupID]);
				menciusSplitter.ackSenders[groupID].addListener(clientStage[groupID]);
			}
		}
		if (tracker != null) {
			// replica can only propose requests in it's primary group
			tracker.flushSender.addListener(orderStage[myPrimaryOrderGroupID]);
		}

		executionStage.checkpointSender.addListener(checkpointStage);
		checkpointStage.progressSender.addListener(executionStage);
		checkpointStage.stableCPSender.addListener(executionStage);
		checkpointStage.stableCPStateSender.addListener(executionStage);

		if (fuzzyCheckpointStage != null) {
			executionStage.updateSender.addListener(fuzzyCheckpointStage);
			executionStage.fuzzyCheckpointNotificationSender.addListener(fuzzyCheckpointStage);

			fuzzyCheckpointStage.fuzzyCheckpointNotificationSender.addListener(executionStage);
			fuzzyCheckpointStage.checkpointSender.addListener(checkpointStage);
		}

		if (updateStage != null) {
			checkpointStage.progressSender.addListener(updateStage);
			executionStage.updateSender.addListener(updateStage);
			updateStage.stableUpdateSender.addListener(executionStage);
		}

		// treat message as directly received from a client
		addClientInbound(REFITMessageType.REQUEST);
	}

	@Override
	protected void assignTasks(REFITScheduler[] schedulers) {
		// Assign tasks to schedulers
		schedulers[1 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.NETWORK_ENDPOINT);
		schedulers[1 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.CLIENT_NETWORK_ENDPOINT_WORKER);
		schedulers[1 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.CLIENT_STAGE);
		schedulers[0 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.REPLICA_NETWORK_ENDPOINT_WORKER);
		schedulers[0 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.ORDER_STAGE);
		schedulers[2 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.CHECKPOINT_STAGE);
		schedulers[3 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.COUPLING_STAGE);
		schedulers[3 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.EXECUTION_STAGE);
		schedulers[3 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.UPDATE_STAGE);
		schedulers[4 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.FUZZY_CHECKPOINT_STAGE);
		schedulers[0 % schedulers.length].assignTasks(group, REFITSchedulerTaskType.PROBE_STAGE);
	}
}
