package refit.replica.checkpoint;

import java.util.Arrays;
import java.util.HashSet;
import java.util.List;

import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITMessageAuthentication;
import refit.util.REFITBallotBox;
import refit.util.REFITPayload;


public class REFITCheckpointCertificate {

	public final long agreementSeqNr;
	private boolean[] legitimateSenders;
	private final REFITBallotBox<Short, REFITCheckpointCertificateVote, REFITCheckpoint> checkpoints;
	private final REFITBallotBox<Short, REFITCheckpointCertificateVote, REFITCheckpoint> lightweightCheckpoints;
	private final HashSet<Short> obsolete;

	public enum AddResult {
		INVALID,
		DUPLICATE,
		ADDED
	}

	public REFITCheckpointCertificate(long agreementSeqNr, boolean[] legitimateSenders) {
		this.agreementSeqNr = agreementSeqNr;
		this.legitimateSenders = legitimateSenders;
		this.checkpoints = new REFITBallotBox<>(REFITConfig.REGULAR_CHECKPOINT_STABILITY_THRESHOLD);
		this.lightweightCheckpoints = new REFITBallotBox<>(REFITConfig.LIGHTWEIGHT_CHECKPOINT_STABILITY_THRESHOLD);
		this.obsolete = new HashSet<>();
	}

	public AddResult add(REFITCheckpoint checkpoint, REFITMessageAuthentication mac) {
		if (checkpoint == null) return AddResult.INVALID;
		if (agreementSeqNr != checkpoint.uid.seqNr) return AddResult.INVALID;
		if (!mac.verifySignature(checkpoint, legitimateSenders)) return AddResult.INVALID;
		REFITCheckpointCertificateVote vote = new REFITCheckpointCertificateVote(checkpoint.isLightweight(), checkpoint.contentHash);
		REFITBallotBox<Short, REFITCheckpointCertificateVote, REFITCheckpoint> ballotBox = (checkpoint.isLightweight()) ? lightweightCheckpoints : checkpoints;
		boolean wasAdded = ballotBox.add(checkpoint.from, vote, checkpoint);
		return wasAdded ? AddResult.ADDED : AddResult.DUPLICATE;
	}

	public boolean isStable() {
		REFITCheckpointCertificateVote decision = checkpoints.getDecision();
		if (decision != null) return true;
		REFITCheckpointCertificateVote lightDecision = lightweightCheckpoints.getDecision();
		return lightDecision != null;
	}

	public List<REFITCheckpoint> getDecidingBallots() {
		List<REFITCheckpoint> correctCheckpoints = checkpoints.getDecidingBallots();
		if (correctCheckpoints != null) {
			return correctCheckpoints;
		}
		return lightweightCheckpoints.getDecidingBallots();
	}

	public REFITCheckpoint getStableCheckpoint() {
		// Check whether the checkpoint has become stable
		List<REFITCheckpoint> correctCheckpoints = getDecidingBallots();
		if (correctCheckpoints == null) return null;

		// Return checkpoint with state if available
		for (REFITCheckpoint correctCheckpoint : correctCheckpoints) {
			if (correctCheckpoint.hasState()) {
				return correctCheckpoint;
			}
		}

		// No full checkpoint available -> Return a correct hash checkpoint
		return correctCheckpoints.get(0);
	}

	public void markSenderObsolete(short sender) {
		if (checkpoints.hasVoted(sender)) {
			obsolete.add(sender);
		}
	}

	public boolean isObsolete() {
		return obsolete.size() == Math.max(checkpoints.getVoteCount(), lightweightCheckpoints.getVoteCount());
	}


	// ######################
	// # CHECKPOINT CONTENT #
	// ######################

	private static class REFITCheckpointCertificateVote {

		private final boolean lightweight;
		private final byte[] hash;


		public REFITCheckpointCertificateVote(boolean lightweight, byte[] hash) {
			this.lightweight = lightweight;
			this.hash = hash;
		}

		@Override
		public boolean equals(Object object) {
			if (this == object) return true;
			if (object == null || getClass() != object.getClass()) return false;
			REFITCheckpointCertificateVote other = (REFITCheckpointCertificateVote) object;
			if (lightweight != other.lightweight) return false;
			if (!Arrays.equals(hash, other.hash)) {
				REFITLogger.logWarning("[CHKPT]", "checkpoint content mismatch: " + REFITPayload.toString(hash) + " vs " + REFITPayload.toString(other.hash));
				return false;
			}
			return true;
		}

		@Override
		public int hashCode() {
			// fixed hash code to ensure that the checkpoint content comparison works
			// the comparison will only ever happen between a few objects so this doesn't get too inefficient
			return 0;
		}
	}

}
