package refit.replica.checkpoint;

import java.util.*;

import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITInstruction;
import refit.message.REFITInstruction.REFITCheckpointAssist;
import refit.message.REFITInstruction.REFITProgressNotification;
import refit.message.REFITMessage;
import refit.replica.*;
import refit.scheduler.REFITSchedulerTaskType;


public class REFITCheckpointStage extends REFITBoundedStage {
	public final REFITObserver<REFITProgressNotification> progressSender;
	public final REFITObserver<REFITCheckpoint> stableCPSender;
	public final REFITObserver<REFITCheckpointState> stableCPStateSender;

	private final boolean[] checkpointerGroup;
	private final boolean[] checkpointReceivers;
	private final REFITIncrementalOutbox outbox;
	private final REFITIncrementalOutbox stableCPOutbox;
	private final REFITIncrementalOutbox stableCPStateOutbox;

	private final boolean[] bootstrapGroup;

	private final Map<Long, REFITCheckpointCertificate> certificates;
	private final Map<Long, REFITCheckpointState> states;
	private REFITCheckpointCertificate latestStableCheckpoint;
	private final long[] latestCheckpointSeqNr;

	private REFITCheckpointAssist assistRequest;


	public REFITCheckpointStage(REFITBaseReplica replica, boolean[] checkpointerGroup, boolean[] checkpointReceivers,
	                            boolean[] bootstrapGroup) {
		super(REFITSchedulerTaskType.CHECKPOINT_STAGE, replica);
//		markInterleaved();
		progressSender = new REFITObserver<>();
		// the checkpoint stage is basically agnostic to the actual content of the checkpoint
		// it just distributes the checkpoint and notifies listening stages once a checkpoint becomes stable
		stableCPSender = new REFITObserver<>();
		stableCPStateSender = new REFITObserver<>();

		this.checkpointerGroup = checkpointerGroup;
		this.checkpointReceivers = checkpointReceivers;
		this.bootstrapGroup = bootstrapGroup;

		int maxCheckpoints = REFITConfig.MAXIMUM_ORDER_INSTANCES_IN_PROGRESS / REFITConfig.CHECKPOINT_INTERVAL + 1;
		outbox = new REFITIncrementalOutbox(replica, this, maxCheckpoints);
		// no need for send limits as the other replicas must always accept checkpoint messages
		outbox.updateSendLimits(Long.MAX_VALUE);

		stableCPOutbox = new REFITIncrementalOutbox(replica, this, REFITConfig.ORDER_GROUP_SIZE);
		stableCPOutbox.updateSendLimits(Long.MAX_VALUE);

		stableCPStateOutbox = new REFITIncrementalOutbox(replica, this, 1);
		stableCPStateOutbox.updateSendLimits(Long.MAX_VALUE);

		certificates = new HashMap<>();
		states = new HashMap<>();
		latestStableCheckpoint = new REFITCheckpointCertificate(-1L, checkpointerGroup);
		latestCheckpointSeqNr = new long[REFITConfig.TOTAL_NR_OF_REPLICAS];
	}

	@Override
	public String toString() {
		return "CHKPT";
	}

	@Override
	protected void handleMessage(REFITMessage message) {
		switch (message.type) {
		case CHECKPOINT:
			handleCheckpoint((REFITCheckpoint) message);
			break;
		case CHECKPOINT_STATE:
			handleCheckpointState((REFITCheckpointState) message);
			break;
		case INSTRUCTION:
			handleInstruction((REFITInstruction) message);
			break;
		default:
			REFITLogger.logError(this, "drop message of unexpected type " + message.type);
		}
	}

	@Override
	protected void stageComplete() {
		if (assistRequest != null && latestStableCheckpoint.agreementSeqNr == assistRequest.minCheckpointSeqNr) {
			if (assistRequest.fullSender == replica.id || assistRequest.fullSender + 1 == replica.id) {
				if (states.containsKey(assistRequest.minCheckpointSeqNr)) {
					forwardAssistCheckpoint();

					boolean[] executors = REFITReplicaGroups.getExecutionReplicas(assistRequest.groupIdx);
					REFITCheckpointState checkpoint = states.get(assistRequest.minCheckpointSeqNr);
					checkpoint.serializeMessage();
					replica.mac.appendNullMAC(checkpoint);
					stableCPStateOutbox.queue(checkpoint, executors);
					assistRequest = null;
				}
			} else {
				forwardAssistCheckpoint();
				assistRequest = null;
			}
		}
	}

	private void forwardAssistCheckpoint() {
		boolean[] executors = REFITReplicaGroups.getExecutionReplicas(assistRequest.groupIdx);
		List<REFITCheckpoint> checkpoints = latestStableCheckpoint.getDecidingBallots();
		for (REFITCheckpoint checkp : checkpoints) {
			stableCPOutbox.queue(checkp, executors);
		}
	}

	private void handleInstruction(REFITInstruction message) {
		switch (message.instructionType) {
		case CHECKPOINT_ASSIST:
			assistRequest = (REFITCheckpointAssist) message;
			break;
		default:
			REFITLogger.logError(this, "drop instruction of unexpected type " + message.instructionType);
		}
	}

	private void handleCheckpoint(REFITCheckpoint checkpoint) {
		// Do not process old checkpoints
		if (checkpoint.uid.seqNr <= latestStableCheckpoint.agreementSeqNr) {
			REFITCheckpoint lsc = latestStableCheckpoint.getStableCheckpoint();
			if (checkpoint.uid.seqNr == latestStableCheckpoint.agreementSeqNr && checkpoint.hasState()
					&& !lsc.hasState()) {
				// forward local checkpoint state if we haven't done so already
				lsc.setState(checkpoint.getState());
				if (!checkpoint.isLightweight()) {
					stableCPStateSender.broadcast(checkpoint.getState());
				}
			}
			return;
		}
		if ((checkpoint.uid.seqNr % REFITConfig.CHECKPOINT_INTERVAL) != 0) {
			REFITLogger.logWarning(this, "dropping checkpoint created out of order " + checkpoint);
			return;
		}

		boolean[] group;
		if (!checkpointerGroup[checkpoint.from] && bootstrapGroup != null && bootstrapGroup[checkpoint.from]
				&& latestStableCheckpoint.agreementSeqNr <= 0) {
			// the initial checkpoint may be sent from another group
			group = bootstrapGroup;
		} else {
			group = checkpointerGroup;
		}
		if (!replica.mac.verifySignature(checkpoint, group)) return;

		// ignore older checkpoints from the sender
		if (checkpoint.uid.seqNr < latestCheckpointSeqNr[checkpoint.from] - REFITConfig.MAXIMUM_ORDER_INSTANCES_IN_PROGRESS) {
			REFITLogger.logWarning(this, "Got old checkpoint from replica " + checkpoint
					+ " already seen up to " + latestCheckpointSeqNr[checkpoint.from]);
			return;
		}
		latestCheckpointSeqNr[checkpoint.from] = Math.max(latestCheckpointSeqNr[checkpoint.from], checkpoint.uid.seqNr);

		// Get checkpoint certificate
		REFITCheckpointCertificate certificate = ensureCertificate(checkpoint.uid.seqNr, group);

		// Store checkpoint
		if (REFITLogger.LOG_CHECKPOINT) REFITLogger.logCheckpoint(this, "received checkpoint " + checkpoint);
		REFITCheckpointCertificate.AddResult addResult = certificate.add(checkpoint, replica.mac);
		if (addResult.equals(REFITCheckpointCertificate.AddResult.INVALID)) {
			REFITLogger.logWarning(this, "bad CHECKPOINT (" + checkpoint + ")");

			// drop certificate again if it is empty
			if (certificate.isObsolete()) {
				certificates.remove(checkpoint.uid.seqNr);
			}
			return;
		}

		// Bound state required for checkpoint certificates, if a sender is already far in the future
		Iterator<Map.Entry<Long, REFITCheckpointCertificate>> it = certificates.entrySet().iterator();
		while (it.hasNext()) {
			Map.Entry<Long, REFITCheckpointCertificate> entry = it.next();
			// Keep only the checkpoints of a sender which could fit in its order window
			if (entry.getKey() < checkpoint.uid.seqNr - REFITConfig.MAXIMUM_ORDER_INSTANCES_IN_PROGRESS) {
				entry.getValue().markSenderObsolete(checkpoint.from);
			}
			if (entry.getValue().isObsolete()) {
				it.remove();
			}
		}

		// Check whether the checkpoint became stable
		if (!certificate.isStable()) return;
		storeStableCheckpoint(certificate);
		forwardLatestCheckpoint();

		if (assistRequest != null && assistRequest.minCheckpointSeqNr < certificate.agreementSeqNr) {
			assistRequest = new REFITCheckpointAssist(assistRequest.groupIdx, certificate.agreementSeqNr, assistRequest.fullSender);
		}
	}

	private REFITCheckpointCertificate ensureCertificate(long seqNr, boolean[] group) {
		REFITCheckpointCertificate certificate = certificates.get(seqNr);
		if (certificate == null) {
			certificate = new REFITCheckpointCertificate(seqNr, group);
			certificates.put(seqNr, certificate);
		}
		return certificate;
	}

	private void storeStableCheckpoint(REFITCheckpointCertificate certificate) {
		// Stable checkpoint reached
		latestStableCheckpoint = certificate;
		if (REFITLogger.LOG_CHECKPOINT)
			REFITLogger.logCheckpoint(this, "stable checkpoint for agreement sequence number " + latestStableCheckpoint.agreementSeqNr);

		// Garbage-collect older certificates
		certificates.entrySet().removeIf(entry -> entry.getKey() <= latestStableCheckpoint.agreementSeqNr);
		outbox.collectGarbageBefore(latestStableCheckpoint.agreementSeqNr);
		states.entrySet().removeIf(entry -> entry.getKey() < latestStableCheckpoint.agreementSeqNr);
	}

	private void forwardLatestCheckpoint() {
		// Distribute progress notification
		REFITProgressNotification progressNotification = new REFITProgressNotification(latestStableCheckpoint.agreementSeqNr);
		progressSender.broadcast(progressNotification);

		// Nothing more to do for stable lightweight checkpoints
		REFITCheckpoint checkpoint = latestStableCheckpoint.getStableCheckpoint();
		if (checkpoint.isLightweight()) return;

		// Forward information that the checkpoint has become stable
		stableCPSender.broadcast(checkpoint);
		if (checkpoint.hasState()) {
			stableCPStateSender.broadcast(checkpoint.getState());
		}

		// Broadcast parts of a stable CP to everyone
		stableCPOutbox.collectGarbageBefore(latestStableCheckpoint.agreementSeqNr);
		for (REFITCheckpoint checkpointPart : latestStableCheckpoint.getDecidingBallots()) {
			stableCPOutbox.queue(checkpointPart, checkpointReceivers);
		}
	}

	private void handleCheckpointState(REFITCheckpointState checkpoint) {
		if (checkpoint.from == replica.id && checkpoint.isInternal()) {
			sendOwnCheckpointShare(checkpoint);

			if (checkpoint.uid.seqNr >= latestStableCheckpoint.agreementSeqNr) {
				states.put(checkpoint.uid.seqNr, checkpoint);
			}
		}

		if (!checkpoint.isInternal()) {
			if (checkpoint.uid.seqNr != latestStableCheckpoint.agreementSeqNr) {
				REFITLogger.logWarning(this, "Expected checkpoint with seqNr " + latestStableCheckpoint.agreementSeqNr + " got " + checkpoint.uid.seqNr);
				return;
			}
			if (!Arrays.equals(checkpoint.getContentHash(), latestStableCheckpoint.getStableCheckpoint().contentHash)) {
				REFITLogger.logWarning(this, "Ignoring checkpoint state with bad hash " + checkpoint);
				return;
			}
			REFITLogger.logWarning(this, "Got cp");
			states.put(checkpoint.uid.seqNr, checkpoint);

			if (!checkpoint.isLightweight()) {
				stableCPStateSender.broadcast(checkpoint);
			}
		}
	}

	private void sendOwnCheckpointShare(REFITCheckpointState state) {
		REFITCheckpoint checkpoint = state.toCheckpoint();
		checkpoint.setState(state);
		checkpoint.serializeMessage();
		replica.mac.appendSignature(checkpoint);
		if (REFITLogger.LOG_CHECKPOINT)
			REFITLogger.logCheckpoint(this, "distribute own checkpoint " + checkpoint);
		outbox.queue(checkpoint, checkpointReceivers);
		// process locally
		handleCheckpoint(checkpoint);
	}
}
