package refit.replica.checkpoint;

import java.nio.ByteBuffer;
import java.security.MessageDigest;

import refit.message.REFITMessage;
import refit.message.REFITMessageType;
import refit.message.REFITUniqueID;
import refit.util.REFITPayload;

import static refit.replica.checkpoint.REFITCheckpoint.LIGHTWEIGHT_CHECKPOINT_MAGIC;
import static refit.replica.checkpoint.REFITCheckpoint.REGULAR_CHECKPOINT_MAGIC;


public class REFITCheckpointState extends REFITMessage {

	public final REFITCheckpointObject[] data;
	private transient byte[] contentHash;


	public REFITCheckpointState(long agreementSeqNr, short from, boolean isLightweight, REFITCheckpointObject[] data) {
		super(REFITMessageType.CHECKPOINT_STATE, new REFITUniqueID(isLightweight ? LIGHTWEIGHT_CHECKPOINT_MAGIC : REGULAR_CHECKPOINT_MAGIC, agreementSeqNr), from);
		this.data = new REFITCheckpointObject[data.length];
		System.arraycopy(data, 0, this.data, 0, data.length);
	}

	public REFITCheckpointState(ByteBuffer buffer) {
		super(REFITMessageType.CHECKPOINT_STATE, buffer);
		int objectCount = buffer.getInt();
		this.data = new REFITCheckpointObject[objectCount];
		for (int i = 0; i < objectCount; i++) {
			final int size = buffer.getInt();
			int oldLimit = buffer.limit();
			buffer.limit(buffer.position() + size);
			ByteBuffer slice = buffer.slice();
			// the REFITObjects may later be converted to the correct type
			data[i] = new REFITCheckpointObject(slice);
			buffer.limit(oldLimit);
			buffer.position(buffer.position() + size);
		}
	}

	@Override
	public AuthenticationType getAuthenticationType() {
		return AuthenticationType.NONE;
	}

	@Override
	public String toString() {
		return String.format("{%s|%d|%d}", type, from, uid.seqNr);
	}

	@Override
	protected void serialize(ByteBuffer buffer) {
		super.serialize(buffer);
		buffer.putInt(data.length);
		for (REFITCheckpointObject object : data) {
			buffer.putInt(object.getMessageSize());
			buffer.put(object.getBuffer());
		}
	}

	@Override
	protected int calculatePayloadSize() {
		int size = super.calculatePayloadSize();
		size += Integer.BYTES;
		for (REFITCheckpointObject object : data) {
			object.serializeMessage();
			size += Integer.BYTES + object.getMessageSize();
		}

		return size;
	}

	public byte[] getContentHash() {
		if (contentHash != null) return contentHash;

		ByteBuffer buffer = ByteBuffer.allocate(Long.BYTES);
		buffer.putLong(uid.seqNr);

		// enforce hash calculation
		for (REFITCheckpointObject object : data) {
			object.getHash();
		}

		MessageDigest digest = REFITPayload.getDigest();
		digest.update(buffer);
		for (REFITCheckpointObject object : data) {
			digest.update(object.getHash());
		}
		contentHash = digest.digest();

		return contentHash;
	}

	public boolean isLightweight() {
		return uid.nodeID == LIGHTWEIGHT_CHECKPOINT_MAGIC;
	}

	public REFITCheckpoint toCheckpoint() {
		return new REFITCheckpoint(uid.seqNr, from, getContentHash(), isLightweight());
	}
}
