package refit.replica.client;

import java.util.ArrayList;

import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.*;
import refit.message.REFITInstruction.REFITConfigurationNotification;
import refit.replica.REFITBaseReplica;
import refit.replica.REFITObserver;
import refit.replica.REFITStage;
import refit.replica.order.REFITOrderGroups;
import refit.replica.stewardorder.REFITStewardGroups;
import refit.scheduler.REFITSchedulerTaskType;
import refit.util.REFITTime;


public class REFITStewardClientStage extends REFITStage {

	public final REFITObserver<REFITRequest> requestSender;
	public final REFITObserver<REFITReply> replySender;
	public final REFITObserver<REFITInstruction.REFITPanicNotification> panicSender;
	public final REFITObserver<REFITRequest> executionSender;

	private final short groupID;
	private final REFITStewardReplyCache replyCache;
	private short contactReplica;

	private long latestPanicNotification;
	private long panicInterval;


	public REFITStewardClientStage(REFITBaseReplica replica, short contactReplica) {
		super(REFITSchedulerTaskType.CLIENT_STAGE, replica);
		markInterleaved();
		this.replyCache = new REFITStewardReplyCache();
		this.contactReplica = contactReplica;

		requestSender = new REFITObserver<>();
		replySender = new REFITObserver<>();
		panicSender = new REFITObserver<>();
		executionSender = new REFITObserver<>();

		groupID = (short) REFITStewardGroups.stewardOrderGroupID(replica.id);

		resetTimeouts();
	}

	@Override
	public String toString() {
		return "OCLINT";
	}

	@Override
	protected void handleMessage(REFITMessage message) {
		switch (message.type) {
		case REQUEST_WRAPPER:
			handleRequestWrapper((REFITRequestWrapper) message);
			break;
		case REPLY:
			handleReply((REFITReply) message);
			break;
		case INSTRUCTION:
			handleInstruction((REFITInstruction) message);
			break;
		default:
			REFITLogger.logError(this, "drop message of unexpected type " + message.type);
		}
	}

	private void handleRequestWrapper(REFITRequestWrapper wrapper) {
		if (!replica.mac.verifyUnicastMAC(wrapper, REFITMessageAuthentication.ANY_SENDER)) return;

		REFITRequest request = wrapper.getRequest();

		// Ignore old requests
		if (replyCache.isOldUID(request.uid, request.mode.isTotalOrder())) return;

		// Resend reply if already available
		// The execution replicas are not always able to verify the request
		REFITReply reply = replyCache.getReply(request.uid, request.mode.isTotalOrder());
		if (reply != null) {
			sendReply(reply);
			return;
		}

		// Ignore request if it was already processed
		if (replyCache.isCurrentUID(request.uid, request.mode.isTotalOrder())) return;

		// Don't handle faulty requests
		if (!request.mode.isTotalOrder()) {
			replica.mac.verifyNullMAC(request);
		} else {
			if (!replica.mac.verifySignature(request, REFITMessageAuthentication.ANY_SENDER)) return;
		}

		// Ignore duplicate non-panic requests
		boolean isNewUID = replyCache.registerRequest(request);
		if (request.mode.isReadOnly()) {
			// retries are not useful for readonly requests
			if (isNewUID) {
				executionSender.broadcast(request);
			}
		} else {
			forwardRequest(request);
		}
	}

	private void forwardRequest(REFITRequest request) {
		if (replica.id != contactReplica && contactReplica >= 0) {
			REFITRequestWrapper wrapper = new REFITRequestWrapper(request.uid, replica.id, request);
			wrapper.serializeMessage();
			replica.mac.appendUnicastMAC(wrapper, contactReplica);

			replica.replicaUnicast(wrapper, contactReplica);
		} else {
			// Forward request to local order stage
			requestSender.broadcast(request);
		}
	}

	private void handleReply(REFITReply reply) {
		if (REFITConfig.COLLECT_MESSAGE_STATISTICS) reply.getContext().statistics.showTrace();

		// prevent overwriting published replies
		REFITReply oldReply = replyCache.getReply(reply.uid, reply.mode.isTotalOrder());
		if (oldReply != null && !oldReply.isInternalOnly) {
			if (!reply.isInternalOnly)
				REFITLogger.logWarning(this, "Trying to send a second reply to a client " + reply + " already sent " + oldReply);
			return;
		}

		replyCache.storeReply(reply);

		sendReply(reply);
	}

	private void sendReply(REFITReply reply) {
		if (reply.isInternalOnly) return;
		if (reply.groupID != -1 && reply.groupID != groupID) return;
		if (REFITConfig.USE_HASHED_REPLIES && reply.sendHashReply) {
			reply = reply.createHashedReply();
		}
		boolean serialized = reply.serializeMessage();
		if (serialized) replica.mac.appendUnicastMAC(reply, reply.uid.nodeID);
		replica.sendMessageToClient(reply, reply.uid.nodeID);
	}


	private void handleInstruction(REFITInstruction instruction) {
		switch (instruction.instructionType) {
		case CONFIGURATION_NOTIFICATION:
			REFITConfigurationNotification notification = (REFITConfigurationNotification) instruction;
			contactReplica = REFITOrderGroups.getGroupPrimary(notification.groupID, notification.viewID);

			// Flush incomplete requests to the order stage / new contact replica, only works with signatures
			ArrayList<REFITRequest> requests = replyCache.flushIncomplete();
			for (REFITRequest request : requests) {
				forwardRequest(request);
			}
			resetTimeouts();
			REFITLogger.logEvent(this, "Configuration update");
			break;
		default:
			REFITLogger.logError(this, "drop instruction of unexpected type " + instruction.instructionType);
		}
	}

	@Override
	public void init() {
		super.init();
		setTimeout(50);
	}

	@Override
	public void timeout() {
		super.timeout();
		// just check whether a panic should be triggered in regular intervals
		// This avoids all the complications related to rescheduling the check
		setTimeout(50);

		// Control the rate at which panic notifications are sent
		if (REFITTime.currentTimeMillis.getAsLong() < latestPanicNotification + panicInterval) return;
		// don't panic if replica itself is the leader
		if (contactReplica == replica.id || !replyCache.shouldPanic()) return;

		// Forward panic to order stage
		panicSender.broadcast(new REFITInstruction.REFITPanicNotification());
		// Update panic information
		latestPanicNotification = REFITTime.currentTimeMillis.getAsLong();
		panicInterval *= 2;
	}

	private void resetTimeouts() {
		latestPanicNotification = REFITTime.currentTimeMillis.getAsLong();
		panicInterval = REFITConfig.REPLICA_PANIC_INTERVAL;
	}
}
