package refit.replica.coupling;

import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.List;

import refit.config.REFITConfig;
import refit.message.REFITMessageAuthentication;
import refit.replica.order.REFITOrderGroups;
import refit.util.REFITBallotBox;

public class REFITOrderedCertificate {

	public enum AddResult {
		INVALID,
		DUPLICATE,
		ADDED
	}

	public final long agreementSeqNr;
	private final boolean localOnly;
	private final REFITBallotBox<Short, ByteBuffer, REFITOrderedRequest> operations;
	private final boolean[] legitimateSenders;
	private REFITOrderedRequest result;

	public REFITOrderedCertificate(long agreementSeqNr, boolean localOnly, boolean[] legitimateSenders) {
		this.agreementSeqNr = agreementSeqNr;
		this.localOnly = localOnly;
		operations = new REFITBallotBox<>(localOnly ? 1 : REFITConfig.EXECUTION_STABILITY_THRESHOLD);
		this.legitimateSenders = legitimateSenders;
		result = null;
	}

	public AddResult add(REFITOrderedRequest operation, REFITMessageAuthentication mac) {
		if (operation == null) return AddResult.INVALID;
		if (agreementSeqNr != operation.uid.seqNr) return AddResult.INVALID;
		if (!localOnly && !REFITOrderGroups.isInGroup(operation.from, agreementSeqNr)) return AddResult.INVALID;
		if (!mac.verifyMulticastMAC(operation, legitimateSenders)) return AddResult.INVALID;

		boolean success = operations.add(operation.from, ByteBuffer.wrap(operation.messageHash), operation);
		checkStable();
		return success ? AddResult.ADDED : AddResult.DUPLICATE;
	}

	private void checkStable() {
		// already stable
		if (result != null) return;

		ByteBuffer decision = operations.getDecision();
		if (decision == null) return;

		for (REFITOrderedRequest request : operations.getDecidingBallots()) {
			if (request.isExecutable()) {
				result = request;
				break;
			}
		}
		if (result == null) {
			result = operations.getDecidingBallots().get(0);
		}
	}

	public boolean isStable() {
		return result != null;
	}

	public REFITOrderedRequest getStableRequest() {
		// FIXME correct viewID calculation
		return result;
	}

	public int getView() {
		int[] views = new int[operations.getDecisionThreshold()];
		List<REFITOrderedRequest> decidingBallots = operations.getDecidingBallots();
		for (int i = 0; i < views.length; i++) {
			views[i] = decidingBallots.get(i).viewID;
		}
		Arrays.sort(views);
		// f+1 largest view
		return views[Math.max(0, views.length - 1 - REFITConfig.FAULTS_TO_TOLERATE)];
	}
}
