package refit.replica.execclient;

import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.*;
import refit.replica.REFITBaseReplica;
import refit.replica.REFITObserver;
import refit.replica.REFITReplicaGroups;
import refit.replica.REFITStage;
import refit.scheduler.REFITSchedulerTaskType;
import refit.util.REFITAssert;


public class REFITExecClientStage extends REFITStage {

	public final REFITObserver<REFITRequest> requestSender;
	public final REFITObserver<REFITRequest> readRequestSender;

	private final REFITReplyCache replyCache;
	private final short groupID;


	public REFITExecClientStage(REFITBaseReplica replica) {
		super(REFITSchedulerTaskType.CLIENT_STAGE, replica);
		markInterleaved();
		REFITAssert.assertTrue(REFITConfig.SIGNED_REQUESTS, "Assumes signed requests inside RequestWrapper messages");

		this.replyCache = new REFITReplyCache();

		requestSender = new REFITObserver<>();
		readRequestSender = new REFITObserver<>();

		groupID = REFITReplicaGroups.getExecutionGroupId(replica.id);
	}

	@Override
	public String toString() {
		return "EXCLI";
	}

	@Override
	protected void handleMessage(REFITMessage message) {
		switch (message.type) {
		case REQUEST_WRAPPER:
			handleWrapper((REFITRequestWrapper) message);
			break;
		case REPLY:
			handleReply((REFITReply) message);
			break;
		default:
			REFITLogger.logError(this, "drop message of unexpected type " + message.type);
		}
	}

	private void handleWrapper(REFITRequestWrapper wrapper) {
		if (!replica.mac.verifyUnicastMAC(wrapper, REFITMessageAuthentication.ANY_SENDER)) return;

		REFITRequest request = wrapper.getRequest();

		// Ignore old requests
		if (replyCache.isOldUID(request.uid, request.mode.isTotalOrder())) return;

		// Check for cached reply only after the request was sent at least once
		if (replyCache.isCurrentUID(request.uid, request.mode.isTotalOrder())) {
			// Resend reply if already available
			// The execution replicas are not always able to verify the request
			REFITReply reply = replyCache.getReply(request.uid, request.mode.isTotalOrder());
			if (reply != null) {
				REFITLogger.logWarning(this, "resend reply " + reply);
				sendReply(reply);
			}
			return;
		}

		// Don't handle faulty requests
		if (!request.mode.isTotalOrder()) {
			replica.mac.verifyNullMAC(request);
		} else {
			if (!replica.mac.verifySignature(request, REFITMessageAuthentication.ANY_SENDER)) return;
		}

		// Ignore duplicate non-panic requests
		boolean isNewUID = replyCache.registerRequest(request);
		if (!request.mode.isTotalOrder()) {
			// retries are not useful for readonly requests
			// silently drop non-total order write requests
			if (request.mode.isReadOnly() && isNewUID) {
				readRequestSender.broadcast(request);
			}
		} else {
			// hand over request to REFITMCClientStage
			requestSender.broadcast(request);
		}
	}

	private void handleReply(REFITReply reply) {
		if (REFITConfig.COLLECT_MESSAGE_STATISTICS) reply.getContext().statistics.showTrace();

		// prevent overwriting published replies
		REFITReply oldReply = replyCache.getReply(reply.uid, reply.mode.isTotalOrder());
		if (oldReply != null && !oldReply.isInternalOnly) {
			if (!reply.isInternalOnly)
				REFITLogger.logWarning(this, "Trying to send a second reply to a client " + reply + " already sent " + oldReply);
			return;
		}

		replyCache.storeReply(reply);
		sendReply(reply);
	}

	private void sendReply(REFITReply reply) {
		if (reply.isInternalOnly) return;
		if (reply.groupID != -1 && reply.groupID != groupID) return;
		if (REFITConfig.USE_HASHED_REPLIES && reply.sendHashReply) {
			reply = reply.createHashedReply();
		}
		boolean serialized = reply.serializeMessage();
		if (serialized) replica.mac.appendUnicastMAC(reply, reply.uid.nodeID);
		replica.sendMessageToClient(reply, reply.uid.nodeID);
	}
}
