package refit.replica.execution;

import java.util.*;

import refit.application.REFITApplicationServer;
import refit.application.REFITSystemApplicationServer;
import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.*;
import refit.message.REFITInstruction.REFITConfigurationNotification;
import refit.message.REFITInstruction.REFITExecuteRequest;
import refit.message.REFITInstruction.REFITExecutionProgressNotification;
import refit.message.REFITInstruction.REFITProgressNotification;
import refit.replica.REFITBaseReplica;
import refit.replica.checkpoint.REFITCheckpoint;
import refit.replica.checkpoint.REFITCheckpointState;
import refit.replica.fuzzycheckpoint.REFITFuzzyCheckpointNotification;
import refit.replica.order.REFITOrderProtocol;
import refit.replica.update.REFITBatchUpdate;
import refit.replica.update.REFITUpdateMessage;


public class REFITExecutionStage extends REFITExecutionStageBase {

	private static final REFITExecutionStagePriorityComparator OPERATION_PRIORITY_COMPARATOR = new REFITExecutionStagePriorityComparator();
	private static final REFITExecutionStageReadRequestComparator READ_REQUEST_COMPARATOR = new REFITExecutionStageReadRequestComparator();

	private final REFITExecutor executor;
	private final Map<Long, REFITCheckpoint> checkpoints;
	private final Map<Long, REFITUpdateMessage> updates;
	private final Map<Long, REFITExecuteRequest> requests;
	private final PriorityQueue<REFITRequest> readOnlyRequests;
	private long nextAgreementSeqNrToProcess;
	private long newestCheckpointSeqNr;

	private boolean checkpointProgress;
	private boolean updateProgress;
	private boolean requestProgress;
	private boolean fuzzyCheckpointProgress;

	public REFITExecutionStage(REFITBaseReplica replica, REFITApplicationServer application, REFITSystemApplicationServer systemApplication) {
		super(replica);

		this.executor = new REFITExecutor(replica.id, application, systemApplication, replySenders, checkpointSender, updateSender,
				fuzzyCheckpointNotificationSender, systemSender);
		this.nextAgreementSeqNrToProcess = 0L;
		this.newestCheckpointSeqNr = -1;
		this.checkpoints = new TreeMap<>(OPERATION_PRIORITY_COMPARATOR);
		this.updates = new HashMap<>();
		this.requests = new HashMap<>();
		this.readOnlyRequests = new PriorityQueue<>(READ_REQUEST_COMPARATOR);
		this.checkpointProgress = false;
		this.updateProgress = false;
		this.requestProgress = false;
	}


	@Override
	public String toString() {
		return "EXECT";
	}

	@Override
	protected void handleMessage(REFITMessage message) {
		switch (message.type) {
		case CHECKPOINT:
			handleCheckpoint((REFITCheckpoint) message);
			break;
		case CHECKPOINT_STATE:
			handleCheckpointState((REFITCheckpointState) message);
			break;
		case INSTRUCTION:
			handleInstruction((REFITInstruction) message);
			break;
		case FUZZY_CHECKPOINT_NOTIFICATION:
			handleFuzzyCheckpointNotification((REFITFuzzyCheckpointNotification) message);
			break;
		case REQUEST:
			handleRequest((REFITRequest) message);
			break;
		case UPDATE:
		case BATCH_UPDATE:
			handleUpdate((REFITUpdateMessage) message);
			break;
		default:
			REFITLogger.logError(this, "drop message of unexpected type " + message.type);
		}
	}

	private void handleRequest(REFITRequest message) {
		readOnlyRequests.add(message);
	}

	private void handleInstruction(REFITInstruction instruction) {
		switch (instruction.instructionType) {
		case CONFIGURATION_NOTIFICATION:
			REFITConfigurationNotification notification = (REFITConfigurationNotification) instruction;
			REFITOrderProtocol protocol = notification.protocol;
			executor.setEnableCheckpoints(protocol.hasCheckpointRecipients(), protocol.useLightweightCheckpoints(), nextAgreementSeqNrToProcess);
			executor.setEnableUpdates(protocol.hasUpdateRecipients());
			break;
		case EXECUTE_REQUEST:
			REFITExecuteRequest operation = (REFITExecuteRequest) instruction;
			// Ignore old operations
			if (operation.agreementSeqNr < nextAgreementSeqNrToProcess) {
				break;
			}

			REFITExecuteRequest op = requests.get(operation.agreementSeqNr);
			if (op != null) {
				break;
			}
			requests.put(operation.agreementSeqNr, operation);
			if (isNextProcessableSeqNr(operation))
				requestProgress = true;
			break;
		case PROGRESS_NOTIFICATION:
			REFITProgressNotification progress = (REFITProgressNotification) instruction;
			if (progress.agreementSeqNr > newestCheckpointSeqNr) {
				newestCheckpointSeqNr = progress.agreementSeqNr;
				checkpointProgress = true;
			}
			break;
		default:
			REFITLogger.logError(this, "drop instruction of unexpected type " + instruction.instructionType);
		}
	}

	private boolean isNextProcessableSeqNr(REFITExecuteRequest operation) {
		if ((REFITConfig.ENABLE_DIFFERENTIAL_CHECKPOINTING || REFITConfig.ENABLE_FUZZY_CHECKPOINTS) && executor.mustBlockExecution()) {
			return false;
		}

		return operation.agreementSeqNr == nextAgreementSeqNrToProcess;
	}


	// #######################
	// # PRIORITY COMPARATOR #
	// #######################

	private static class REFITExecutionStagePriorityComparator implements Comparator<Long> {

		@Override
		public int compare(Long agreementSeqNrA, Long agreementSeqNrB) {
			return (int) (agreementSeqNrB - agreementSeqNrA);
		}
	}

	private static class REFITExecutionStageReadRequestComparator implements Comparator<REFITRequest> {
		@Override
		public int compare(REFITRequest a, REFITRequest b) {
			return Long.compare(a.executeAfter, b.executeAfter);
		}
	}


	// ###############
	// # CHECKPOINTS #
	// ###############

	private void handleCheckpoint(REFITCheckpoint checkpoint) {
		// Ignore old checkpoints they won't be used to update the application state
		if (checkpoint.uid.seqNr <= nextAgreementSeqNrToProcess) return;
		if (checkpoints.containsKey(checkpoint.uid.seqNr)) return;
		checkpoints.put(checkpoint.uid.seqNr, checkpoint);
		checkpointProgress = true;
	}

	private void handleCheckpointState(REFITCheckpointState state) {
		// Ignore old checkpoints they will not lead to progress
		if (state.uid.seqNr <= nextAgreementSeqNrToProcess) return;

		// Get checkpoint
		REFITCheckpoint checkpoint = checkpoints.get(state.uid.seqNr);
		if (checkpoint == null) {
			throw new InternalError("no checkpoint cert for checkpoint state, agreementNr " + state.uid.seqNr);
		}

		// Attach full state to checkpoint
		boolean success = checkpoint.setState(state);
		if (!success && !checkpoint.hasState()) {
			REFITLogger.logWarning(this, "full checkpoint did not lead to the corresponding operation becoming executable " + state + " " + checkpoint);
			return;
		}
		checkpointProgress = true;
	}

	private void garbageCollectCheckpoints() {
		checkpoints.values().removeIf(checkpoint -> checkpoint.uid.seqNr <= nextAgreementSeqNrToProcess);
	}


	// ###########
	// # UPDATES #
	// ###########

	private void handleUpdate(REFITUpdateMessage update) {
		// Ignore old updates they will not lead to progress
		if (update.agreementSeqNr() < nextAgreementSeqNrToProcess) return;

		REFITUpdateMessage operation = updates.get(update.agreementSeqNr());
		if (operation != null && operation.isFullUpdate()) {
			// update is already complete
			return;
		}

		updates.put(update.agreementSeqNr(), update);
		if (update.agreementSeqNr() == nextAgreementSeqNrToProcess) updateProgress = true;
	}


	// #####################
	// # FUZZY CHECKPOINTS #
	// #####################

	private void handleFuzzyCheckpointNotification(REFITFuzzyCheckpointNotification notification) {
		switch (notification.notificationType) {
		case COLLECTION_DONE:
			handleCollectionDoneNotification();
			break;
		default:
			REFITLogger.logError(this, "drop notification of unexpected type " + notification.type);
			break;
		}
	}

	private void handleCollectionDoneNotification() {
		REFITLogger.logDebug(this, "Notified about finished collection while @ "
				+ nextAgreementSeqNrToProcess);
		executor.setFuzzyCheckpointCompleted(nextAgreementSeqNrToProcess);
		fuzzyCheckpointProgress = true;
	}

	// ############
	// # PROGRESS #
	// ############

	@Override
	protected void stageComplete() {
		boolean progress = false;
		if (checkpointProgress) {
			// Reset flags
			checkpointProgress = false;
			progress = processCheckpoints();
		}

		if (updateProgress || requestProgress || fuzzyCheckpointProgress || progress) {
			// Reset flags
			updateProgress = false;
			requestProgress = false;
			fuzzyCheckpointProgress = false;
			processOperations();
			garbageCollectCheckpoints();
		}

		processReadOnlyOperations();
	}

	private boolean processCheckpoints() {
		// Check whether there is a stable checkpoint that allows the replica to make progress
		REFITCheckpoint newestStableState = null;
		REFITCheckpoint newestStableCheckpoint = null;
		for (REFITCheckpoint checkpointOperation : checkpoints.values()) {
			if (checkpointOperation.hasState() && (newestStableState == null
					|| checkpointOperation.uid.seqNr > newestStableState.uid.seqNr)) {
				newestStableState = checkpointOperation;
			}

			if (newestStableCheckpoint == null
					|| checkpointOperation.uid.seqNr > newestStableCheckpoint.uid.seqNr) {
				newestStableCheckpoint = checkpointOperation;
			}
		}

		if (newestStableState == null) {
			if (newestStableCheckpoint != null) {
				REFITLogger.logWarning(this, "TODO: implement checkpoint fetching"); // Somewhere above: Mark the latest stable checkpoint for transfer
				garbageCollectRequests(newestStableCheckpoint.uid.seqNr - REFITConfig.MAXIMUM_ORDER_INSTANCES_IN_PROGRESS);
			}
			return false;
		}

		// Apply checkpoint
		REFITCheckpointState checkpoint = newestStableState.getState();
		executor.applyCheckpoint(checkpoint);
		nextAgreementSeqNrToProcess = newestStableState.uid.seqNr;
		sendLastSeqNr();

		garbageCollectCheckpoints();
		garbageCollectRequests(nextAgreementSeqNrToProcess);

		return true;
	}

	private void garbageCollectRequests(long nextSeqNr) {
		// Garbage collect updates
		updates.keySet().removeIf(aLong -> aLong < nextSeqNr);

		// Garbage collect requests
		requests.keySet().removeIf(aLong -> aLong < nextSeqNr);
	}

	private void processOperations() {
		long startTime = System.nanoTime();

		// Check whether there are stable updates or requests that allow the replica to make progress
		while (true) {
			// Try to apply a full stable update
			REFITUpdateMessage update = updates.get(nextAgreementSeqNrToProcess);
			// FIXME TODO: implement update fetching?
			if (update != null && update.isFullUpdate()) {
				updates.remove(nextAgreementSeqNrToProcess);

				// Apply update
				executor.applyUpdate(update);
				nextAgreementSeqNrToProcess = ((update.type == REFITMessageType.BATCH_UPDATE) ? ((REFITBatchUpdate) update).getMaxAgreementSeqNr() : update.agreementSeqNr()) + 1;
				continue;
			}
			if (nextAgreementSeqNrToProcess >= newestCheckpointSeqNr + REFITConfig.MAXIMUM_ORDER_INSTANCES_IN_PROGRESS)
				break;

			// Try to process a full stable request
			REFITExecuteRequest requestOperation = requests.get(nextAgreementSeqNrToProcess);
			if (requestOperation == null || executor.mustBlockExecution()) break;

			// Process latest stable request
			requests.remove(nextAgreementSeqNrToProcess);
			executor.processRequest((REFITOrderRequest) requestOperation.message, requestOperation.agreementSeqNr, requestOperation.viewID);
			nextAgreementSeqNrToProcess++;
			sendLastSeqNr();

			long endTime = System.nanoTime();
			// yield when execution took already more than 1ms
			if (endTime - startTime > 1000 * 1000) {
				// make sure to run on the next scheduler sub-iteration
				requestProgress = true;
				progress();
				break;
			}
		}
	}

	private void processReadOnlyOperations() {
		while (!readOnlyRequests.isEmpty() && readOnlyRequests.peek().executeAfter < nextAgreementSeqNrToProcess) {
			REFITRequest request = readOnlyRequests.poll();
			executor.processReadOnlyRequest(request, nextAgreementSeqNrToProcess - 1);
		}
	}

	private void sendLastSeqNr() {
		REFITExecutionProgressNotification notification = new REFITExecutionProgressNotification(nextAgreementSeqNrToProcess);
		lastSeqNrSender.broadcast(notification);
	}

}
