package refit.replica.mccoupling;

import java.util.*;

import refit.client.REFITClientLibrary.RequestMode;
import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.*;
import refit.message.REFITInstruction.*;
import refit.replica.REFITBaseReplica;
import refit.replica.REFITObserver;
import refit.replica.REFITStage;
import refit.replica.checkpoint.REFITCheckpointObject;
import refit.replica.checkpoint.REFITCheckpointState;
import refit.replica.multichannel.REFITChannelItem;
import refit.scheduler.REFITSchedulerTaskType;

import static refit.replica.mccoupling.REFITOrderCouplingState.INSTANCE_HISTORY;

public class REFITMCCouplingStage extends REFITStage {

	private static final int CLIENT_TOKENS = 1;

	public final REFITObserver<REFITReply> replySender;
	public final REFITObserver<REFITChannelItem> channelRequestSender;
	public final REFITObserver<REFITCheckpointState> checkpointSender;
	public final REFITObserver<REFITProgressNotification> progressSender;
	public final REFITObserver<REFITChannelConfig> channelConfigSender;
	public final REFITObserver<REFITChannelProgress> channelProgressSender;

	private final Map<Long, REFITOrderRequest> requestQueue;
	private long checkpointSeqNr = 0;
	private long nextSeqNr = 0;
	private long seqNrLimit = 0;
	private final ArrayDeque<REFITOrderRequest> backlog;
	private final long[] requestedProgresses; // move_window request from the executor
	private final long[] pendingProgresses; // latest request of a client waiting to be ordered
	private final long[] nodeProgresses; // ordered requests

	static {
		if (REFITConfig.MAXIMUM_EXECUTION_INSTANCES_IN_PROGRESS + REFITConfig.CHECKPOINT_INTERVAL > REFITConfig.MAXIMUM_ORDER_INSTANCES_IN_PROGRESS)
			throw new AssertionError("Order window must be larger than the execution window");
	}

	public REFITMCCouplingStage(REFITBaseReplica replica) {
		super(REFITSchedulerTaskType.COUPLING_STAGE, replica);
		replySender = new REFITObserver<>();
		channelRequestSender = new REFITObserver<>();
		checkpointSender = new REFITObserver<>();
		progressSender = new REFITObserver<>();
		channelConfigSender = new REFITObserver<>();
		channelProgressSender = new REFITObserver<>();

		requestQueue = new HashMap<>();
		nodeProgresses = new long[REFITConfig.ADDRESSES.length];
		Arrays.fill(nodeProgresses, -1L);
		requestedProgresses = new long[REFITConfig.ADDRESSES.length];
		Arrays.fill(requestedProgresses, -1L);
		pendingProgresses = new long[REFITConfig.ADDRESSES.length];
		Arrays.fill(pendingProgresses, -1L);

		backlog = new ArrayDeque<>();
	}

	@Override
	public String toString() {
		return "MCPL-SRC";
	}

	@Override
	public void init() {
		super.init();
		if (REFITConfig.GROUP_LATE_START_SEQNR > 0) {
			channelConfigSender.broadcast(new REFITChannelConfig(REFITConfig.GROUP_LATE_START_GID, false));
		}
	}

	@Override
	protected void handleMessage(REFITMessage message) {
		switch (message.type) {
		case CHECKPOINT_STATE:
			handleCheckpoint((REFITCheckpointState) message);
			break;
		case REQUEST:
			handlePendingRequest((REFITRequest) message);
			break;
		case INSTRUCTION:
			handleInstruction((REFITInstruction) message);
			break;
		default:
			REFITLogger.logError(this, "drop message of unexpected type " + message.type);
		}
	}

	private void handleCheckpoint(REFITCheckpointState stableCheckpoint) {
		REFITProgressNotification progress = new REFITProgressNotification(stableCheckpoint.uid.seqNr);
		progressSender.broadcast(progress);

		// garbage collect request queue
		long checkpointSeqNr = progress.agreementSeqNr;
		requestQueue.keySet().removeIf(aLong -> aLong < checkpointSeqNr);

		// no need to further process old checkpoints
		if (checkpointSeqNr < nextSeqNr) {
			this.checkpointSeqNr = Math.max(checkpointSeqNr, this.checkpointSeqNr);
			return;
		}

		REFITOrderCouplingState state = new REFITOrderCouplingState(stableCheckpoint.uid.seqNr, stableCheckpoint.data[0].getBuffer());

		for (short i = REFITConfig.TOTAL_NR_OF_REPLICAS; i < state.nodeProgresses.length; i++) {
			// the checkpoint only contains client progresses, just block panics
			REFITReply reply = new REFITReply(new REFITUniqueID(i, state.nodeProgresses[i]), replica.id, (short) -1, RequestMode.WRITE);
			replySender.broadcast(reply);
			// update stored node progresses
			nodeProgresses[i] = state.nodeProgresses[i];
		}
		backlog.clear();
		Collections.addAll(backlog, state.requests);

		long skipped = (checkpointSeqNr - nextSeqNr);
		int requeue = (int) Math.max(0, backlog.size() - skipped);

		for (int i = requeue; i < backlog.size(); i++) {
			requestQueue.put(checkpointSeqNr - backlog.size() + i, state.requests[i]);
		}

		nextSeqNr = checkpointSeqNr - backlog.size() + requeue;
		this.checkpointSeqNr = checkpointSeqNr;
		sendChannelProgress();
	}

	private void handlePendingRequest(REFITRequest message) {
		// track timestamp of latest client request
		pendingProgresses[message.uid.nodeID] = Math.max(pendingProgresses[message.uid.nodeID], message.uid.seqNr);
	}

	private void handleInstruction(REFITInstruction message) {
		switch (message.instructionType) {
		case CHANNEL_PROGRESS:
			REFITChannelProgress progress = (REFITChannelProgress) message;
			seqNrLimit = progress.baseSeqNr + progress.tokens;
			break;
		case CHANNEL_SKIP:
			REFITChannelSkip skip = (REFITChannelSkip) message;
			short clientID = (short) (skip.subChannelID + REFITConfig.TOTAL_NR_OF_REPLICAS);
			requestedProgresses[clientID] = skip.baseSeqNr;
			sendSingleChannelProgress(clientID);
			break;
		case EXECUTE_REQUEST:
			REFITExecuteRequest operation = (REFITExecuteRequest) message;
			// no need to process already completed sequence numbers
			if (operation.agreementSeqNr < nextSeqNr || operation.agreementSeqNr < checkpointSeqNr) {
				break;
			}
			REFITOrderRequest request = ((REFITOrderRequest) operation.message);
			requestQueue.put(operation.agreementSeqNr, request);

			if (operation.agreementSeqNr == REFITConfig.GROUP_LATE_START_SEQNR) {
				channelConfigSender.broadcast(new REFITChannelConfig(REFITConfig.GROUP_LATE_START_GID, true));
			}
			break;
		default:
			REFITLogger.logError(this, "drop instruction of unexpected type " + message.type);
		}
	}

	@Override
	protected void stageComplete() {
		// linear iteration over sequence numbers is only necessary for proper timeout handling
		while (requestQueue.containsKey(nextSeqNr) && nextSeqNr < seqNrLimit) {
			REFITOrderRequest request = requestQueue.remove(nextSeqNr);

			if (nextSeqNr >= checkpointSeqNr) {
				backlog.add(request);
				if (backlog.size() > INSTANCE_HISTORY) {
					backlog.pollFirst();
				}
				// No reply yet, but we already have the guarantee that the request will be processed some time in the future
				// Requires that if one correct replica delivers the request, then eventually all correct replicas will deliver it
				blockPanics(request);
			}

			REFITChannelItem channelRequest = new REFITChannelItem(nextSeqNr, (short) 0, request);
			channelRequestSender.broadcast(channelRequest);

			nextSeqNr++;
			if (nextSeqNr >= checkpointSeqNr) {
				createCheckpointIfNecessary();
			}
		}
	}

	private void blockPanics(REFITOrderRequest request) {
		if (request.type == REFITMessageType.BATCH) {
			for (REFITRequest req : ((REFITBatch) request).requests) {
				blockSinglePanic(req);
			}
		} else if (request.type != REFITMessageType.REQUEST_NO_OP) {
			blockSinglePanic(request);
		}
	}

	private void blockSinglePanic(REFITOrderRequest request) {
		REFITUniqueID uid = request.uid;
		REFITRequest req = (REFITRequest) request;
		REFITReply reply = new REFITReply(uid, replica.id, req.groupID, req.mode);
		replySender.broadcast(reply);
		nodeProgresses[uid.nodeID] = Math.max(nodeProgresses[uid.nodeID], uid.seqNr);
		sendSingleChannelProgress(uid.nodeID);
	}

	private void createCheckpointIfNecessary() {
		if ((nextSeqNr % REFITConfig.CHECKPOINT_INTERVAL) == 0) {
			REFITOrderRequest[] reqs = backlog.toArray(new REFITOrderRequest[0]);
			REFITOrderCouplingState state = new REFITOrderCouplingState(nextSeqNr, nodeProgresses, reqs);
			state.serializeMessage();
			REFITCheckpointState checkpoint = new REFITCheckpointState(nextSeqNr, replica.id, false, new REFITCheckpointObject[]{state});
			checkpointSender.broadcast(checkpoint);
		}
	}

	private void sendChannelProgress() {
		for (short i = REFITConfig.TOTAL_NR_OF_REPLICAS; i < nodeProgresses.length; i++) {
			sendSingleChannelProgress(i);
		}
	}

	private void sendSingleChannelProgress(int clientID) {
		// exclusive
		// the really correct variant would be to use the oldest pending request of a client as basis,
		// that is not yet covered by nodeProgresses, if there's none fall-back to requestedProgresses
		// For CLIENT_TOKENS = 1, it is sufficient to just use the id of latest pending request
		long progressLimit = CLIENT_TOKENS
				+ ((pendingProgresses[clientID] > nodeProgresses[clientID])
				? pendingProgresses[clientID] : requestedProgresses[clientID]);

		int tokens = (progressLimit > requestedProgresses[clientID]) ? (int) (progressLimit - requestedProgresses[clientID]) : 0;
		channelProgressSender.broadcast(new REFITChannelProgress((short) (clientID - REFITConfig.TOTAL_NR_OF_REPLICAS),
				requestedProgresses[clientID], -1, tokens));
	}
}
