package refit.replica.multichannel;

import java.util.HashMap;
import java.util.HashSet;

import refit.config.REFITConfig;
import refit.message.REFITMessageType;
import refit.replica.REFITBoundedBaseReplica;
import refit.replica.REFITBoundedStage;
import refit.replica.REFITOutbox;

public class REFITChannelAckOutbox implements REFITOutbox {
	private final REFITBoundedBaseReplica replica;
	private final REFITBoundedStage stage;
	private final REFITMessageType ackType;

	private final HashMap<Short, Long> ackProgress;
	private final HashSet<Short>[] perSenderAckProgress;
	private long ackMsgSeqNr = 0;
	private boolean newEntries;
	private boolean[] writeBlocked;
	private boolean[] targetDisabled;

	public REFITChannelAckOutbox(REFITBoundedBaseReplica replica, REFITBoundedStage stage, REFITMessageType ackType, boolean[] ackReceivers) {
		this.replica = replica;
		this.stage = stage;
		this.ackType = ackType;

		ackProgress = new HashMap<>();
		//noinspection unchecked
		this.perSenderAckProgress = (HashSet<Short>[]) new HashSet[REFITConfig.TOTAL_NR_OF_REPLICAS];
		for (int i = 0; i < ackReceivers.length; i++) {
			if (ackReceivers[i]) {
				perSenderAckProgress[i] = new HashSet<>();
			}
		}

		targetDisabled = new boolean[REFITConfig.TOTAL_NR_OF_REPLICAS];
		writeBlocked = new boolean[REFITConfig.TOTAL_NR_OF_REPLICAS];
		stage.registerOutbox(this);
	}

	@Override
	public REFITBoundedStage getStage() {
		return stage;
	}

	// The user must make sure to queue the AckOutbox for sending
	public void addProgress(short subChannelId, long seqNr) {
		if (subChannelId < 0) throw new IllegalArgumentException();
		// noop if no progress was made
		if (ackProgress.containsKey(subChannelId) && seqNr <= ackProgress.get(subChannelId)) {
			return;
		}

		storeProgress(subChannelId, seqNr);
	}

	protected void storeProgress(short subChannelId, long seqNr) {
		ackProgress.put(subChannelId, seqNr);
		for (HashSet<Short> set : perSenderAckProgress) {
			if (set != null) {
				set.add(subChannelId);
			}
		}
		newEntries = true;
	}

	// The user must make sure to queue the AckOutbox for sending
	public void addSpecialProgress(short subChannelId, long seqNr) {
		if (subChannelId >= 0) throw new IllegalArgumentException();
		// noop if nothing has changed
		if (ackProgress.containsKey(subChannelId) && seqNr == ackProgress.get(subChannelId)) {
			return;
		}

		storeProgress(subChannelId, seqNr);
	}

	public boolean hasNewEntries() {
		return newEntries;
	}

	// return true if there are messages ready for sending when a CONNECTION_WRITEABLE_NOTIFICATION instruction arrives
	public void trySend() {
		newEntries = false;

		for (short replicaID = 0; replicaID < perSenderAckProgress.length; replicaID++) {
			HashSet<Short> ackProgress = perSenderAckProgress[replicaID];
			if (ackProgress == null || writeBlocked[replicaID] || targetDisabled[replicaID] || ackProgress.isEmpty()) {
				continue;
			}

			short[] changedSubChannels = new short[ackProgress.size()];
			long[] changedAckSeqNr = new long[changedSubChannels.length];
			int idx = 0;
			for (Short channelID : ackProgress) {
				changedSubChannels[idx] = channelID;
				changedAckSeqNr[idx] = this.ackProgress.get(channelID);
				idx++;
			}

			REFITChannelAck ack = new REFITChannelAck(ackType, ackMsgSeqNr, replica.id, changedSubChannels, changedAckSeqNr);
			ack.serializeMessage();
			replica.mac.appendUnicastMAC(ack, replicaID);
			boolean wasAccepted = replica.replicaBoundedUnicast(ack, replicaID, this);

			if (wasAccepted) {
				// sequence number is not used by the other side, thus its doesn't matter whether the sequence number is per replica or not
				ackMsgSeqNr++;
				ackProgress.clear();
			} else {
				writeBlocked[replicaID] = true;
			}
		}
	}

	@Override
	public void notifyWriteable(short replicaID) {
		writeBlocked[replicaID] = false;
	}

	public void setTargetDisable(short replicaID, boolean isDisabled) {
		targetDisabled[replicaID] = isDisabled;
		if (isDisabled) {
			newEntries = true;
		}
	}

	@Override
	public void requeue(short replicaID) {
		HashSet<Short> progress = this.perSenderAckProgress[replicaID];
		if (progress == null) {
			return;
		}

		progress.addAll(ackProgress.keySet());
		stage.enqueueOutbox(this);
	}
}
