package refit.replica.multichannel;

import java.nio.ByteBuffer;
import java.security.MessageDigest;
import java.util.function.Function;

import refit.message.*;
import refit.util.REFITPayload;

public class REFITShare extends REFITMessage {

	private static final int PAYLOAD_HEADER_SIZE = Byte.BYTES;

	private final REFITMessage content;
	private byte[] payloadHash;

	public REFITShare(REFITMessageType type, long seqNr, short subChannel, short from, REFITMessage content) {
		this(type, new REFITUniqueID(subChannel, seqNr), from, content, null);
	}

	private REFITShare(REFITMessageType type, REFITUniqueID uid, short from, REFITMessage content, byte[] hash) {
		super(type, uid, from);
		this.content = content;
		this.payloadHash = hash;
	}

	public REFITShare(REFITMessageType type, ByteBuffer buffer) {
		super(type, buffer);
		byte payloadType = buffer.get();
		if (payloadType == 0) {
			content = REFITMessage.createMessage(buffer.slice());
			buffer.position(buffer.position() + content.getMessageSize());
			payloadHash = null;
		} else if (payloadType == 1) {
			content = REFITNoOpRequest.INSTANCE;
			payloadHash = null;
		} else {
			content = null;
			payloadHash = new byte[REFITPayload.HASH_SIZE];
			buffer.get(payloadHash);
		}
	}

	@Override
	public REFITMessageType groupType() {
		return REFITMessageType.CHANNEL_MSG;
	}

	@Override
	public AuthenticationType getAuthenticationType() {
		return AuthenticationType.SIGNATURE;
	}

	@Override
	public String toString() {
		return String.format("{%s|%d|%s|%s}", type, from, uid, REFITPayload.toString(getPayloadHash()));
	}

	public REFITMessage getContent() {
		return content;
	}

	@Override
	public boolean equals(Object object) {
		if (this == object) return true;
		if (object == null || getClass() != object.getClass()) return false;
		if (!super.equals(object)) return false;
		REFITShare other = (REFITShare) object;
		return content.equals(other.content);
	}

	@Override
	protected void serialize(ByteBuffer buffer) {
		super.serialize(buffer);
		if (content == null) {
			buffer.put((byte) 2);
			buffer.put(payloadHash);
		} else if (content.type == REFITMessageType.REQUEST_NO_OP) {
			buffer.put((byte) 1);
		} else {
			buffer.put((byte) 0);
			content.serializeInto(buffer);
		}
	}

	@Override
	protected int calculatePayloadSize() {
		int contentSize;
		if (content == null) {
			contentSize = payloadHash.length;
		} else if (content.type == REFITMessageType.REQUEST_NO_OP) {
			contentSize = 0;
		} else {
			contentSize = content.calculateMessageSize();
		}
		return (super.calculatePayloadSize() + PAYLOAD_HEADER_SIZE + contentSize);
	}

	public REFITShare toFilteredBundle(Function<REFITMessage, REFITMessage> filter) {
		if (filter == null) return this;

		REFITMessage message = filter.apply(content);
		REFITShare share = new REFITShare(type, uid, from, message, payloadHash);
		share.serializeMessage();
		share.setPadding(getPadding());
		share.markVerified();
		return share;
	}

	public boolean isHashOnly() {
		return content == null;
	}

	public REFITShare toHashOnly() {
		if (content.type == REFITMessageType.REQUEST_NO_OP) {
			return this;
		}

		REFITShare share = new REFITShare(type, uid, from, null, getPayloadHash());
		share.serializeMessage();

		// Copy signature from original message, implicitly checks that messages were already serialized
		ByteBuffer hashBuffer = share.getBuffer();
		hashBuffer.position(share.getMessageSize() - share.getPaddingSize());
		ByteBuffer myBuffer = getBuffer();
		myBuffer.position(getMessageSize() - getPaddingSize());
		hashBuffer.put(myBuffer);

		share.markVerified();

		return share;
	}

	@Override
	protected byte[] calculateHash() {
		// Prepare message header
		ByteBuffer messageHeader = getInternalBuffer();
		// Skip size field
		messageHeader.position(REFITBaseMessage.HEADER_SIZE);
		messageHeader.limit(getHeaderSize());

		// Make sure that payload hash has been calculated (must be ensured in advance)
		getPayloadHash();

		// Calculate hash
		MessageDigest digest = REFITPayload.getDigest();
		digest.update(messageHeader);
		digest.update(payloadHash);
		return digest.digest();
	}

	public byte[] getPayloadHash() {
		if (payloadHash == null) {
			payloadHash = content.getHash();
		}
		return payloadHash;
	}

}
