package refit.replica.multichannel;

import java.nio.ByteBuffer;
import java.util.List;

import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITMessageAuthentication;
import refit.message.REFITUniqueID;
import refit.util.REFITBallotBox;

public class REFITShareCollectionCertificate implements REFITChannelInboxCertificate<REFITShare> {

	final int STABILITY_THRESHOLD;

	private final REFITMessageAuthentication mac;

	private final REFITUniqueID uid;
	private boolean[] legitimateSenders;
	private final REFITBallotBox<Short, ByteBuffer, REFITShare> shares;
	private boolean hasFullShare = false;

	private REFITShare[] result;

	public REFITShareCollectionCertificate(REFITUniqueID uid, REFITMessageAuthentication mac, boolean[] legitimateSenders, int senderFaults) {
		this.mac = mac;

		this.uid = uid;
		this.legitimateSenders = legitimateSenders;
		STABILITY_THRESHOLD = senderFaults + 1;
		this.shares = new REFITBallotBox<>(STABILITY_THRESHOLD);
	}

	@Override
	public boolean add(REFITShare share) {
		// Check reply
		if (share == null)
			return false;
		if (!uid.equals(share.uid))
			return false;
		if (shares.getDecision() != null && !hasFullShare && share.isHashOnly())
			return true; // Ignore hash-only shares if we just need a full share
		if (!mac.verifySignature(share, legitimateSenders))
			return false;

		// Prepare vote
		ByteBuffer vote = ByteBuffer.wrap(share.getPayloadHash());
		hasFullShare |= !share.isHashOnly();

		// Cast vote
		boolean success = shares.add(share.from, vote, share);
		if (success) checkResult();
		return success;
	}

	@Override
	public boolean isStable() {
		return (result != null);
	}

	@Override
	public REFITShare[] getResult() {
		return result;
	}

	private void checkResult() {
		if (isStable()) return;
		if (REFITConfig.ENABLE_DEBUG_CHECKS) if (shares.votesDiffer())
			REFITLogger.logWarning("ACS-DST", "received different executes for request " + uid);
		List<REFITShare> resultSet = shares.getDecidingBallots();
		if (resultSet == null || !hasFullShare) return;

		// move the full share to the front
		result = new REFITShare[STABILITY_THRESHOLD];
		for (REFITShare share : resultSet) {
			if (!share.isHashOnly()) {
				result[0] = share;
				break;
			}
		}
		// fill up with remaining shares
		int idx = 1;
		for (REFITShare share : resultSet) {
			if (share != result[0]) {
				result[idx++] = share;
			}
			if (idx == result.length) {
				break;
			}
		}
	}
}
