package refit.replica.order;

import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;

import refit.agreement.weave.REFITWeaveMessages.REFITWeaveMessagePrepare;
import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.*;
import refit.message.REFITInstruction.*;
import refit.replica.REFITBaseReplica;
import refit.replica.REFITObserver;
import refit.replica.REFITStage;
import refit.scheduler.REFITSchedulerTaskType;


public class REFITOrderStage extends REFITStage {

	public final REFITObserver<REFITExecuteRequest> requestSender;
	public final REFITObserver<REFITConfigurationNotification> viewChangeSender;
	public final REFITObserver<REFITInternalViewChangeNotification> internalViewChangeSender;
	private final REFITOrderProtocol initialProtocol;

	private int targetMinView;
	private long targetAfterSeq;
	private final short groupID;
	private final short limitedGroupID;

	private static final int THROTTLE_TIMER_INTERVAL = 25;

	private final REFITTimingTracker timingTracker;
	public final REFITObserver<REFITAckReply> ackSender;
	private final long[] noopedUpTo = new long[REFITOrderGroups.COUNT];

	public REFITOrderStage(REFITBaseReplica replica, REFITOrderProtocol initialProtocol, short groupID, REFITTimingTracker timingTracker) {
		// Initialize data structures
		super(REFITSchedulerTaskType.ORDER_STAGE, replica);
		if (groupID == -1 && !REFITConfig.MENCIUS_MODE) throw new IllegalStateException("Invalid order group id");
		this.groupID = groupID;
		this.limitedGroupID = (short) Math.max(0, groupID);

		this.readySlots = new TreeSet<>();
		this.proposals = new LinkedList<>();
		this.waitingProposals = new LinkedList<>();
		this.activeSlotLower = 0;
		this.activeSlotLimit = (REFITConfig.USE_SPINNING_LEADER) ? REFITConfig.ACTIVE_PROPOSALS : REFITConfig.MAXIMUM_ORDER_INSTANCES_IN_PROGRESS;
		this.nextProposerID = 0;
		this.viewID = REFITConfig.INITIAL_VIEW;
		this.viewState = REFITTransitionProtocolState.VIEW_STABLE;

		requestSender = new REFITObserver<>();
		viewChangeSender = new REFITObserver<>();
		internalViewChangeSender = new REFITObserver<>();
		ackSender = new REFITObserver<>();
		this.timingTracker = timingTracker;

		// Create protocol-instance slots
		this.slots = new REFITOrderStageSlot[REFITConfig.MAXIMUM_ORDER_INSTANCES_IN_PROGRESS];
		this.windowStartID = -slots.length;
		this.initialProtocol = initialProtocol;
		for (int i = 0; i < slots.length; i++) {
			slots[i] = new REFITOrderStageSlot(replica, this, this.initialProtocol, requestSender);
		}

		proposedCounter = null;
		if (REFITConfig.ENABLE_THROTTLING) {
			// split limit between groups according to assigned performance
			int groupWeightSum = 0;
			for (int i = 0; i < REFITOrderGroups.COUNT; i++) {
				if (REFITOrderGroups.isInGroupByID(replica.id, i)) {
					groupWeightSum += REFITConfig.ORDER_CLIENT_WEIGHTS[i];
				}
			}
			relativeWeight = (double) REFITConfig.ORDER_CLIENT_WEIGHTS[groupID] / groupWeightSum;

			int requestPerSecond = (int) (REFITConfig.MAX_REQUESTS_PER_SECOND[replica.id] * relativeWeight);
			throttleIntervalLimit = Math.max(1,
					requestPerSecond * THROTTLE_TIMER_INTERVAL / 1000);
		} else {
			throttleIntervalLimit = -1;
		}
	}

	@Override
	public void init() {
		// Initialize order stage
		REFITProgressNotification initProgressNotification = new REFITProgressNotification(0);
		insertMessage(initProgressNotification);
		setupNewView(initialProtocol);
	}

	private double relativeWeight;
	private long lastForeignCounter = 0;
	private AtomicLong foreignCounter = new AtomicLong(0);
	private AtomicInteger proposedCounter;
	private final int throttleIntervalLimit;

	private void drainThrottle() {
		int amount = Math.max(1, throttleIntervalLimit); // ensure that at least one proposal is freed
		int ctr, newCtr;
		do {
			ctr = proposedCounter.get();
			newCtr = Math.max(0, ctr - amount);
		} while (!proposedCounter.compareAndSet(ctr, newCtr));
	}

	private void initThrottleTimer() {
		proposedCounter = new AtomicInteger(0);
		// create timer to wakeup order thread despite throttling
		Timer timer = new Timer(true);
		timer.schedule(new TimerTask() {
			@Override
			public void run() {
				drainThrottle();
				progress();
			}
		}, THROTTLE_TIMER_INTERVAL, THROTTLE_TIMER_INTERVAL);
	}

	@Override
	public void execute() {
		if (!REFITConfig.ENABLE_THROTTLING) {
			super.execute();
			return;
		}

		if (proposedCounter == null) {
			initThrottleTimer();
		}

		// distribute mac costs of other stages between the order groups
		long newForeignCounter = foreignCounter.addAndGet(replica.mac.processedDigestCount());
		proposedCounter.addAndGet((int) ((newForeignCounter - lastForeignCounter) * relativeWeight));
		lastForeignCounter = newForeignCounter;

		long startTime = (isInterleaved) ? System.nanoTime() : 0;
		// Process messages
		REFITMessage message;
		// assumes costs of 2 MAC for each message
		int remaining = (throttleIntervalLimit - proposedCounter.get() + 1) / 2;
		while ((remaining > 0) && (message = incomingMessages.poll()) != null) {
			if (REFITConfig.ENABLE_MESSAGE_TRACES) {
				try {
					handleMessage(message);
				} catch (Throwable t) {
					Exception e = new Exception("Fake exception to capture message source");
					e.setStackTrace(message.getSte());

					t.addSuppressed(e);
					throw t;
				}
			} else {
				handleMessage(message);
			}
			event();
			remaining--;
			if (isInterleaved) {
				long currentTime = System.nanoTime();
				// yield when execution took already more than 1ms
				if (currentTime - startTime > 1000 * 1000 && !incomingMessages.isEmpty()) {
					progress();
					break;
				}
			}
		}
		// Add costs for processed MACs
		proposedCounter.addAndGet(replica.mac.processedDigestCount());

		// Complete stage. Messages inserted into the queue by this method will be processed
		// on the next scheduler sub-iteration
		stageComplete();

		// return if throttled
		if (proposedCounter.get() < throttleIntervalLimit && !incomingMessages.isEmpty()) {
			// request another execution round
			progress();
		}
	}


	@Override
	public String toString() {
		return "ORDER";
	}

	protected void handleMessage(REFITMessage message) {
		switch (message.groupType()) {
		case ORDER:
			// Forward message to the corresponding slot
			REFITOrderProtocolMessage orderMessage = (REFITOrderProtocolMessage) message;
			REFITOrderStageSlot slot = slots[(int) (orderMessage.uid.seqNr % slots.length)];
			boolean ready = slot.insertMessage(orderMessage);
			if (ready) readySlots.add(slot);

			if (REFITConfig.MENCIUS_MODE && (orderMessage.type == REFITMessageType.WEAVE_PREPARE || orderMessage.type == REFITMessageType.WEAVE_COMMIT)) {
				synthesizeNoops(orderMessage.uid.seqNr, orderMessage.from);
				// skip own slots with lower sequence number
				synthesizeNoops(orderMessage.uid.seqNr, replica.id);
			}
			break;
		case INSTRUCTION:
			handleInstruction((REFITInstruction) message);
			break;
		case REQUEST:
			handleProposal(message);
			break;
		default:
			REFITLogger.logError(this, "drop message of unexpected type " + message.type);
		}
	}

	private void synthesizeNoops(long seqNr, short from) {
		short group = REFITOrderGroups.getPrimaryGroup(from);
		boolean isPrimary = REFITOrderGroups.getGroupPrimary(group, 0) == replica.id;
		long last = -1;
		for (long instanceID = Math.max(windowStartID, noopedUpTo[group]); (instanceID < activeSlotLimit)
				&& (instanceID < seqNr); instanceID++) {
			if (REFITOrderGroups.getGroupID(instanceID) != group) continue;
			last = instanceID;

			final REFITOrderStageSlot slot = slots[(int) (instanceID % slots.length)];
			if (slot.hasProposal()) continue;

			if (isPrimary) {
				slot.assignProposal(REFITNoOpRequest.INSTANCE);
				readySlots.add(slot);
			} else {
				short leaderID = REFITOrderGroups.getGroupPrimary(group, 0);
				REFITWeaveMessagePrepare prepare = new REFITWeaveMessagePrepare(instanceID, 0, REFITNoOpRequest.INSTANCE, leaderID);
//				REFITLogger.logWarning(this, "Skipping " + prepare);
				boolean ready = slot.insertMessage(prepare);
				if (ready) readySlots.add(slot);
			}
		}
		noopedUpTo[group] = last;
	}

	private void handleInstruction(REFITInstruction instruction) {
		switch (instruction.instructionType) {
		case FLUSH_REQUEST:
			REFITFlushRequestNotification flush = (REFITFlushRequestNotification) instruction;
			flushSlotsBeforeSeqNr = flush.flushSlotsBeforeSeqNr;
			partialBatchingBeforeSeqNr = flush.partialBatchingBeforeSeqNr;
			if (flush.proposeOnlyBeforeSeqNr >= 0) {
				proposeOnlyBeforeSeqNr = flush.proposeOnlyBeforeSeqNr;
			}
			break;
		case PROGRESS_NOTIFICATION: {
			REFITProgressNotification notification = (REFITProgressNotification) instruction;
			moveWindow(notification.agreementSeqNr);
			break;
		}
		case PANIC_NOTIFICATION:
			REFITLogger.logEvent(this, "protocol abort");
			abort();
			break;
		case SYSTEM_NOTIFICATION: {
			REFITSystemNotification notification = (REFITSystemNotification) instruction;
			targetMinView = notification.targetMinView;
			targetAfterSeq = notification.targetAfterSeq;
			triggerViewChangeIfRequested();
			break;
		}
		default:
			REFITLogger.logError(this, "drop instruction of unexpected type " + instruction.instructionType);
		}
	}


	// ###########################
	// # PROTOCOL-INSTANCE SLOTS #
	// ###########################

	private final REFITOrderStageSlot[] slots;
	private long windowStartID;
	private long targetStartID;

	private long activeSlotLower;
	private long activeSlotLimit;

	private void moveWindow(long newestCheckpointSeqNr) {
		// Keep a certain amount of old slots to allow replicas to catch up
		long newWindowStartID = Math.max(0, newestCheckpointSeqNr + REFITConfig.INSTANCES_AFTER_CHECKPOINT - REFITConfig.MAXIMUM_ORDER_INSTANCES_IN_PROGRESS);

		// Only move window forward
		if (newWindowStartID <= windowStartID) return;

		// smoothly move window forwards
		if (windowStartID > 0) {
			targetStartID = newWindowStartID;
			newWindowStartID = Math.min(newWindowStartID, windowStartID + 1000);
		}

		if (REFITLogger.LOG_ORDER)
			REFITLogger.logOrder(this, "move window from start id " + windowStartID + " to " + newWindowStartID);

		// Calculate which protocol instances to initialize
		long initInstancesStartID = Math.max(windowStartID + slots.length, newWindowStartID);
		long initInstancesEndID = newWindowStartID + slots.length;
		if (REFITLogger.LOG_ORDER)
			REFITLogger.logOrder(this, "initialize protocol instances from " + initInstancesStartID + " to " + (initInstancesEndID - 1));

		activeSlotLimit = Math.max(activeSlotLimit, (REFITConfig.USE_SPINNING_LEADER) ? newestCheckpointSeqNr + REFITConfig.ACTIVE_PROPOSALS : initInstancesEndID);
		// Initialize new protocol instances
		for (long instanceID = initInstancesStartID; instanceID < initInstancesEndID; instanceID++) {
			REFITOrderStageSlot slot = slots[(int) (instanceID % slots.length)];
			boolean ready = slot.init(instanceID, groupID != -1 && REFITOrderGroups.getGroupID(instanceID) != groupID, timingTracker != null);
			if (ready && instanceID < activeSlotLimit) readySlots.add(slot);
			else readySlots.remove(slot);
		}

		// Update window information
		windowStartID = newWindowStartID;
		// Nothing to propose before the checkpoint
		nextProposerID = Math.max(nextProposerID, newestCheckpointSeqNr);

		if (windowStartID < targetStartID) {
			if (!hasTimeout()) {
				setTimeout(2);
			}
		}

		triggerViewChangeIfRequested();
	}

	private void triggerViewChangeIfRequested() {
		if (windowStartID >= targetAfterSeq && viewID < targetMinView) {
			transitionToView(targetMinView);
			internalViewChangeSender.broadcast(new REFITInternalViewChangeNotification());
		}
	}

	@Override
	public void timeout() {
		moveWindow(targetStartID);

		if (REFITConfig.PREFER_FULL_BATCHES) {
			assignProposals(true);
		}
	}

	// ################################
	// # PROTOCOL-INSTANCE INVOCATION #
	// ################################

	private final Set<REFITOrderStageSlot> readySlots;

	@Override
	protected void stageComplete() {
		if (REFITConfig.USE_SPINNING_LEADER) {
			stageCompleteSpinning();
			return;
		}
		stageCompleteStandard();
	}

	private void stageCompleteStandard() {
		// Assign proposals to idle proposers
		assignProposals(false);

		long startTime = System.nanoTime();
		outer:
		do {
			if (viewIDUpdated && readySlots.isEmpty()) {
				viewIDUpdated = false;
				readyAllSlots();
			}
			// Invoke ready protocol instances in the order of their IDs (-> use of TreeSet)
			Iterator<REFITOrderStageSlot> it = readySlots.iterator();
			while (it.hasNext()) {
				REFITOrderStageSlot readySlot = it.next();
				readySlot.execute();
				it.remove();

				long endTime = System.nanoTime();
				// yield when processing took already more than 1ms
				if (endTime - startTime > 5 * 1000 * 1000) {
					// make sure to run on the next scheduler sub-iteration
					progress();
					break outer;
				}
			}
		} while (viewIDUpdated);
	}

	private void stageCompleteSpinning() {
		do {
			stageCompleteStandard();

			while (activeSlotLimit < windowStartID + slots.length) {
				REFITOrderStageSlot bottomSlot = slots[(int) ((activeSlotLower) % slots.length)];
				REFITOrderStageSlot topSlot = slots[(int) ((activeSlotLimit - 1) % slots.length)];
				if (bottomSlot.isComplete()) {
					activeSlotLower++;
					activeSlotLimit = Math.max(activeSlotLimit, activeSlotLower + REFITConfig.ACTIVE_PROPOSALS * REFITOrderGroups.COUNT);
					REFITOrderStageSlot nextSlot = slots[(int) ((activeSlotLimit - 1) % slots.length)];
					if (!nextSlot.isComplete()) {
						readySlots.add(nextSlot);
					}
				} else if (topSlot.isPrepared()) {
					// piggyback next PRE-PREPARE to COMMIT <=> send COMMIT and PRE-PREPARE at the same time
					// <=> increase activeSlotLimit a bit earlier on the next primary
					long nextGroupSeqNr = REFITOrderGroups.nextSequenceNumber(activeSlotLimit - 1, groupID);
					REFITOrderStageSlot nextSlot = slots[(int) (nextGroupSeqNr % slots.length)];
					//System.out.println("Expanding to " + nextGroupSeqNr + " from " + (activeSlotLimit-1) + " " + nextSlot.isProposer());
					if (nextSlot.isProposer()) {
						activeSlotLimit = nextGroupSeqNr + 1;
						readySlots.add(nextSlot);
					}
					break;
				} else {
					break;
				}
			}

		} while (!readySlots.isEmpty());
	}

	private void readyAllSlots() {
		for (long instanceID = Math.max(0, windowStartID); instanceID < activeSlotLimit; instanceID++) {
			REFITOrderStageSlot slot = slots[(int) (instanceID % slots.length)];
			readySlots.add(slot);
		}
	}


	// ############################
	// # VIEW-CHANGE NOTIFICATION #
	// ############################

	private int viewID;

	public void ackStable(REFITAckReply ack) {
		ackSender.broadcast(ack);
	}

	private enum REFITTransitionProtocolState {
		VIEW_CHANGE_STARTED,
		VIEW_CHANGE_COMPLETABLE,
		VIEW_STABLE
	}

	private REFITTransitionProtocolState viewState;
	private int pendingViewChanges;
	private int nonCompletableViewChanges;
	private boolean viewIDUpdated = false;


	private void abort() {
		if (REFITConfig.MENCIUS_MODE) throw new UnsupportedOperationException("Not implemented for mencius");
		REFITLogger.logWarning(this, "abort state " + viewState + " next view " + viewID);
		if (viewState == REFITTransitionProtocolState.VIEW_CHANGE_STARTED) {
			REFITLogger.logWarning(this, "ignoring abort for view "
					+ viewID + " due to view state " + viewState);
			return;
		}
		transitionToView(viewID + 1);
		if (timingTracker != null) timingTracker.updateView(groupID, viewID, false);
	}

	public void transitionToView(int newViewID) {
		if (newViewID <= viewID) return;
		REFITLogger.logWarning(this, "new view " + newViewID);
		viewID = newViewID;
		viewState = REFITTransitionProtocolState.VIEW_CHANGE_STARTED;
		int ownSlots = 0;
		for (long instanceID = windowStartID; instanceID < activeSlotLimit; instanceID++) {
			REFITOrderStageSlot slot = slots[(int) (instanceID % slots.length)];
			ownSlots += slot.isSkipped() ? 0 : 1;
		}
		nonCompletableViewChanges = ownSlots;
		pendingViewChanges = ownSlots;
		viewIDUpdated = true;
		// Abort slots in order
		if (REFITLogger.LOG_ORDER) REFITLogger.logOrder(this, "update view of instances " + windowStartID
				+ " to " + (activeSlotLimit - 1) + " to view " + newViewID);
		for (long instanceID = windowStartID; instanceID < activeSlotLimit; instanceID++) {
			REFITOrderStageSlot slot = slots[(int) (instanceID % slots.length)];
			slot.applyNewView(viewID);
		}
	}

	public void learnCompletableViewChange() {
		nonCompletableViewChanges -= 1;
		if (nonCompletableViewChanges == 0) {
			viewState = REFITTransitionProtocolState.VIEW_CHANGE_COMPLETABLE;
		} else if (nonCompletableViewChanges < 0) {
			throw new InternalError("Too many completable view changes!");
		}
	}

	public void learnViewChange(REFITOrderProtocol protocol) {
		pendingViewChanges -= 1;
		if (pendingViewChanges == 0) {
			setupNewView(protocol);
		} else if (pendingViewChanges < 0) {
			throw new InternalError("Too many completed view changes!");
		}
	}

	private void setupNewView(REFITOrderProtocol protocol) {
		if (viewState == REFITTransitionProtocolState.VIEW_CHANGE_STARTED && nonCompletableViewChanges != 0) {
			throw new InternalError("Some view instances skipped the completable state " + nonCompletableViewChanges);
		}

		viewState = REFITTransitionProtocolState.VIEW_STABLE;
		viewIDUpdated = true;
		// TODO clean protocol handling

		short contactReplicaID = REFITOrderGroups.getGroupPrimary(limitedGroupID, viewID);
		REFITLogger.logEvent(this, "switch to " + protocol.getClass().getSimpleName() + " (view " + viewID + ")");
		REFITLogger.logEvent(this, contactReplicaID + " is now the contact replica for group " + groupID);

		// Reset next proposer id
		nextProposerID = windowStartID;

		// Clear outstanding proposals
		proposals.clear();

		// Distribute replica-role information
		if (groupID >= 0) {
			viewChangeSender.broadcast(new REFITConfigurationNotification(protocol, groupID, viewID));
		} else {
			// Hack to support mencius
			for (int i = 0; i < REFITOrderGroups.COUNT; i++) {
				viewChangeSender.broadcast(new REFITConfigurationNotification(protocol, i, viewID));
			}
		}

		if (timingTracker != null) timingTracker.updateView(groupID, viewID, true);
	}


	// #############
	// # PROPOSALS #
	// #############

	private final Queue<REFITRequest> proposals;
	private final LinkedList<REFITRequest> waitingProposals;
	private long nextProposerID;
	private long flushSlotsBeforeSeqNr = -REFITConfig.MAXIMUM_ORDER_INSTANCES_IN_PROGRESS;
	private long partialBatchingBeforeSeqNr = -REFITConfig.MAXIMUM_ORDER_INSTANCES_IN_PROGRESS;
	private long proposeOnlyBeforeSeqNr = Long.MAX_VALUE;


	private void handleProposal(REFITMessage proposal) {
		waitingProposals.add((REFITRequest) proposal);
	}

	private void moveReadyProposals() {
		Iterator<REFITRequest> iter = waitingProposals.iterator();
		while (iter.hasNext()) {
			REFITRequest proposal = iter.next();
			if (nextProposerID > proposal.executeAfter) {
				proposals.add(proposal);
				iter.remove();
			}
		}
	}

	private void assignProposals(boolean retry) {
		moveReadyProposals();

		if (REFITLogger.LOG_ORDER) REFITLogger.logOrder(this, "Trying to assign proposals, proposals "
				+ proposals.size() + " next " + nextProposerID + " window " + windowStartID);
		int minimumBatchSize = REFITConfig.MINIMUM_BATCH_SIZE;

		int active = 0;
		for (long instanceID = windowStartID; instanceID < nextProposerID; ++instanceID) {
			REFITOrderStageSlot slot = slots[(int) (instanceID % slots.length)];
			if (!slot.isComplete()) active++;
		}
		final int activeProposalsLimit = (REFITConfig.ACTIVE_PROPOSALS > 0) ? REFITConfig.ACTIVE_PROPOSALS : Integer.MAX_VALUE;

		// Find idle proposer instances
		for (long instanceID = nextProposerID; (instanceID < activeSlotLimit) && (instanceID < proposeOnlyBeforeSeqNr)
				&& (proposals.size() >= minimumBatchSize || instanceID < flushSlotsBeforeSeqNr)
				&& (active < activeProposalsLimit); instanceID++) {
			// Ignore slot if its current instance is not a proposer
			REFITOrderStageSlot slot = slots[(int) (instanceID % slots.length)];
			if (!slot.isProposer()) continue;
			if (REFITConfig.PREFER_FULL_BATCHES && !retry && instanceID >= flushSlotsBeforeSeqNr
					&& instanceID >= partialBatchingBeforeSeqNr && !fullBatchAvailable()) {
				if (!hasTimeout()) {
					setTimeout(2);
				}
				return;
			}

			// immediately update timing tracker with latest proposal
			if (timingTracker != null) timingTracker.receivedPrepare(instanceID, replica.id);
			slot.assignProposal(assembleRequestBatch(instanceID));
			// Assign proposal and trigger protocol processing
			readySlots.add(slot);
			active++;

			// Update proposer selection information
			nextProposerID = instanceID + 1;
		}

		// Update proposer selection information if not enough proposers have been available
		if (proposals.size() >= minimumBatchSize && active < activeProposalsLimit
				&& nextProposerID < proposeOnlyBeforeSeqNr) {
			nextProposerID = activeSlotLimit;
			if (REFITConfig.USE_SPINNING_LEADER) return;
			int complete = 0;
			for (REFITOrderStageSlot slot : slots) {
				if (slot.isComplete()) {
					complete++;
				}
			}
			REFITLogger.logWarning(this, "No free slots in the agreement window. completed "
					+ complete + "/" + slots.length);
		}
	}

	private boolean fullBatchAvailable() {
		if (proposals.size() >= REFITConfig.MAXIMUM_BATCH_SIZE[limitedGroupID]) {
			return true;
		}

		int batchBytes = 0;
		for (REFITRequest proposal : proposals) {
			batchBytes += proposal.getMessageSize();
			if (batchBytes >= REFITConfig.MAXIMUM_BATCH_BYTES) {
				return true;
			}
		}

		return false;
	}

	private REFITOrderRequest assembleRequestBatch(long instanceID) {
		// Check whether there are messages to be proposed
		if (proposals.isEmpty()) {
			return REFITNoOpRequest.INSTANCE;
		}

		int maximumBatchSize = REFITConfig.MAXIMUM_BATCH_SIZE[limitedGroupID];

		final REFITRequest head = proposals.peek();

		// Handle single proposals
		if ((proposals.size() == 1) || (maximumBatchSize == 1)
				|| (head.getMessageSize() >= REFITConfig.MAXIMUM_BATCH_BYTES)) {

			REFITRequest request = proposals.poll();
			if (REFITConfig.BATCH_ALWAYS) {
				LinkedList<REFITRequest> ll = new LinkedList<>();
				ll.add(request);
				return new REFITBatch(instanceID, replica.id, ll);
			}
			return request;
		}

		int batchBytes = 0;
		// Handle batch proposals
		Collection<REFITRequest> proposalBatch = new LinkedList<>();
		for (int i = 0; i < maximumBatchSize; i++) {
			final REFITRequest proposal = proposals.peek();
			if (proposal == null) {
				break;
			}

			// ensure that at least one message ends up in the batch
			// while limiting the max batch size
			if (i > 0 && batchBytes + proposal.getMessageSize() > REFITConfig.MAXIMUM_BATCH_BYTES) {
				break;
			}

			proposals.poll();
			proposalBatch.add(proposal);
			batchBytes += proposal.getMessageSize();
		}

		// avoid batching overhead if only a single message fits in the batch
		if (!REFITConfig.BATCH_ALWAYS && proposalBatch.size() == 1) {
			return proposalBatch.iterator().next();
		}

		return new REFITBatch(instanceID, replica.id, proposalBatch);
	}

	void prepareReceived(long instanceID, short from) {
		if (timingTracker != null) timingTracker.receivedPrepare(instanceID, from);
	}

	void commitReceived(long instanceID, short from) {
		if (timingTracker != null) timingTracker.receivedCommit(instanceID, from);
	}

	public void messageStable(REFITUniqueID uid) {
		for (Iterator<REFITRequest> iterator = proposals.iterator(); iterator.hasNext(); ) {
			REFITRequest proposal = iterator.next();
			if (!proposal.uid.equals(uid)) continue;
			iterator.remove();
			break;
		}
	}

}
