package refit.replica.order;

import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;

import refit.config.REFITConfig;
import refit.message.*;
import refit.message.REFITInstruction.REFITExecuteRequest;
import refit.replica.REFITBaseReplica;
import refit.replica.REFITObserver;


public class REFITOrderStageSlot implements Comparable<REFITOrderStageSlot> {

	protected final REFITBaseReplica replica;
	private final REFITObserver<REFITExecuteRequest> requestSender;
	private final REFITOrderStage orderStage;

	private REFITOrderProtocol currentProtocol;
	private REFITOrderProtocolInstance currentInstance;
	private REFITOrderProtocolTransitionInstance transitionInstance;
	private boolean skipped;
	private boolean activeAfterComplete;
	private long currentInstanceID;
	private int currentViewID;


	public REFITOrderStageSlot(REFITBaseReplica replica, REFITOrderStage orderStage, REFITOrderProtocol initialProtocol,
	                           REFITObserver<REFITExecuteRequest> requestSender) {
		// Initialize data structures
		this.replica = replica;
		this.requestSender = requestSender;
		this.orderStage = orderStage;
		this.messageStore = new REFITOrderStageSlotMessageStore();
		this.instanceMessages = new HashMap<>();
		this.currentProtocol = initialProtocol;
		this.currentInstanceID = -1L;
		this.currentViewID = REFITConfig.INITIAL_VIEW;
	}


	@Override
	public String toString() {
		return "ORDST[" + currentInstanceID + "-" + currentViewID + "]";
	}

	@Override
	public int compareTo(REFITOrderStageSlot slot) {
		return Long.compare(currentInstanceID, slot.currentInstanceID);
	}


	// ######################
	// # LIFE-CYCLE METHODS #
	// ######################

	public boolean init(long instanceID, boolean skip, boolean activeAfterComplete) {
		currentInstanceID = instanceID;
		skipped = skip;
		this.activeAfterComplete = activeAfterComplete;

		if (skip) {
			currentInstance = null;
			transitionInstance = null;
			return false;
		} else {
			// Garbage-collect buffered messages of skipped instances
			instanceMessages.keySet().removeIf(messageInstanceId -> messageInstanceId < instanceID);

			// Prepare message store for the new instance and demand instance execution if there are any buffered messages
			Collection<REFITOrderProtocolMessage> currentInstanceMessages = instanceMessages.remove(instanceID);
			messageStore.init(currentInstanceMessages);

			// Update current protocol instance
			currentInstance = currentProtocol.createInstance(this);
			currentInstance.init(instanceID, currentViewID);

			transitionInstance = currentProtocol.createTransitionInstance(this);
			transitionInstance.init(instanceID, currentViewID);
			return (currentInstanceMessages != null);
		}
	}

	public boolean isSkipped() {
		return skipped;
	}

	public boolean isComplete() {
		return skipped || currentInstance != null && currentInstance.isComplete();
	}

	public boolean isPrepared() {
		return currentInstance != null && currentInstance.isPrepared();
	}

	public void execute() {
		if (skipped || !activeAfterComplete && isComplete()) return;
		if (currentInstance != null) currentInstance.execute();
		transitionInstance.execute();
	}

	public void applyNewView(int viewID) {
		if (REFITConfig.ENABLE_DEBUG_CHECKS && viewID <= currentViewID) {
			throw new InternalError("Order stage has set a stale view " + viewID + " " + currentViewID);
		}
		currentViewID = viewID;
		if (skipped) return;
		if (currentInstance != null) {
			REFITOrderProtocolData protocolData = currentInstance.abort();
			transitionInstance.updateProtocolData(protocolData);
			currentInstance = null;
		}
		transitionInstance.updateView(viewID);
	}

	protected void synchronizeNewView(int viewID) {
		orderStage.transitionToView(viewID);
	}

	protected void learnCompletableViewChange() {
		orderStage.learnCompletableViewChange();
	}

	protected void learnViewChange(REFITOrderProtocol protocol, REFITOrderProtocolInstance successorInstance) {
		currentProtocol = protocol;
		orderStage.learnViewChange(protocol);

		currentInstance = successorInstance;
		// Invoke new instance
		execute();
	}

	// #####################
	// # PROPOSAL HANDLING #
	// #####################

	public boolean isProposer() {
		return currentInstance != null && currentInstance.isProposer();
	}

	public boolean hasProposal() {
		return currentInstance.hasProposal();
	}

	public void assignProposal(REFITOrderRequest message) {
		currentInstance.assignProposal(message);
	}

	public void prepareReceived(short from) {
		orderStage.prepareReceived(currentInstanceID, from);
	}

	public void commitReceived(short from) {
		orderStage.commitReceived(currentInstanceID, from);
	}

	public void ackStable(REFITAckReply ack) {
		orderStage.ackStable(ack);
	}

	public void messageStable(REFITUniqueID uid) {
		orderStage.messageStable(uid);
	}

	public void sendRequest(REFITExecuteRequest request) {
		requestSender.broadcast(request);
	}

	// ####################
	// # MESSAGE HANDLING #
	// ####################

	private final REFITOrderStageSlotMessageStore messageStore;
	private final Map<Long, Collection<REFITOrderProtocolMessage>> instanceMessages;


	public boolean insertMessage(REFITOrderProtocolMessage message) {
		// Ignore messages for old instances and old views
		if (message.uid.seqNr < currentInstanceID) return false;
		if (message.viewID < currentViewID) return false;

		// Buffer messages for future instances
		if (message.uid.seqNr != currentInstanceID) {
			instanceMessages.computeIfAbsent(message.uid.seqNr, k -> new LinkedList<>()).add(message);
			return false;
		}

		// Ignore message for current instance if the instance is already complete
		if (message.viewID == currentViewID && (skipped || !activeAfterComplete && isComplete())) return false;

		// Store message for current instance
		messageStore.add(message);
		return true;
	}

	protected REFITMessage fetchMessage(REFITMessageType type, int viewID) {
		return messageStore.remove(type, viewID);
	}


}
