package refit.replica.probe;

import java.util.*;

import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITMessage;
import refit.message.REFITRequest;
import refit.replica.REFITBaseReplica;
import refit.replica.REFITStage;
import refit.scheduler.REFITSchedulerTaskType;


public class REFITProbeStage extends REFITStage {

	private final REFITProbeSlot[] slots;
	private final HashSet<REFITProbeSlot> readySlots;
	private final TreeMap<Long, LinkedList<REFITProbeSlot>> slotTimeouts;


	public REFITProbeStage(REFITBaseReplica replica) {
		// Initialize data structures
		super(REFITSchedulerTaskType.PROBE_STAGE, replica);

		// Create protocol-instance slots
		this.slots = new REFITProbeSlot[REFITConfig.TOTAL_NR_OF_CLIENTS];
		for (int i = 0; i < slots.length; i++) {
			slots[i] = new REFITProbeSlot(replica, this);
		}
		this.readySlots = new HashSet<>();
		this.slotTimeouts = new TreeMap<>();
	}

	@Override
	public String toString() {
		return "PROBE";
	}

	protected void handleMessage(REFITMessage message) {
//		System.out.println("" + System.currentTimeMillis() + " got message " + message);
		switch (message.groupType()) {
		case PROBE_ORDER:
			// Forward message to the corresponding slot
			sendToSlot((REFITProbeOrderMessage) message);
			break;
		case REQUEST:
			assignRequest((REFITRequest) message);
			break;
		default:
			REFITLogger.logError(this, "Unexpected message " + message);
		}
	}

	private void sendToSlot(REFITProbeOrderMessage message) {
		REFITProbeSlot slot = safeGetSlot(message);
		boolean ready = slot.insertMessage(message);
		if (ready) {
			readySlots.add(slot);
		}
	}

	private REFITProbeSlot safeGetSlot(REFITMessage message) {
		short nodeID = message.uid.nodeID;
		return slots[nodeID];
	}

	private void assignRequest(REFITRequest request) {
		REFITProbeSlot slot = safeGetSlot(request);
		slot.setProposal(request);
		readySlots.add(slot);
	}

	public void addSlotTimeout(REFITProbeSlot slot, long timeout) {
		if (!slotTimeouts.containsKey(timeout)) {
			slotTimeouts.put(timeout, new LinkedList<>());
		}
		LinkedList<REFITProbeSlot> slotList = slotTimeouts.get(timeout);
		slotList.add(slot);
	}

	protected void stageComplete() {
		long currentTime = System.currentTimeMillis();
		Iterator<Map.Entry<Long, LinkedList<REFITProbeSlot>>> it = slotTimeouts.entrySet().iterator();
		while (it.hasNext()) {
			Map.Entry<Long, LinkedList<REFITProbeSlot>> timeouts = it.next();
			if (timeouts.getKey() <= currentTime) {
				readySlots.addAll(timeouts.getValue());
				it.remove();
			} else {
				break;
			}
		}

		if (!slotTimeouts.isEmpty()) {
			long timeout = slotTimeouts.firstKey() - System.currentTimeMillis();
			setTimeout(timeout);
		}

		for (REFITProbeSlot readySlot : readySlots) readySlot.execute();

		// Reset ready slots
		readySlots.clear();
	}
}
