package refit.replica.stewardorder;

import java.nio.ByteBuffer;

import refit.config.REFITLogger;
import refit.message.REFITMessageAuthentication;
import refit.replica.stewardorder.REFITStewardOrderMessages.REFITStewardAccept;
import refit.replica.stewardorder.REFITStewardOrderMessages.REFITStewardProposal;
import refit.util.REFITBallotBox;

public class REFITStewardAcceptCertificate {

	public enum AddResult {
		INVALID,
		DUPLICATE,
		ADDED
	}

	private final int leaderGroupID;
	private int COMMITTED_THRESHOLD;

	private long instanceID;
	private int viewID;

	public REFITStewardProposal prePrepare;
	private REFITBallotBox<Short, ByteBuffer, REFITStewardAccept> commits;

	public REFITStewardAcceptCertificate(int leaderGroupID, int siteCount) {
		this.leaderGroupID = leaderGroupID;
		// votes: 1 propsal + floor(sites/2) accepts from other sites
		COMMITTED_THRESHOLD = (siteCount / 2);
		this.commits = new REFITBallotBox<>(COMMITTED_THRESHOLD);
	}


	@Override
	public String toString() {
		return "{" + viewID + " " + instanceID + ": " + isCommitted() + "}";
	}

	public void init(long instanceID, int viewID) {
		this.instanceID = instanceID;
		this.viewID = viewID;
		this.prePrepare = null;
		this.committed = false;

		commits.clear();
	}


	// ###################
	// # ADDING MESSAGES #
	// ###################

	public boolean addPrePrepare(REFITStewardProposal prePrepare, REFITMessageAuthentication mac) {
		if (prePrepare == null) return false;
		if (this.prePrepare != null) return false;
		if (!mac.verifyNullMAC(prePrepare)) return false;
		if (REFITStewardGroups.stewardOrderGroupID(prePrepare.from) != leaderGroupID) return false;
		// already check by proposalcertificate
//		if (!mac.verifyCombined(leaderGroupID, prePrepare.calcShortHash(), prePrepare.thressig)) {
//			REFITLogger.logWarning(this, "Invalid thressig on " + prePrepare);
//			return false;
//		}
		this.prePrepare = prePrepare;
		checkCommitted();

		return true;
	}

	public AddResult addCommit(REFITStewardAccept commit, REFITMessageAuthentication mac) {
		if (commit == null) return AddResult.INVALID;
		if (!mac.verifyNullMAC(commit)) return AddResult.INVALID;
		short gid = (short) REFITStewardGroups.stewardOrderGroupID(commit.from);
		// the leader group has already voted via the prepare
		// this restriction does not apply when newer view are in play
		if (gid == leaderGroupID) return AddResult.DUPLICATE;
		if (commits.hasVoted(gid)) return AddResult.DUPLICATE;
		// skip signature check for internal message
		if (!commit.isInternal() && !mac.verifyCombined(gid, commit.prepareHash, commit.thressig)) {
			REFITLogger.logWarning(this, "Invalid thressig on " + commit);
			return AddResult.INVALID;
		}
		commits.add(gid, ByteBuffer.wrap(commit.prepareHash), commit);

		checkCommitted();
		return AddResult.ADDED;
	}


	// ######################
	// # CERTIFICATE STATES #
	// ######################

	private boolean committed;

	public boolean isCommitted() {
		return committed;
	}

	private void checkCommitted() {
		if (committed) return;
		if (prePrepare == null) return;

		ByteBuffer decision = commits.getDecision();
		if (decision == null) return;
		// In the current implementation, a non-faulty replica that
		// has been provided with a faulty PRE_PREPARE will never
		// reach the status "committed" in the corresponding instance,
		// even if it is able to obtain 2f+1 correct COMMITs; see the
		// following code line. This is not a safety problem but
		// implementing a solution (which would require obtaining
		// the correct message) would lead to a more robust protocol.
		byte[] shortHash = prePrepare.calcShortHash();
		if (!decision.equals(ByteBuffer.wrap(shortHash))) return;

		committed = true;
	}
}
