package refit.replica.stewardorder;

import java.math.BigInteger;
import java.nio.ByteBuffer;
import java.security.MessageDigest;

import refit.message.REFITMessage;
import refit.message.REFITMessageType;
import refit.message.REFITNoOpRequest;
import refit.message.REFITOrderRequest;
import refit.replica.order.REFITOrderProtocolMessage;
import refit.util.REFITBigIntHelper;
import refit.util.REFITPayload;
import threshsig.SigShare;
import threshsig.Verifier;


public class REFITStewardOrderMessages {

	// #####################
	// # ABSTRACT MESSAGES #
	// #####################

	public static abstract class REFITStewardMessage extends REFITOrderProtocolMessage {

		public static final int STEWARD_HEADER_SIZE = Integer.BYTES;

		protected ByteBuffer command;


		protected REFITStewardMessage(REFITMessageType pbftType, long instanceID, int viewID, ByteBuffer command, short from) {
			super(pbftType, instanceID, viewID, from);
			this.command = command;
		}

		protected REFITStewardMessage(REFITMessageType expectedType, ByteBuffer buffer) {
			super(expectedType, buffer);
			int cmdSize = buffer.getInt();
			this.command = buffer.slice();
			this.command.limit(cmdSize);
			buffer.position(buffer.position() + cmdSize);
		}

		@Override
		public REFITMessageType groupType() {
			return REFITMessageType.STEWARD_WAN;
		}

		@Override
		public AuthenticationType getAuthenticationType() {
			return AuthenticationType.SIGNATURE;
		}

		@Override
		protected void serialize(ByteBuffer buffer) {
			super.serialize(buffer);
			serializeCommand(buffer);
		}

		protected void serializeCommand(final ByteBuffer buffer) {
			int cmdSize = command.remaining();
			buffer.putInt(cmdSize);
			// replace command ByteBuffer after serialization to store the data only once
			ByteBuffer newCommand = buffer.slice();
			buffer.put(command);
			newCommand.limit(cmdSize);
			command = newCommand;
		}

		public ByteBuffer getCommand() {
			return command.slice();
		}

		@Override
		protected int getHeaderSize() {
			return (super.getHeaderSize() + STEWARD_HEADER_SIZE);
		}

		@Override
		protected int calculatePayloadSize() {
			return (super.calculatePayloadSize() + STEWARD_HEADER_SIZE + calculateCommandSize());
		}

		protected int calculateCommandSize() {
			return command.remaining();
		}

		@Override
		public String toString() {
			return String.format("{%s|%d|%d|%d}", type, from, uid.seqNr, viewID);
		}

		protected void putBigInt(ByteBuffer buffer, BigInteger bigint) {
			REFITBigIntHelper.putBigInt(buffer, bigint);
		}

		protected BigInteger getBigInt(ByteBuffer buffer) {
			return REFITBigIntHelper.getBigInt(buffer);
		}

		protected int getLen(BigInteger bigint) {
			return REFITBigIntHelper.getLen(bigint);
		}
	}

	// #####################
	// # SPECIFIC MESSAGES #
	// #####################

	public static class REFITStewardProposal extends REFITStewardMessage {

		public final transient REFITOrderRequest message;
		public final BigInteger thressig;
		public static final int SHORT_HASH_LENGTH = REFITPayload.HASH_SIZE + Integer.BYTES + Integer.BYTES + Long.BYTES;


		public REFITStewardProposal(long instanceID, int viewID, short from, REFITOrderRequest message, BigInteger thressig) {
			super(REFITMessageType.STEWARD_PROPOSAL, instanceID, viewID, null, from);
			this.message = message;
			this.thressig = thressig;
		}

		public REFITStewardProposal(ByteBuffer buffer) {
			super(REFITMessageType.STEWARD_PROPOSAL, buffer);
			// create buffer slice to keep the invariant that the ByteBuffer passed
			// to a message constructor must be at position zero
			ByteBuffer cmd = getCommand();
			if (cmd.hasRemaining()) {
				this.message = (REFITOrderRequest) REFITMessage.createMessage(cmd);
			} else {
				this.message = REFITNoOpRequest.INSTANCE;
			}
			thressig = getBigInt(buffer);
		}

		@Override
		public AuthenticationType getAuthenticationType() {
			return AuthenticationType.NONE;
		}

		@Override
		public String toString() {
			return String.format("{%s|%d|%d|%d|%s}", type, from, uid.seqNr, viewID, message);
		}

		@Override
		protected void serializeCommand(final ByteBuffer buffer) {
			if (message.type == REFITMessageType.REQUEST_NO_OP) {
				command = ByteBuffer.allocate(0);
				return;
			}

			int cmdSize = message.calculateMessageSize();
			buffer.putInt(cmdSize);
			// replace command ByteBuffer after serialization to store the data only once
			ByteBuffer newCommand = buffer.slice();
			message.serializeInto(buffer);
			newCommand.limit(cmdSize);
			command = newCommand;
			putBigInt(buffer, thressig);
		}

		@Override
		protected int calculateCommandSize() {
			if (message.type == REFITMessageType.REQUEST_NO_OP) {
				return 0;
			} else {
				return message.calculateMessageSize();
			}
		}

		@Override
		protected int calculatePayloadSize() {
			return super.calculatePayloadSize() + getLen(thressig);
		}

		@Override
		protected byte[] calculateHash() {
			// Prepare message header
			ByteBuffer messageHeader = getInternalBuffer();
			messageHeader.limit(getHeaderSize());

			// Prepare message (must be done in advance in order to prevent the hash from
			// being computed during the calculation of the REFITStewardPrePrepare hash)
			byte[] messageHash = message.getHash();

			// Calculate hash
			MessageDigest digest = REFITPayload.getDigest();
			digest.update(messageHeader);
			digest.update(messageHash);
			return digest.digest();
		}

		public byte[] calcShortHash() {
			ByteBuffer buffer = ByteBuffer.allocate(SHORT_HASH_LENGTH);
			buffer.putInt(0);
			buffer.putInt(viewID);
			buffer.putLong(uid.seqNr);
			buffer.put(message.getHash());
			return buffer.array();
		}
	}

	public static class REFITStewardAcceptShare extends REFITStewardMessage {

		public final SigShare share;
		public final byte[] prepareHash;

		public REFITStewardAcceptShare(REFITStewardProposal prePrepare, short from, SigShare share) {
			super(REFITMessageType.STEWARD_ACCEPT_SHARE, prePrepare.uid.seqNr, prePrepare.viewID,
					null, from);
			this.share = share;
			this.prepareHash = prePrepare.calcShortHash();
		}

		public REFITStewardAcceptShare(ByteBuffer buffer) {
			super(REFITMessageType.STEWARD_ACCEPT_SHARE, buffer);
			// create buffer slice to keep the invariant that the ByteBuffer passed
			// to a message constructor must be at position zero
			ByteBuffer cmd = getCommand();
			prepareHash = new byte[REFITStewardProposal.SHORT_HASH_LENGTH];
			cmd.get(prepareHash);

			int shareId = cmd.getInt();
			BigInteger sig = getBigInt(cmd);
			BigInteger z = getBigInt(cmd);
			BigInteger c = getBigInt(cmd);

			// FIXME: should reconstruct share and group verifier using the signers info
			Verifier verif = new Verifier(z, c, null, null);
			share = new SigShare(shareId, sig, verif);
		}

		@Override
		protected void serializeCommand(ByteBuffer buffer) {
			int cmdSize = calculateCommandSize();
			buffer.putInt(cmdSize);
			// replace command ByteBuffer after serialization to store the data only once
			ByteBuffer newCommand = buffer.slice();
			buffer.put(prepareHash);

			buffer.putInt(share.getId());
			putBigInt(buffer, share.getSig());

			Verifier verif = share.getSigVerifier();
			putBigInt(buffer, verif.getZ());
			putBigInt(buffer, verif.getC());

			newCommand.limit(cmdSize);
			command = newCommand;
		}

		@Override
		protected int calculateCommandSize() {
			Verifier verif = share.getSigVerifier();
			return prepareHash.length + Integer.BYTES + getLen(share.getSig()) + getLen(verif.getZ()) + getLen(verif.getC());
		}
	}


	public static class REFITStewardAccept extends REFITStewardMessage {

		public final byte[] prepareHash;
		public final BigInteger thressig;

		public REFITStewardAccept(REFITStewardProposal prePrepare, short from, BigInteger thressig) {
			super(REFITMessageType.STEWARD_ACCEPT, prePrepare.uid.seqNr, prePrepare.viewID, null, from);
			this.thressig = thressig;
			this.prepareHash = prePrepare.calcShortHash();
		}

		public REFITStewardAccept(ByteBuffer buffer) {
			super(REFITMessageType.STEWARD_ACCEPT, buffer);
			// create buffer slice to keep the invariant that the ByteBuffer passed
			// to a message constructor must be at position zero
			ByteBuffer cmd = getCommand();
			prepareHash = new byte[REFITStewardProposal.SHORT_HASH_LENGTH];
			cmd.get(prepareHash);
			thressig = getBigInt(buffer);
		}

		@Override
		public AuthenticationType getAuthenticationType() {
			return AuthenticationType.NONE;
		}

		@Override
		public String toString() {
			return String.format("{%s|%d|%d|%d|%s}", type, from, uid.seqNr, viewID, REFITPayload.toString(prepareHash));
		}

		@Override
		protected void serializeCommand(final ByteBuffer buffer) {
			int cmdSize = calculateCommandSize();
			buffer.putInt(cmdSize);
			// replace command ByteBuffer after serialization to store the data only once
			ByteBuffer newCommand = buffer.slice();
			buffer.put(prepareHash);

			newCommand.limit(cmdSize);
			command = newCommand;

			putBigInt(buffer, thressig);
		}

		@Override
		protected int calculateCommandSize() {
			return prepareHash.length;
		}

		@Override
		protected int calculatePayloadSize() {
			return super.calculatePayloadSize() + getLen(thressig);
		}
	}
}
