package refit.replica.stewardorder;

import java.math.BigInteger;
import java.nio.ByteBuffer;
import java.util.List;

import refit.config.REFITConfig;
import refit.config.REFITLogger;
import refit.message.REFITMessageAuthentication;
import refit.replica.stewardorder.REFITStewardOrderMessages.REFITStewardAcceptShare;
import refit.replica.stewardorder.REFITStewardOrderMessages.REFITStewardProposal;
import refit.util.REFITBallotBox;
import threshsig.SigShare;
import threshsig.ThresholdSigException;

public class REFITStewardProposalCertificate {

	public enum AddResult {
		INVALID,
		DUPLICATE,
		ADDED
	}

	private final int leaderGroupID;
	private final int localGroupID;
	private int COMMITTED_THRESHOLD;
	private boolean[] localGroup;

	private long instanceID;
	private int viewID;

	public REFITStewardProposal prePrepare;
	private REFITBallotBox<Short, ByteBuffer, REFITStewardAcceptShare> commits;

	private BigInteger combinedSig;

	public REFITStewardProposalCertificate(int leaderGroupID, int localGroupID, int groupSize) {
		this.leaderGroupID = leaderGroupID;
		this.localGroupID = localGroupID;
		COMMITTED_THRESHOLD = (groupSize + REFITConfig.FAULTS_TO_TOLERATE + 1 + 1) / 2;
		this.localGroup = REFITStewardGroups.stewardOrderGroupByID(localGroupID);

		if (groupSize < COMMITTED_THRESHOLD) {
			throw new AssertionError(
					String.format("Group size %d is too small for the expected number of faults %d",
							groupSize, REFITConfig.FAULTS_TO_TOLERATE));
		}
		this.commits = new REFITBallotBox<>(COMMITTED_THRESHOLD);
	}


	@Override
	public String toString() {
		return "{" + viewID + " " + instanceID + ": " + isCommitted() + "}";
	}

	public void init(long instanceID, int viewID) {
		this.instanceID = instanceID;
		this.viewID = viewID;
		this.prePrepare = null;
		this.committed = false;
		this.combinedSig = null;

		commits.clear();
	}


	// ###################
	// # ADDING MESSAGES #
	// ###################

	public boolean addPrePrepare(REFITStewardProposal prePrepare, REFITMessageAuthentication mac) {
		if (prePrepare == null)
			return false;
		if (this.prePrepare != null)
			return false;
		if (!mac.verifyNullMAC(prePrepare))
			return false;
		// skip signature check for internal message
		if (!prePrepare.isInternal() && !mac.verifyCombined(leaderGroupID, prePrepare.calcShortHash(), prePrepare.thressig)) {
			REFITLogger.logWarning(this, "Invalid thressig on " + prePrepare);
			return false;
		}
		this.prePrepare = prePrepare;

		return true;
	}

	public AddResult addCommit(REFITStewardAcceptShare commit, REFITMessageAuthentication mac) {
		if (commit == null) return AddResult.INVALID;
		if (commits.hasVoted(commit.from)) return AddResult.DUPLICATE;
		if (!mac.verifySignature(commit, localGroup)) return AddResult.INVALID;
		commits.add(commit.from, ByteBuffer.wrap(commit.prepareHash), commit);

		checkCommitted(mac);
		return AddResult.ADDED;
	}


	// ######################
	// # CERTIFICATE STATES #
	// ######################

	private boolean committed;

	public boolean isCommitted() {
		return committed;
	}

	private void checkCommitted(REFITMessageAuthentication mac) {
		if (committed) return;
		if (prePrepare == null) return;

		ByteBuffer decision = commits.getDecision();
		if (decision == null) return;
		// In the current implementation, a non-faulty replica that
		// has been provided with a faulty PRE_PREPARE will never
		// reach the status "committed" in the corresponding instance,
		// even if it is able to obtain 2f+1 correct COMMITs; see the
		// following code line. This is not a safety problem but
		// implementing a solution (which would require obtaining
		// the correct message) would lead to a more robust protocol.
		byte[] shortHash = prePrepare.calcShortHash();
		if (!decision.equals(ByteBuffer.wrap(shortHash))) return;

		SigShare[] sigs = new SigShare[COMMITTED_THRESHOLD];
		List<REFITStewardAcceptShare> desc = commits.getDecidingBallots();
		for (int i = 0; i < sigs.length; i++) {
			sigs[i] = desc.get(i).share;
		}

		try {
			combinedSig = mac.combineShares(localGroupID, shortHash, sigs);
		} catch (ThresholdSigException e) {
			REFITLogger.logError(this, "Failed to combine signature shares");
			// FIXME: implement handling of corrupted signatures
			return;
		}

		committed = true;
	}

	public BigInteger getCombinedSig() {
		return combinedSig;
	}

}
