package refit.replica.update;

import java.nio.ByteBuffer;
import java.security.MessageDigest;
import java.util.Arrays;
import java.util.Collection;
import java.util.LinkedList;
import java.util.Objects;

import refit.message.REFITMessageType;
import refit.message.REFITUniqueID;
import refit.util.REFITPayload;

public class REFITBatchUpdate extends REFITUpdateMessage {

	public final LinkedList<REFITUpdate> updates;

	public REFITBatchUpdate(long agreementSeqNr, short from, Collection<REFITUpdate> updates) {
		super(REFITMessageType.BATCH_UPDATE, new REFITUniqueID((short) -1, agreementSeqNr), from);
		this.updates = new LinkedList<>(updates);
		if (uid.seqNr != this.updates.getFirst().agreementSeqNr()) {
			throw new IllegalStateException("BatchUpdate got wrong seqNr");
		}
	}

	protected REFITBatchUpdate(REFITUniqueID uid, short from, byte[] payloadHash) {
		super(REFITMessageType.BATCH_UPDATE, uid, from);
		updates = null;
		this.payloadHash = payloadHash;
	}

	public REFITBatchUpdate(ByteBuffer buffer) {
		super(REFITMessageType.BATCH_UPDATE, buffer);
		if (buffer.get() == 0) {
			this.updates = null;
			this.payloadHash = new byte[REFITPayload.HASH_SIZE];
			buffer.get(payloadHash);
			return;
		}
		this.updates = new LinkedList<>();
		short batchSize = buffer.getShort();
		for (int i = 0; i < batchSize; i++) {
			REFITUpdate update = new REFITUpdate(buffer.slice());
			updates.add(update);
			buffer.position(buffer.position() + update.getMessageSize());
		}
		if (uid.seqNr != this.updates.getFirst().agreementSeqNr()) {
			throw new IllegalStateException("BatchUpdate got wrong seqNr");
		}
	}

	@Override
	public AuthenticationType getAuthenticationType() {
		return AuthenticationType.MAC_AUTHENTICATOR;
	}

	@Override
	public REFITBatchUpdate toHashedUpdate() {
		return new REFITBatchUpdate(uid, from, getPayloadHash());
	}

	@Override
	public long agreementSeqNr() {
		return uid.seqNr;
	}

	public long getMaxAgreementSeqNr() {
		if (updates == null) return -1;
		return updates.getLast().agreementSeqNr();
	}

	@Override
	public boolean isFullUpdate() {
		return updates != null;
	}

	@Override
	public String toString() {
		return String.format("{%s|%d|%s|%d}", type, from, uid, (updates != null) ? updates.size() : -1);
	}

	@Override
	public boolean equals(Object o) {
		if (this == o) return true;
		if (o == null || getClass() != o.getClass()) return false;
		if (!super.equals(o)) return false;
		REFITBatchUpdate that = (REFITBatchUpdate) o;
		return Objects.equals(updates, that.updates) &&
				Arrays.equals(payloadHash, that.payloadHash);
	}

	@Override
	public int hashCode() {
		int result = Objects.hash(super.hashCode(), updates);
		result = 31 * result + Arrays.hashCode(payloadHash);
		return result;
	}

	// #################
	// # SERIALIZATION #
	// #################

	@Override
	protected void serialize(ByteBuffer buffer) {
		super.serialize(buffer);
		buffer.put((byte) (isFullUpdate() ? 1 : 0));
		if (isFullUpdate()) {
			buffer.putShort((short) updates.size());
			for (REFITUpdate update : updates) buffer.put(update.getBuffer());
		} else {
			buffer.put(payloadHash);
		}
	}

	@Override
	protected int getHeaderSize() {
		return super.getHeaderSize() + Byte.BYTES;
	}

	@Override
	protected int calculatePayloadSize() {
		int size = super.calculatePayloadSize() + Byte.BYTES;
		if (!isFullUpdate()) {
			return size + REFITPayload.HASH_SIZE;
		}
		size += Short.BYTES;
		for (REFITUpdate update : updates) size += update.getMessageSize();
		return size;
	}

	// ###########
	// # HASHING #
	// ###########

	private byte[] payloadHash;

	@Override
	protected byte[] calculateHash() {
		// Prepare message header
		ByteBuffer messageHeader = getBuffer();
		messageHeader.limit(getHeaderSize());

		getPayloadHash();

		// Calculate hash
		MessageDigest digest = REFITPayload.getDigest();
		digest.update(messageHeader);
		digest.update(getPayloadHash());
		return digest.digest();
	}

	@Override
	public byte[] getPayloadHash() {
		if (payloadHash == null) {
			// Make sure that update hashes have been calculated
			for (REFITUpdate update : updates) update.getPayloadHash();

			// Calculate payload hash
			MessageDigest digest = REFITPayload.getDigest();
			for (REFITUpdate update : updates) digest.update(update.getPayloadHash());
			payloadHash = digest.digest();
		}
		return payloadHash;
	}
}
