package refit.replica.update;

import java.nio.ByteBuffer;
import java.security.MessageDigest;
import java.util.Arrays;

import refit.message.REFITMessageType;
import refit.message.REFITUniqueID;
import refit.util.REFITPayload;

public class REFITUpdate extends REFITUpdateMessage {

	public static final int UPDATE_HEADER_SIZE = Byte.BYTES + Long.BYTES + Short.BYTES;

	private final ByteBuffer payload;
	private final boolean isFullUpdate;
	public final boolean isSystem;
	private final long agreementSeqNr;
	public final byte[] replyHash;


	public REFITUpdate(REFITUniqueID uid, short from, long agreementSeqNr, ByteBuffer payload, byte[] replyHash, boolean isSystem) {
		this(uid, from, agreementSeqNr, payload, replyHash, true, isSystem);
	}

	private REFITUpdate(REFITUniqueID uid, short from, long agreementSeqNr, ByteBuffer payload, byte[] replyHash, boolean isFullUpdate, boolean isSystem) {
		super(REFITMessageType.UPDATE, uid, from);
		this.agreementSeqNr = agreementSeqNr;
		this.payload = payload;
		this.replyHash = replyHash;
		this.isFullUpdate = isFullUpdate;
		this.isSystem = isSystem;
		if (payload == null) {
			throw new AssertionError("Missing payload");
		}
		if (isFullUpdate && replyHash == null) {
			throw new AssertionError("Missing reply");
		}
		if (replyHash != null && replyHash.length != REFITPayload.HASH_SIZE) {
			throw new AssertionError("Reply hash has wrong length!");
		}
	}

	public REFITUpdate(ByteBuffer buffer) {
		super(REFITMessageType.UPDATE, buffer);
		byte flags = buffer.get();
		this.isFullUpdate = (flags & 1) != 0;
		this.isSystem = (flags & 2) != 0;
		this.agreementSeqNr = buffer.getLong();
		short payloadLen = buffer.getShort();

		this.payload = buffer.slice();
		this.payload.limit(payloadLen);
		buffer.position(buffer.position() + payloadLen);

		if (isFullUpdate) {
			this.replyHash = new byte[REFITPayload.HASH_SIZE];
			buffer.get(replyHash);
		} else {
			this.replyHash = null;
		}
	}

	@Override
	public AuthenticationType getAuthenticationType() {
		return AuthenticationType.MAC_AUTHENTICATOR;
	}

	@Override
	public REFITUpdate toHashedUpdate() {
		return new REFITUpdate(uid, from, agreementSeqNr, ByteBuffer.wrap(getPayloadHash()), null, false, isSystem);
	}

	public ByteBuffer getPayload() {
		return payload.slice();
	}

	@Override
	public boolean isFullUpdate() {
		return isFullUpdate;
	}

	@Override
	public long agreementSeqNr() {
		return agreementSeqNr;
	}

	@Override
	public String toString() {
		return String.format("{%s|%d|%s|%d|%s|%d}", type, from, uid, agreementSeqNr, isFullUpdate, payload.remaining());
	}

	@Override
	public boolean equals(Object o) {
		if (this == o) return true;
		if (o == null || getClass() != o.getClass()) return false;
		if (!super.equals(o)) return false;
		REFITUpdate that = (REFITUpdate) o;
		return isFullUpdate == that.isFullUpdate &&
				agreementSeqNr == that.agreementSeqNr &&
				payload.equals(that.payload) &&
				Arrays.equals(replyHash, that.replyHash) &&
				Arrays.equals(payloadHash, that.payloadHash);
	}

	@Override
	protected void serialize(ByteBuffer buffer) {
		super.serialize(buffer);
		buffer.put((byte) ((isFullUpdate ? 1 : 0) | (isSystem ? 2 : 0)));
		buffer.putLong(agreementSeqNr);
		buffer.putShort((short) payload.remaining());
		payload.mark();
		buffer.put(payload);
		payload.reset();
		if (isFullUpdate) {
			buffer.put(replyHash);
		}
	}

	@Override
	protected int getHeaderSize() {
		return super.getHeaderSize() + UPDATE_HEADER_SIZE;
	}

	@Override
	protected int calculatePayloadSize() {
		return super.calculatePayloadSize() + UPDATE_HEADER_SIZE + payload.remaining() + ((isFullUpdate) ? replyHash.length : 0);
	}


	// ###########
	// # HASHING #
	// ###########

	private transient byte[] payloadHash;


	@Override
	protected byte[] calculateHash() {
		// Prepare message header
		ByteBuffer messageHeader = getBuffer();
		messageHeader.limit(getHeaderSize());

		// Make sure that payload hash has been calculated (must be ensured in advance)
		getPayloadHash();

		// Calculate hash
		MessageDigest digest = REFITPayload.getDigest();
		digest.update(messageHeader);
		digest.update(payloadHash);
		return digest.digest();
	}

	@Override
	public byte[] getPayloadHash() {
		if (payloadHash == null) {
			if (isFullUpdate) {
				MessageDigest hashFunction = REFITPayload.getDigest();

				// enforce binding to uid
				ByteBuffer buf = ByteBuffer.wrap(new byte[Short.BYTES + Long.BYTES]);
				buf.putShort(uid.nodeID);
				buf.putLong(uid.seqNr);
				buf.flip();
				hashFunction.update(buf);

				payload.mark();
				hashFunction.update(payload);
				payload.reset();
				hashFunction.update(replyHash);
				payloadHash = hashFunction.digest();
			} else {
				payloadHash = new byte[payload.remaining()];
				payload.mark();
				payload.get(payloadHash);
				payload.reset();
			}
		}
		return payloadHash;
	}

}
