package refit.scheduler;

import java.io.IOException;
import java.nio.channels.SelectableChannel;
import java.nio.channels.SelectionKey;

import refit.util.REFITAssert;
import refit.util.REFITTime;


public abstract class REFITSchedulerTask {

	public final REFITSchedulerTaskType taskType;
	private short taskID;

	private REFITScheduler scheduler;
	private long timeout;


	public REFITSchedulerTask(REFITSchedulerTaskType taskType, REFITSchedulerGroup group) {
		REFITAssert.assertTrue(taskType != null, "Missing task type");
		this.taskType = taskType;
		this.taskID = -1;
		clearTimeout();
		this.throughput = 0L;
		group.registerTask(this);
	}


	// ################
	// # TASK METHODS #
	// ################

	public void setTaskID(short id) {
		if (this.taskID != -1) {
			throw new IllegalStateException("May set id only once");
		}
		this.taskID = id;
	}

	public final short taskID() {
		return this.taskID;
	}

	public void setScheduler(REFITScheduler scheduler) {
		this.scheduler = scheduler;
		if (hasTimeout()) {
			scheduler.notifySetTimeout(this, timeout);
		}
	}

	public SelectionKey registerIO(SelectableChannel channel, int operations) throws IOException {
		return scheduler.registerIOTask(this, channel, operations);
	}

	public void deregisterIO(SelectableChannel channel) {
		scheduler.deregisterIOTask(channel);
	}

	public boolean sameScheduler(REFITSchedulerTask other) {
		return scheduler == other.scheduler;
	}

	public void selectKey(SelectionKey key) {
		throw new IllegalStateException("This scheduler task does not own a SelectionKey " + this);
	}

	protected void progress() {
		if (scheduler != null) scheduler.notifyProgress(this);
	}

	protected void clearTimeout() {
		// doesn't cleanup the notification sent to the scheduler
		timeout = -1L;
	}

	// replaces the previous timeout
	protected void setTimeout(long delay) {
		setTimeoutAbsolute(delay + REFITTime.currentTimeMillis.getAsLong());
	}

	// replaces the previous timeout
	protected void setTimeoutAbsolute(long newTimeout) {
		timeout = newTimeout;
		if (scheduler != null) scheduler.notifySetTimeout(this, timeout);
	}

	protected boolean hasTimeout() {
		return timeout != -1L;
	}

	public void init() {
		// Override in sub classes if necessary
	}

	public void work(boolean allowTimeout) {
		if (allowTimeout && hasTimeout() && timeout <= REFITTime.currentTimeMillis.getAsLong()) {
			clearTimeout();
			timeout();
		}
		execute();
	}

	public abstract void execute();

	public void timeout() {
		// Override in sub classes if necessary
	}


	// ##############
	// # STATISTICS #
	// ##############

	private long throughput;


	protected void event() {
		throughput++;
	}

	public long collectThroughput() {
		long tmp = throughput;
		throughput = 0L;
		return tmp;
	}

}
