package refit.test;

import java.nio.ByteBuffer;
import java.util.Queue;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.LockSupport;

public class QueueMicrobenchmark {
	public static void main(String[] args) {
		final Queue<ByteBuffer[]> dataQueue = new ConcurrentLinkedQueue<>();
		// Up to a Batch size of 5 the speed up is quite noticable
		// afterwards it start to slow down -> Use at least a bit more
		// Too large batchSize (depends on server/java) cause the creation of much more object which
		// seem to slow down garbage collection / instance creation (-> GC nursery)
		final int batchSize = 20;
		final int dataAmount = 200000000 / batchSize;
		final AtomicInteger queueSize = new AtomicInteger();

		Thread producer = new Thread(new Runnable() {
			@Override
			public void run() {
				for (int counter = 0; counter < dataAmount; counter++) {
					while (queueSize.get() > Math.max(5, 10000 / 256 / batchSize)) {
						LockSupport.parkNanos(10000);
					}

					ByteBuffer[] datas = new ByteBuffer[batchSize];
					for (int i = 0; i < datas.length; i++) {
						ByteBuffer data = ByteBuffer.allocate(4);
						data.putInt(counter);
						data.flip();
						datas[i] = data;
					}
					dataQueue.add(datas);
					// reduce counter contention, especially beneficial for small batch sizes
					// f.e. > 20% speedup for batch = 5
					if ((counter & 0xff) == 0) {
						queueSize.incrementAndGet();
					}
				}
				// kill byte
				ByteBuffer[] eof = new ByteBuffer[0];
				dataQueue.add(eof);
			}
		});

		Thread consumer = new Thread(new Runnable() {
			@Override
			public void run() {
				// Just make sure that all of this can't be optimized away
				long sum = 0;
				long counter = 0;
				while (true) {
					ByteBuffer[] buffer = dataQueue.poll();
					if (buffer == null) {
						// > 30% speedup
						// seems to reduce congestion by a large amount
						LockSupport.parkNanos(10000);
					} else if (buffer.length == 0) {
						break;
					} else {
						if ((counter & 0xff) == 0) {
							queueSize.decrementAndGet();
						}
						for (ByteBuffer bb : buffer) {
							int value = bb.getInt();
							sum += value;
						}
						counter++;
					}
				}

				long expected = (long) dataAmount * (dataAmount - 1) / 2 * batchSize;
				if (expected != sum) {
					System.err.println("Data corruption: got " + sum + " instead of " + expected);
				}
			}
		});

		long startTime = System.nanoTime();
		producer.start();
		consumer.start();

		try {
			producer.join();
			consumer.join();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		long endTime = System.nanoTime();
		long diff = endTime - startTime;

		System.out.println("Finished in " + diff / 1000 + " us");
		long throughput = (long) ((double) dataAmount * batchSize / diff * 1000000000);
		System.out.println("Throughput: " + throughput + " messages per second");
	}
}
